package Liszt
//class used to build stack frame contexts, there are a few pieces in the code were we use this functionality
//and trying to refactor them might be a good idea

class ContextStack[K,V] {
	class Frame(val name : String) {
		val map = new scala.collection.mutable.HashMap[K,V]
		var unnamed_frames = 0
	}
	class Context(cur_stk : List[Frame]) {
		lazy val name = for(a <- cur_stk.reverse) yield a.name
	}
	private var stk : List[Frame] = List()
	
	def frame = stk.head
	
	def push() {
		val n = frame.unnamed_frames
		frame.unnamed_frames += 1
		push(n.toString)
	}
	def push(frame : String) {
		stk = new Frame(frame) :: stk
	}
	def pop() {
		stk = stk.tail
	}
	def context() : Context = new Context(stk)
	
	def apply(k : K) : V = {
		get(k).get
	}
	def get(k : K) : Option[V] = {
		for( (_,v) <- getInContext(k) )	
			yield v
	}
	//same as get but also returns the context in which the value was found
	def getInContext(k : K) : Option[(Context,V)] = {
		def g(frames : List[Frame]) : Option[(Context,V)] = {
			frames match {
				case h :: t => {
					h.map.get(k) match {
						case Some(v) => Some( (new Context(frames),v) )
						case None => g(t)
					}
				}
				case _ => None
			}
		}
		g(stk)
	}
	def update(k : K, v : V) {
		frame.map(k) = v
	}
	
	//TODO(zach): Provide the option to keep around old contexts such that you can ask
	//for the value of something in a particular old context
	push("<top>")
}