package scala.LisztPlugin

import scala.tools.nsc
import nsc.Global

import Liszt.Types.{Type => LType, _}
import Liszt.{Trees => L}
import Liszt.Log
import Liszt.Ops
import Liszt.ASTUtil

//TODO(zach): Error out if a ValDef for a mesh object/field/set is a var
trait TreeTranslation extends TypeTranslation {
private object Log extends Liszt.Log("TreeTranslation")
val global : Global
import global._

	object PApply {
		def unapply(t : Tree) :  Option[(Tree,List[Tree])] = t match {
			case Apply(ta @ TypeApply(fn,targs),args) => Some  ( (fn.duplicate copyAttrs ta, args) )
			case Apply(fn,args) => Some( (fn,args) )
			case _ => None
		}
	}
	def scalaToLiszt(rawtree : Tree, module_name : String) : (L.Block,List[String]) = { // (tree,used modules)
		
		object ctx {
			def push() {
				Log.log("entering block")
				stk.push(newbuf())
			}
			def pop() : List[L.Stmt] = {
				Log.log("exiting block")
				stk.pop().stmts.toList
			}
			def insert(s : L.Stmt) {
				stk.top.stmts += s
			}
			def newVar() : String = {
				name += 1
				"temp%d".format(name)
			}
			def typeOfOption(s : String) : Option[LType] = for(c <- stk.find(_.types contains s)) yield c.types(s)
			def typeOf(s : String) : LType = typeOfOption(s) match {
				case Some(s) => s
				case None => Log.ierror("variable not found: %s",s)
			}
			def defVar(s : String, t : LType) {
				stk.top.types(s) = t
			}
			class Context {
				val stmts = new scala.collection.mutable.ListBuffer[L.Stmt]
				val types = new scala.collection.mutable.HashMap[String,LType]
			}
			private def newbuf() = new Context
			private var stk = new scala.collection.mutable.Stack[Context]
			private var name = 0
			def isTopLevel : Boolean = stk.size == 1
		}
		val extern_modules = new scala.collection.mutable.HashSet[String]()
		
		def sanitizeVar(s : String) : String = s.replace("$","__")
		def mkVar(t : LType, s : String) : L.Var = L.Var(t,sanitizeVar(s))
		
		//perform some cleaning of the tree before match against it
		
		object Preprocess extends Transformer {
			def pretransform(tree : Tree) : Tree = {
				val post = tree match {
				case PApply(Select(Ident(predef),_),List(arg))  //remove implicit converstions in predef
					if predef.toString == "Predef" => pretransform(arg)
				case PApply(fn,args) if isImplicitType(fn.tpe) => pretransform(fn.duplicate copyAttrs tree)  //remove implicit applies, they are not needed after typechecking
				//case Apply(typeApply @ TypeApply(fn, targs),args) =>
				//	pretransform(Apply(fn.duplicate copyAttrs typeApply,args) copyAttrs tree) //remove the type application and pass its type down the tree
				case PApply(fn,List(arg)) if isImplicitPromotionFunc(fn) => pretransform(arg)
				case sel @ Select(_ : This, name) => pretransform(Ident(name) copyAttrs sel)
				case _ => tree 
				}
				//Log.log("Pree %s Post %s",tree,post)
				post
			}
			override def transform(tree : Tree) : Tree = pretransform(super.transform(tree))
		}
		
		object MatchName {
			def unapply(tree : Tree) : Option[String] = tree match {
				case Ident(x) => Some(x.toString)
				case Select(obj,nme) => matchSelect(tree,obj,nme.toString)
				case _ => None
			}
			def matchSelect(tree : Tree, obj : Tree, field : String) : Option[String] = {
				if(ToLType.isLisztObjectSym(obj.tpe.typeSymbol)) {
					val Ident(x) = obj
					extern_modules += x.toString
					Some(Liszt.ASTUtil.moduleName(x.toString,field))
				} else None
			}
		}
		object MatchLValue {
			def unapply(tree : Tree) : Option[L.Var] = 
				for(e <- MatchName.unapply(tree))
					yield mkVar(tree.tpe.lisztType,e)
		}
		
		object ::! { // match init ::! last
			def unapply[A](as : List[A]) : Option[(List[A],A)] = {
				if(!as.isEmpty)
					Some((as.init, as.last))
				else
					None
			}
		}
		object UpdateEntry {

			def unapply(t : Tree) : Option[(Tree,List[Tree],Ops.Reduce,Tree)] = {
			//field writes: f(x) = 4, f(x) += 4, get translated to f.update(x,4) and f.update(x,f(x) + 4)
				t match {
					case PApply( Select(objv @ MatchName(v1),update),
								args ::! 
									 PApply(
										Select(
											PApply(
												Select(MatchName(v2),apply),
												args2
											),
											BOp(binary_op)
										),
										List(v)
									 )
								
							   ) if (isUpdateable(objv.tpe.lisztType) && v1 == v2 && 
									apply.toString == "apply" && 
									update.toString == "update") => {
						val sameArgs = {
							if(args.size != args2.size)
								false
							else
								(args zip args2) forall {
									case (Ident(a),Ident(b)) => a == b
									case (Literal(a),Literal(b)) => a == b
									case _ => false
								}
						}
						if(sameArgs)
							Some((objv,args,Ops.Reduce(binary_op),v))
						else
							None
					}
					case _ => None
				}
			}
		}
		//f(exp1,exp2) += exp3 is translated to 
		// {
		//   val a = exp1
		//   val b = exp2
		//   f.update(a,b,f(a,b) + exp3)
		// }
		// we need to unroll this
		object ReduceUpdateBlock {
			def unapply(t : Tree) : Option[ (Tree, List[Tree], Ops.Reduce, Tree) ] = t match {
				case Block(vds, UpdateEntry(obj,args,op,v)) => {
					val nameswithexpOp = ToLType.liftOption { 
						vds map {
							case ValDef(mods,name,tpt,rhs) if mods.hasFlag(nsc.symtab.Flags.SYNTHETIC) => Some( (name,rhs) )
							case _ => None
						}
					}
					val namesOp = ToLType.liftOption  {
						args map {
							case Ident(n) => Some(n)
							case _ => None
						}
					}
					(nameswithexpOp,namesOp) match {
						case(Some(nameswithexp),Some(names)) => {
							if(nameswithexp.size != names.size)
								None
							else if( (nameswithexp zip names) forall {
								case ( (n1,_) , n2) if n1 == n2 => true
								case _ => false
							}) {
								Some((obj, nameswithexp map {_._2}, op, v))
							} else
								None
						}
						case _ => None
					}
				}
				case _ => None
			}
		}
		def tAssign(lval : L.LValue, rhs : Tree) : L.Assign = {
			def check(lval : L.LValue, lval2 : L.LValue) = {
				//Log.log("CHECKING: %s === %s : %s".format(lval,lval2,lval sameRef lval2))
				lval sameRef lval2
			}
			rhs match {
				case PApply(Select(MatchLValue(lval2),BOp(op)),List(rhs2)) if check(lval,lval2) => {
					L.Assign(lval,Ops.Reduce(op),tVar(rhs2))
				}
				case _ => L.Assign(lval,Ops.Assign,tVar(rhs))
			}
		}
		def tBlockStmt(b : Tree) : L.Block = {
			val result = if(willBeBlock(b)) {
				ASTUtil.blockWrap(tStmt(b))
			} else {
				ctx.push()
				ctx insert tStmt(b)
				L.Block(ctx.pop(),L.EmptyExp(UnitT))
			} 
			result withPosition scalaPosToLisztPos(b.pos)
		}
		
		def tBlockExp(b : Tree) : L.Block = {
			ctx.push()
			val exp = tVar(b)
			L.Block(ctx.pop(),exp) withPosition scalaPosToLisztPos(b.pos)
		}
		def tBlockOrExp(tree : Tree) : L.Block = {
			tree match {
				case Block(es,e) => {
					ctx.push()
					for(e <- es)
						ctx insert tStmt(e)
					val exp = tSimpleExp(e)
					val stmts = ctx.pop()
					L.Block(stmts,exp)
				}
				case _ => {
					ctx.push()
					val v = tSimpleExp(tree)
					L.Block(ctx.pop(),v)
				}
			}
		}
		def tFuncBlock(tree : Tree) : L.Block = {
			ctx.push()
			val exp = tree match {
				case Block(es,exp) => {
					for(e <- es)
						ctx insert tStmt(e)
					tStmt(exp)
				}
				case _ => {
					tStmt(tree)
				}
			}
			val stmt = exp match {
				case L.ExpStmt(e) => {
					if(e.t != Liszt.Types.UnitT)
						L.Return(asVar(e))
					else
						exp
				}
				case x =>  x
			}
			ctx insert stmt
			L.Block(ctx.pop(),L.EmptyExp(UnitT))
		}
		def tUpdateStmt(obj : Tree, args : List[Tree], op : Ops.AssignOp, v : Tree) : L.Stmt = {
			obj.tpe.lisztType match {
				case _ : FieldT => 
					L.FieldWrite(tVar(obj),tVar(args.head),op,tVar(v))
				case _ : SparseVectorType => {
					val idx = tVar(args.head)
					val vec = tVar(obj)
					val value = tVar(v)
					L.Assign(L.SparseVectorAccess(value.t,vec, idx),op,value)
				}
				case _ : SparseMatrixType => {
					val row = tVar(args(0))
					val col = tVar(args(1))
					val mat = tVar(obj)
					val value = tVar(v)
					L.Assign(L.SparseMatrixAccess(value.t,mat,row,col),op,value)
				}
				case vt : VectorT => 				
					L.Assign(L.VecExtract(vt.t,tVar(obj),tMetaIntOrIntVec(args.head)),op,tVar(v))
				case vt : MatrixT => 
					L.Assign(L.MatExtract(vt.t,tVar(obj),tMetaIntOrInt(args(0)),tMetaIntOrInt(args(1))),op,tVar(v))
				case a => Log.ierror("unexpected type %s",a)
			}
		}
		
		def willBeBlock(tree : Tree) : Boolean = tree match {
			case ReduceUpdateBlock(_,_,_,_) => false
			case Block(es,e) => true
			case _ => false
		}
		
		def tStmt(tree : Tree) : L.Stmt = { 
			//val tree = preprocess(ttree)
			lazy val ttype = tree.tpe.lisztType //lazy to prevent exception when we don't translate the type
			lazy val stype = tree.symbol.tpe.lisztType
			Log.log("Matching Stmt: %s",tree.toString.take(60))
			val result = tree match {
				case ReduceUpdateBlock(objv,argv,rop,v) => tUpdateStmt(objv,argv,rop,v)
				case Block(es,e) => {
					ctx.push()
					//add all the variable definitions to the context
					//currently not needed as we can get all type from the tree directly
					//for(ValDef(mods,name,tpt,rhs) <- es)
					//	ctx.defVar(name.toString,rhs.tpe.lisztType)
					for(e <- es)
						ctx insert tStmt(e)
					val stmt = tStmt(e)
					L.Block(ctx.pop() ::: List(stmt),L.EmptyExp(UnitT))
				}
				case ValDef(mods,name,tpt,rhs) => {
					val dv = L.DefVar(sanitizeVar(name.toString),tExp(rhs))	
					
					
					if(mods.isVariable)
						dv.init.t match {
							case _ : BasicType => 
							case _ : VectorT =>
							case _ : MatrixT =>
							case _ => reportError(tree,"only basic types (int,float,string,vector,matrix) can be vars, use vals instead")
						}
					dv
					
				}
				case UpdateEntry(objv,argv,rop,v) => tUpdateStmt(objv,argv,rop,v)

				case PApply(Select(obj,update), as ::! v) if update.toString == "update" && 
					isUpdateable(obj.tpe.lisztType)
					 => tUpdateStmt(obj,as,Ops.Assign,v)
				case Return(v) => {
					val blk = tBlockOrExp(v)
					L.Block(blk.stmts ::: List(L.Return(blk.exp)),L.EmptyExp(UnitT))
				}
				case If(p,t,e) => {
					val pl = tBlockExp(p)
					val tl = tBlockStmt(t)
					val el = tBlockStmt(e)
					L.If(pl,tl,el)
				} 
				case _ => tree match {//https://lampsvn.epfl.ch/trac/scala/ticket/425
				case While(p,e) => {
					val pl = tBlockExp(p)
					val el = tBlockStmt(e)
					L.While(pl,el)
				}
				case PApply(Select(st,foreach),List(Function(List(ValDef(_,nme,_,_)),blk))) if foreach.toString == "foreach" &&
					isSet(st.tpe.lisztType) => {
					//Log.log("Set: %s",st)
					//Log.log("Blk: %s",blk)
					(st,blk) match {
						case ( PApply(Select(st,filter),List(_:Function)) ,
						      Match(_,List(CaseDef(
						                             UnApply( Apply(
						                                             TypeApply(Select(withLabel,_),_)
						                                             ,lst
						                                           ),
						                                      List(
						                                            Bind(mo,_), 
						                                            Bind(lbl,_)
						                                          )
						                                    ),
						                             guard,
						                             blk
						                           ) 
						                    )
						            )
						     ) if filter.toString == "filter" && isLanguageName(withLabel,"withLabel") => {
						  val set = tVar(st)
						  val mov = ASTUtil.emptyDefVar(sanitizeVar(mo.toString),set.t.asInstanceOf[Set].t)
						  val lblv = ASTUtil.emptyDefVar(sanitizeVar(lbl.toString),IntT)
						  val block = tBlockStmt(blk)
						  L.Forall(mov,Some(lblv),set,block)
						}
						case _ => {
							val block = tBlockStmt(blk)
							val set = tVar(st)
							val mov = ASTUtil.emptyDefVar(sanitizeVar(nme.toString),set.t.asInstanceOf[Set].t)
							L.Forall(mov,None,set,block)
						}
					}
				}
				case PApply(sel @ Select(st,nonzeroes),List(Function(List(ValDef(_,a_nme,_,_)),blk))) if nonzeroes.toString == "nonzeroes" &&
					sel.symbol.owner == definitions.getClass("Liszt.Language.LinearSystem") => {
					
					val block = tBlockStmt(blk)
					val ls = st.symbol.name.toString
					val a = ASTUtil.emptyDefVar(sanitizeVar(a_nme.toString),NonzeroesSpecifierType(ls))
					L.LinearSystemNonzeroes(ls,a,block)
				}
				case _ => tree match {
				case PApply(x,args) if isLanguageName(x,"Print") => L.Print(args.map(tVar))
				case PApply(x,args) if isLanguageName(x,"FieldSave") => L.FieldSave(tVar(args(0)),tVar(args(1)))
				case DefDef(mods,name,tparams,vparamss,tpt,dd) if isLanguageName(dd,"__") =>  {
					val ftype = stype.asInstanceOf[Func]
					val def_vars = for( (name,typ) <- (vparamss.head.map(_.name.toString) zip ftype.args) )
						yield ASTUtil.emptyDefVar(name,typ)
					L.DefFunc(stype,name.toString,def_vars,L.ExternStmt)
				}
				case DefDef(mods,name,tparams,vparamss,tpt,rhs) =>  {
					stype match {
						case Func(_, _ : Set) => reportError(tree, "Mesh sets cannot be returned from functions %s".format(tree))
						case _ =>
					}
					if(vparamss.length != 1)
						reportError(tree, "Functions must have exactly 1 argument list.")
					val ftype = stype.asInstanceOf[Func]
					val arg_list = for( (name,typ) <- vparamss.head.map(_.name.toString) zip ftype.args)
						yield ASTUtil.emptyDefVar(name,typ)
					val fn_name = name.toString
					if(isLanguageName(rhs,"AutoIndex")) {
						val Func(topo_list,ret) = stype
						L.Index(fn_name, topo_list, ret)
					} else {
						L.DefFunc(stype,fn_name,arg_list,tFuncBlock(rhs))
					}
				}
				case app @ PApply(sel @ Select(obj,field),List(rhs)) if app.symbol.hasFlag(nsc.symtab.Flags.ACCESSOR) => {
					val field_name_m = "(.*)_\\$eq".r
					val field_name_m(field_name) = field.toString
					val lname = MatchName.matchSelect(rhs,obj,field_name).get
					tAssign(mkVar(rhs.tpe.lisztType,lname),rhs)
				}
				case PApply(Select(nz,x),args) if isNonzero(nz.tpe) => {
					val i = x.toString match {
						case "A" => L.LS_A
						case "x" => L.LS_X
						case "b" => L.LS_B
					}
					L.LinearSystemPlaceNonzero(tVar(nz),i,args.map(tVar))
				}
				case Assign(MatchLValue(lval),rhs) => tAssign(lval,rhs)
				case _ => L.ExpStmt(tExp(tree))
				}
				}
			}
			Log.log("Matched: ---> %s",result.toShortString)
			result withPosition scalaPosToLisztPos(tree.pos)
		}
		//create a new variable and bind it to exp, return the variable
		def asVar( e : L.Exp ) : L.Var = e match {
			//val a = e
			//e
			case v @ L.Var(_,_) => v
			case _ => {
				val name = ctx.newVar()
				ctx insert (L.DefVar(name,e) withSynthetic true)
				mkVar(e.t,name)
			}
		}
		//utility: first translate to expression then make a var
		def tVar(tree : Tree) : L.Var = asVar(tExp(tree))
		def tSimpleExp(tree : Tree) : L.SimpleExp = tExp(tree) match {
			case s : L.SimpleExp => s
			case u if u.t == UnitT => {
				ctx insert L.ExpStmt(u)
				L.EmptyExp(UnitT)
			}
			case x => asVar(x)
		}
		def tMetaIntOrIntVec(tree : Tree) : L.VecExtractArgument = tree.tpe.lisztTypeOption match {
			case Some(IntT) => tExp(tree) match {
				case x : L.IntL => x
				case x => asVar(x)
			}
			case _ => L.IntL(mIntValueToInt(tree)) //deprecated
		}
		def tMetaIntOrInt(tree : Tree) : L.Var = tree.tpe.lisztTypeOption match {
			case Some(IntT) => tVar(tree)
			case _ => asVar(L.IntL(mIntValueToInt(tree))) //deprecated
		}
		def tExp(tree : Tree) : L.Exp = {
			//val tree = preprocess(ttree)
			lazy val ttype = tree.tpe.lisztType //lazy to prevent exception when we don't translate the type
			lazy val stype = tree.symbol.tpe.lisztType
			

			//Log.log("Matching Exp: %s",tree.toString.take(60))
			val result = tree match {
				case Literal(x) => ttype match {
					case IntT => L.IntL(x.intValue)
					case FloatT => L.FloatL(x.floatValue)
					case DoubleT => L.DoubleL(x.doubleValue)
					case BoolT => L.BoolL(x.booleanValue)
					case StringT => L.StringL(x.stringValue)
					case UnitT => L.EmptyExp(UnitT)
				}
				
				case PApply(Select(a,BOp(op)),List(b)) => {
					val av = tVar(a)
					val bv = tVar(b)
					if(av.t == StringT || bv.t == StringT) {
						reportError(tree,"String concatenation is not supported")
					}
					def isNumeric(t : LType) = t == IntT || t == FloatT || t == DoubleT
					if( (op == Ops.Eq || op == Ops.NEq) && av.t != bv.t && (!isNumeric(av.t) || !isNumeric(bv.t))) {
						reportError(tree,"type mismatch in equality comparison: %s %s".format(av.t,bv.t))
					}
					L.Binary(ttype,av,op,bv) // a + b
				}
				case Select(a,UOp(op)) => L.Unary(ttype,op,tVar(a)) // -a
				case PApply(Select(x,apply),args) if isLanguageName(x,"Vec") && apply.toString == "apply" =>
					L.VectorL(ttype,args.map(tVar))
				case PApply(Select(x,apply),args) if isLanguageName(x,"Mat") && apply.toString == "apply" => {
					val largs = args map tExp
					if(largs.forall(x => x.isInstanceOf[L.VectorL])) {
						val inits = largs.flatMap {
							case L.VectorL(_,as) => as
							case _ => Log.ierror("unexpected case")
						}
						L.MatrixL(ttype,inits)
					} else {
						L.MatrixFromVec(ttype,largs map asVar)
					}
				}
				case PApply(Select(obj,apply),args) if apply.toString == "apply" =>
					obj.tpe.lisztType match {
						case FieldT(_,_) => L.FieldRead(ttype,tVar(obj),tVar(args.head))
						case MatrixT(_,_,_) => L.MatExtract(ttype,tVar(obj),tMetaIntOrInt(args(0)),tMetaIntOrInt(args(1)))
						case VectorT(_,_) =>  L.VecExtract(ttype,tVar(obj),tMetaIntOrIntVec(args.head))
						case _ : SparseVectorType => {
							L.SparseVectorAccess(ttype,tVar(obj),tVar(args.head))
						}
						case _ : SparseMatrixType => {
							val row = tVar(args(0))
							val col = tVar(args(1))
							L.SparseMatrixAccess(ttype,tVar(obj),row,col)
						}
						case x => Log.ierror("unexpected type: %s",x)
					}
				
				case PApply(x,List(arg)) if isLanguageName(x,"FieldWithConst") => {
					if(!ctx.isTopLevel)
						reportError(tree,"field initialization must occur at the object level only")
					else
						L.FieldInit(ttype,L.ConstSource(tVar(arg)))
				}
				case PApply(x,List(Literal(str))) if isLanguageName(x,"FieldWithLabel") && str.tag == StringTag => {
					if(!ctx.isTopLevel)
						reportError(tree,"field initialization must occur at the object level only")
					else
						L.FieldInit(ttype,L.StringSource(str.stringValue))
				}
				case PApply(x,List(Literal(str))) if isLanguageName(x,"BoundarySet") && str.tag == StringTag => {
					if(!ctx.isTopLevel)
						reportError(tree,"set initialization must occur at the object level only")
					else
						L.SetInit(ttype,str.stringValue)
				}
				case x if isLanguageName(x,"mesh") => 
					L.Call(Mesh,L.Var(Func(List(),Mesh),"mesh"),List())
				
				case _ => tree match { ////https://lampsvn.epfl.ch/trac/scala/ticket/425
				case PApply(TopCall(tfn),args) => L.TopoCall(ttype,tfn(args.map(_.tpe.lisztType)),args.map(tVar)) //cells(args)
				
				
				case Apply(ta @ TypeApply(fn,targs),args) => {
					val types = targs.map(_.tpe.lisztType)
					L.Call(ttype,tVar(fn.duplicate copyAttrs ta),args.map(tVar)) withTypeParameters(types)
				}
				case Apply(fn,args) => L.Call(ttype,tVar(fn),args.map(tVar)) //misc_fn(args)
				case Select(v,x) if isVector(v.tpe) => {
					val i = x.toString match {
							case "x" => 0
							case "y" => 1
							case "z" => 2
							case "w" => 3
							case _ => Log.ierror("unknown vector extractor: %s",x)
						}
					L.VecExtract(ttype,tVar(v), L.IntL(i))
				}

				case Select(x,typeConvert) if isTypeConversion(typeConvert) => L.Call(ttype,mkVar(Func(List(x.tpe.lisztType),ttype),typeConvert.toString),List(tVar(x))) 
				case If(p,t,e) => {
					val bl = tBlockExp(p)
					val tl = tBlockOrExp(t)
					val el = tBlockOrExp(e)
					assert(tl.exp.t == el.exp.t)
					val tpt = tl.exp.t 
					L.IfExp(tpt,bl,tl,el) //we cannot use ttype here because sometimes if-statments are given a less specific type
				}
				case _ => tree match {
					case MatchLValue(v) => v
					case _  => asLanguageName(tree) match {
						case Some(name) => mkVar(ttype,name)
						case None => Log.ierror("unhandled scala expression: %s",tree)
					} 
				}
					
				
				
				//case TypeApply(Select(mapinit,asInstanceOf),List(arg))
				//	if isLanguageName(mapinit,"MapInit") && asInstanceOf.toString == "asInstanceOf" =>
				//		L.MapInit(ttype)
				
				
				}
			}
			//Log.log("Matched Exp: %s -----> %s",tree.toString.take(60),result.toShortString)
			result withPosition scalaPosToLisztPos(tree.pos)
		}
		
		val tree = Preprocess.transform(rawtree)
		val toplevel = tBlockStmt(tree)
		(L.Block(toplevel.stmts,toplevel.exp),extern_modules.toList)
	}
	def isField(t : LType) : Boolean = t match {
		case FieldT(_,_) => true
		case _ => false
	}
	def isUpdateable(t : LType) : Boolean = t match {
		case _ : FieldT => true
		case _ : SparseVectorType => true
		case _ : SparseMatrixType => true
		case _ : VectorT => true
		case _ : MatrixT => true
		case _ => false
	}
	def isSet(t : LType) : Boolean = t match {
		case Set(_) => true
		case _ => false
	}
	def isVector(t : Type) : Boolean = t.lisztTypeOption match {
		case Some(VectorT(_,_)) => true
		case _ => false
	}
	def isNonzero(t : Type) : Boolean = t.lisztTypeOption match {
		case Some(_ : NonzeroesSpecifierType) => true
		case _ => false
	}
	
	def asLanguageName(t : Tree) : Option[String] = t match {
		case Select(Select(Ident(liszt),lang),x) if liszt.toString == "Liszt" &&
			lang.toString == "Language" => Some(x.toString)
		case _ => None
	}
	def isImplicitPromotionFunc(t : Tree) : Boolean = asLanguageName(t) match {
		case Some(x) => x startsWith "implicit"
		case None => false
	}
	def isLanguageName(t : Tree, str : String) : Boolean = asLanguageName(t) match {
		case Some(x) => x == str
		case None => false
	}
	def isTypeConversion(t : Name) : Boolean = {
		val conversions = List("toByte","toChar","toDouble","toFloat","toInt","toLong","toShort")
		conversions.contains(t.toString)
	}
	def scalaPosToLisztPos(p : nsc.util.Position) : Option[Liszt.Trees.Position] = {
		if(p.isDefined) {
			Some(Liszt.Trees.Position(p.source.toString,p.line,p.column,p.lineContent))
		} else {
			None
		}
	}
	
	def reportError(tree : Tree, msg : String) : Nothing = {
		val str = scalaPosToLisztPos(tree.pos) match {
			case Some(x) => x.formatString.format(msg)
			case None => msg
		}
		Log.error(msg)
	}
	object TopCall {	
		import Ops._
		type result_t = List[LType] => TopoFn
		val funcs = List( "vertices" -> Vertex,
			                  "edges" -> Edge,
			                  "faces" -> Face,
			                  "cells" -> Cell )
		val dirs = List( "" -> NoDir,
			             "CCW" -> CCW,
			             "CW" -> CW)
		val mapping : Map[String,result_t] = Map.empty ++ {
			for((f,ff) <- funcs; (d,dd) <- dirs) 
				yield f+d -> ( (args:List[LType]) => TopoSet(ff,dd,args.head.asInstanceOf[Topo]))
		}
		def unapply(tree : Tree) : Option[result_t] = {
			asLanguageName(tree) match {
				case Some(x) => x match {
					case "head" => Some(args => Head(args.head.asInstanceOf[Topo]))
					case "tail" => Some(args => Tail(args.head.asInstanceOf[Topo]))
					case "inside" => Some(args => Tail(args.head.asInstanceOf[Topo]))
					case "outside" => Some(args => Head(args.head.asInstanceOf[Topo]))
					case "flip" => Some(args => Flip(args.head.asInstanceOf[Topo]))
					case "towards" => Some(args => Towards(args(0).asInstanceOf[Topo],args(1).asInstanceOf[Topo]))
					//label extractors
					case "vertex" => Some(args => LabelExtract(Vertex,args(0).asInstanceOf[Topo]))
					case "edge" => Some(args => LabelExtract(Edge,args(0).asInstanceOf[Topo]))
					case "face" => Some(args => LabelExtract(Face,args(0).asInstanceOf[Topo]))
					case "cell" => Some(args => LabelExtract(Cell,args(0).asInstanceOf[Topo]))
					case "ID" => Some(args => IDLookup(args(0).asInstanceOf[Topo]))
					case s => mapping.get(s)
				}
				case _ => None
			}
		}
	}
	object UOp {
		import Ops._
		val pattern = "unary_(.+)".r
		def unapply(nme : Name) : Option[Ops.UnOp] = nme.toString match {
			case pattern(x) => x match {
				case "$tilde" => Some(BNot)
				case "$minus" => Some(Neg)
				case "$bang" => Some(Not)
				case _ => None
			}
			case _ => None
		}
	}
	object BOp {
		import Ops._
		def unapply(nme : Name) : Option[Ops.BinOp] = nme.toString match {
			case "$plus" => Some(Plus)
			case "$minus" => Some(Minus)
			case "$times" => Some(Times)
			case "$div" => Some(Div)
			case "$percent" => Some(Mod)
			case "$eq$eq" => Some(Eq)
			case "$bang$eq" => Some(NEq)
			case "$less" => Some(Less)
			case "$less$eq" => Some(LessEq)
			case "$greater" => Some(Greater)
			case "$greater$eq" => Some(GreaterEq)
			case "$times$less$times" => Some(Min)
			case "$times$greater$times" => Some(Max)
			case "min" => Some(Min)
			case "max" => Some(Max)
			case "$less$less" => Some(LShift)
			case "$greater$greater" => Some(RShift)
			case "$bar" => Some(BOr)
			case "$amp" => Some(BAnd)
			case "$up" => Some(BXor)
			case "$amp$amp" => Some(And)
			case "$bar$bar" => Some(Or)
			case _ => None
		}
	}
	object While {
		def unapply(tree : Tree) : Option[(Tree,Tree)] = tree match {
			//this matches the de-sugared while loop that scala's parser generates
			case LabelDef(name,List(),If(cond,Block(es,PApply(Ident(name2),List())),Literal(l))) 
			if name == name2 && l.tag == UnitTag => {
				val blk = es.length match {
					case 0 => Literal(())
					case 1 => es.last
					case n => Block(es.init,es.last)
				}
				Some((cond,blk))
			}
			case _ => None
		}
	}
} //trait TreeTranslation
