package Liszt
import scala.collection.immutable.{Set => SSet}
import Liszt.Types._
import Liszt.Trees._

//
//Introduces a copy-by-value of variables used inside forall loops,
//thus writing to globals from a nested context will not work.
class CloseForalls(top : CompilationUnit) {
	
		
	// PASS 1:
	//Generate the list of nested and unnested funcs,
	//and create the AST nodes for each context of the function
	val (nested_funcs, unnested_funcs, renamed_code) = {
		val defined_functions_map = Map.empty ++ { for(df @ DefFunc(_,name,_,stmt) <- top.code if stmt != ExternStmt) yield name -> df }
		val defined_functions = defined_functions_map.keySet
		case class Status(var nested : Option[String], var unnested : Option[String])
		val status = Map.empty ++ { for(df <- defined_functions) yield df -> new Status(None,None) }
		val main_fn = "_liszt_start"
		def examineFunction(fn_name : String, nested : Boolean) {
			if(nested && status(fn_name).nested == None) {
				status(fn_name).nested = Some(ASTUtil.Names.tag(fn_name,"nested"))
				for(Call(_,fn,_) <- defined_functions_map(fn_name) if defined_functions.contains(fn.name)) {
					examineFunction(fn.name,true)
				}
			} else if(!nested && status(fn_name).unnested == None) {
				status(fn_name).unnested = Some(if(fn_name == main_fn) fn_name else ASTUtil.Names.tag(fn_name,"unnested"))
				
				def find_calls(tree : Tree) {
					tree match {
						case Call(_,fn,_) if defined_functions.contains(fn.name) =>
							examineFunction(fn.name,false)
						case Forall(_,_,_,blk) =>
							for(Call(_,fn,_) <- blk if defined_functions.contains(fn.name))
								examineFunction(fn.name,true)
						case _ => tree match {
							case ShapeAny(t,names,exps) => exps.foreach(find_calls)
						}
					}
				}
				
				find_calls(defined_functions_map(fn_name))
			}
		}
		//generate initial states forall functions calls
		examineFunction(main_fn,false)
		//println(status)
		def descend(tree : Tree, nested : Boolean) : Tree = {
			tree match {
				case fa @ Forall(_,_,_,block) => fa.copy(block = descend(block,true).asInstanceOf[Block])
				case call @ Call(_,fn,_) if defined_functions.contains(fn.name)=> {
					val new_name = if (nested) {
						status(fn.name).nested.get
					} else {
						status(fn.name).unnested.get
					}
					call.copy(func = fn.copy(name = new_name))
				}
				case _ => { tree match {
					case ShapeAny(t,names,exps) => ShapeAny.trees(tree,for(e <- exps) yield descend(e,nested))
				}}
			}
		}
		val non_defs = for(stmt <- top.code.stmts if !stmt.isInstanceOf[DefFunc])
			yield stmt
		
		val defs = status flatMap {
			case (fn_name,s) =>
			def mkFunc(nested : Boolean, list : List[String] ) : List[Stmt] = {
				for(name <- list) yield {
					val cp = ASTUtil.duplicate(defined_functions_map(fn_name)).asInstanceOf[DefFunc].copy(name = name)
					descend(cp,nested).asInstanceOf[Stmt]
				}
			}
			val nested_fn = mkFunc(true,s.nested.toList)
			val unnested_fn = mkFunc(false,s.unnested.toList)
			nested_fn ::: unnested_fn
		}
		val nested_fns  = SSet.empty ++ { for( (k,s) <- status; e <- s.nested) yield e }
		val unnested_fns  = SSet.empty ++ { for( (k,s) <- status; e <- s.unnested) yield e }
		(nested_fns,unnested_fns, Block(defs.toList ::: non_defs,top.code.exp))
	}

	// PASS 2:
	// In unnested functions, replace Foralls with Closures 
	var forallFuncts : List[DefFunc] = List()
	val global_variables = SSet.empty ++ { for(dv @ DefVar(name,_) <- renamed_code.stmts if !dv.synthetic) yield name }
	val defined_variables = SSet.empty ++ { for(DefVar(name,_) <- renamed_code) yield name }
	val (code,global_writes_map) = {
		val global_writes = new GlobalWritesForWorkgroup(renamed_code)
		var global_writes_info : List[(Int,SSet[String])] = List.empty
		def decend(tree : Tree) : Tree = tree match {
			case df : DefFunc if nested_funcs contains df.name => df
			case fa @ Forall(name, lblop, set, block) => {
				//find all variables in block that are not local to nor global
				val local_vars = SSet.empty ++ { for(DefVar(name,_) <- fa) yield name}
				val local_uses = SSet.empty ++ { for(v @ Var(t,name) <- block) yield (t, name, v.global)}
				val not_fvars = local_vars ++ global_variables
				val function_vars = for ((u_t, u_n,u_g) <- local_uses.toList if ! not_fvars.contains(u_n) && defined_variables.contains(u_n)) yield (u_t, u_n, u_g)
				val function_var_names = SSet.empty ++ { set.name :: { for ((u_t, u_n,_) <- function_vars) yield u_n} }
				
				def uniqueify(tree : Tree) : Tree = tree match {
					case v : Var if function_var_names.contains(v.name)=> v.copy(name = ASTUtil.Names.tag(v.name, "passed")) withGlobal v.global
					case dv : DefVar if function_var_names.contains(dv.name) => dv.copy(name = ASTUtil.Names.tag(dv.name, "passed")) withGlobal dv.global
					case _ => { tree match {
						case ShapeAny(_,names,exps) => ShapeAny.trees(tree,for(e <- exps) yield uniqueify(e))
					} }
				}
				
				//make new function with those as variables including set as the first argument
				val argtypes = set.t :: { for((u_t, u_n, u_g) <- function_vars) yield u_t }
				val argnames = DefVar(set.name,EmptyExp(set.t)) :: { for((u_t, u_n, u_g) <- function_vars) yield DefVar(u_n,EmptyExp(u_t)) withGlobal u_g }
				val ftype = Func(argtypes, UnitT)
				
				//TODO(zach): For our first pass at synchronization, we are adding workgroups for each forall loop, even if many can execute at once
				//The workgroup objects are inserted here so that abstract interpretation does the right thing.
				//eventually the workgroup objects should be inserted using a better algorithm that is seperate from the compiler pass
				
				val wg = Workgroup(List(fa))
				global_writes_info ::= wg.id -> global_writes.writesInTree(fa)
				val fname = "forallname_%d".format(wg.id)
				
				val df = DefFunc(ftype, fname, argnames, wg)
				val uniquifiedDf = uniqueify(df).asInstanceOf[DefFunc]
				
				//add the new function to global_variables
				forallFuncts ::= uniquifiedDf
				
				//return a new call statement calling the function 
				val callargs = Var(set.t, set.name) :: { for((u_t, u_n, u_g) <- function_vars) yield Var(u_t, u_n) withGlobal(u_g) }
				ExpStmt(Call(UnitT, Var(ftype, fname), callargs))
			}
			case _ => ShapeAny.trees(tree,tree.children map decend)
		}
		val treeSmts = decend(renamed_code).asInstanceOf[Block].stmts
		(Block(forallFuncts ::: treeSmts, EmptyExp(UnitT)), Map.empty ++ global_writes_info)
	}

}