/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */
package Liszt
import scala.collection.immutable.{Set => SSet}
import Liszt.Types._
import Liszt.Trees._

class CodeGen(val top : CompilationUnit, globals : SSet[String], runtime_header : String, output_dir : String) extends CLikeCodeGen {
	import ASTUtil._
	import CodeLineImplicits._
	val Log = new Log("CodeGen")
	val liszt_obj = "__liszt"
	val fluent_file = "__fluent"
	
	
	val all_indices_size = "<<Unimplemented>>"
	val all_indices_offset = "<<Unimplemented>>"
	
	def tFunctionHeader(e : DefFunc, isDecl : Boolean) = {
		val fnt = getCType(e).asInstanceOf[Func]
		val name = e.name
		val args = for(a <- e.args) yield a.name
		val template_header = if(fnt.args.exists(_.isInstanceOf[SpecialType])) {
			val template_args = for( SpecialType(_,typ) <- fnt.args) yield "typename %s".format(typ) 
			"template< %s > ".format(template_args.mkString(","))
		} else {
			""
		}
		val argwt = for( (t,a) <- fnt.args zip args) yield "%s %s".format(tType(t),a)
		val return_type = tType(fnt.range)
		if(isDecl)
			"%sstatic %s %s(%s)".format(template_header,return_type,name,argwt.mkString(","))
		else
			"template<bool __liszt_nested>%s %s LisztCode<__liszt_nested>::%s(%s)".format(template_header,return_type,name,argwt.mkString(","))
	}
	var forall_depth = 0 //depth of static forall nesting, used to determine what argument to pass
	                     //to nested template
	
	def tIndex(e : Index) = Log.ierror("unexpected tIndex")
	def isPointerType(t : Type) = t match {
		case _ : SparseVectorType => true
		case _ : SparseMatrixType => true
		case _ => false
	}
	override def tPrintStream() = if (forall_depth > 0) "std::cout" else "if(__mpi_rank == 0 || __liszt_nested) std::cout"
	
	override def tCall(e : Call) = {
		val prefix = if(!isExternal(e.func.name)) {
			val nest = if(forall_depth > 0) "true" else "__liszt_nested"
			"LisztCode<%s>::".format(nest,tVar(e.func))
		} else ""
		prefix + super.tCall(e)
	}
	
	def tFieldInit(e : FieldInit) = Log.ierror("field init unexpected.")
	def tSetInit(e : SetInit) = Log.ierror("set init unexpected.")
	def tIfExp(e : IfExp) = Log.ierror("unexpected ifexp")
	
	override def tTypeFor(e : Tree) = e match {
		case dv : DefVar if globals.contains(dv.name) => 
			"LISZT_RUNTIME_NAMESPACE::Global< %s , %s >".format(super.tTypeFor(e),if(forall_depth > 0) "true" else "__liszt_nested")
		case _ => super.tTypeFor(e)
	}
	
	override def tVar(e : Var) = if(globals.contains(e.name)) "%s.readEntry()".format(super.tVar(e)) else super.tVar(e)
	
	override def tDefVar(e : DefVar) = {
		val DefVar(name,init) = e
		def declaration = super.tDefVar(DefVar(e.name,EmptyExp(init.t)) copyAttr e)
		def v = tVar(Var(init.t,name)) //variable refering to this declaration
		val result : List[CodeLine] = init match {
			case fi @ FieldInit(_,init) => {
				//TODO(zach): handle cases where field init is not immediately assigned to a variable
				//the name matching nonsense below is a hack to initialize fields correctly
				//by connecting them to descriptors.  In the future, descriptors will be allocation-context
				//based rather than variable-based to remove this issue.
				val real_name = ASTUtil.nameFromTempHack(name)
				init match {
					case ConstSource(exp) => "%s.buildFieldWithConst(%d,\"%s\",%s,&%s);".format(liszt_obj,fi.id,real_name,tVar(exp),real_name)
					case StringSource(url) => if(url == "position") {
						"%s.buildFieldWithPositions(%d,\"%s\",&%s);".format(liszt_obj,fi.id,real_name,real_name)
					} else {
						Log.error(fi.report("url not supported"))
					}
				}
			}
			case SetInit(t,set_name) => {
				val real_name = ASTUtil.nameFromTempHack(name)
				"%s %s;".format(tTypeFor(e),name) :::
				"%s.buildSet<%sType>(\"%s\",&%s);".format(liszt_obj,tType(t.asInstanceOf[Set].t),set_name,name)
			}
			/* TODO: Replace with LinearySystemAllocate node
			case New(svt @ SparseVectorType(t)) => {
				val temp = Names.create(name + "_newobj")
				//"boost::numeric::ublas::mapped_vector< %s > %s(%s);".format(tType(t),temp,all_indices.size) :::
				"PETSc::SparseVector< %s > %s(%s,%s);".format(tType(t),temp,all_indices_size,all_indices_offset) :::
				"%s %s = & %s;".format(tType(svt),name,temp)
			}
			case New(svt @ SparseMatrixType(t)) => {
				val temp = Names.create(name + "_newobj")
				//"boost::numeric::ublas::mapped_matrix< %s > %s(%s,%s);".format(tType(t),temp,all_indices.size,all_indices.size) :::
				"PETSc::SparseMatrix< %s > %s(%s,%s);".format(tType(t),temp,all_indices_size,all_indices_size) :::
				"%s %s = & %s;".format(tType(svt),name,temp)
			}
			case exp if isSparseVector(exp.t) => {
				val t = exp.t.asInstanceOf[SparseVectorType].t
				val temp = Names.create(name + "_newobj")
				//"boost::numeric::ublas::mapped_vector< %s > %s = %s;".format(tType(t),temp,tExp(exp)) :::
				val e1 = exp match {
					case Binary(_,Var(SparseMatrixType(_),lhs),op,Var(SparseVectorType(_),rhs)) => {
						op match {
							case Ops.Times => "MatMult(%s->internal(),%s->internal(),%s.internal());".format(lhs,rhs,temp)
							case Ops.Div => "PETScSolve(*%s,%s,*%s);".format(lhs,temp,rhs)
							case _ => Log.ierror("unimplemented")
						}
					}
					case _ => Log.ierror("unimplemented")
				}
				"PETSc::SparseVector< %s > %s(%s,%s);".format(tType(t),temp,all_indices_size,all_indices_offset) :::
				e1 :::
				"%s %s = & %s;".format(tType(exp.t),name,temp)
			}
			case exp if isSparseMatrix(exp.t) => {
				val t = exp.t.asInstanceOf[SparseMatrixType].t
				val temp = Names.create(name + "_newobj")
				//"boost::numeric::ublas::mapped_matrix< %s > %s = %s;".format(tType(t),temp,tExp(exp)) :::
				Log.ierror("unimplemented")
				"%s %s = & %s;".format(tType(exp.t),name,temp)
			} */
			case _ => super.tDefVar(e)
		}
		result
	}
	
	override def tAssign(e : Assign) = {
		val Assign(name,op,rhs) = e
		if(isFieldT(rhs.t)) List() //we ignore field assignments generated by seperating defs from decls
		else name match {
			case v : Var if globals.contains(v.name) => {
				val lhs = if(forall_depth > 0) "%s.writeEntry()" else "%s.readEntry()"
				super.tAssign(e.copy(lv = Var(v.t,lhs.format(v.name))))
			}
			case ve @ VecExtract(t,v,_) if globals.contains(v.name) => {
				val lhs = if(forall_depth > 0) "%s.writeEntry()" else "%s.readEntry()"	
				super.tAssign(e.copy(lv = ve.copy(vec = Var(v.t,lhs.format(v.name)))))
			}
			case ve @ MatExtract(t,v,_,_) if globals.contains(v.name) => {
				val lhs = if(forall_depth > 0) "%s.writeEntry()" else "%s.readEntry()"	
				super.tAssign(e.copy(lv = ve.copy(mat = Var(v.t,lhs.format(v.name)))))
			}
			case _ => super.tAssign(e)
		}
	}
	def tForall(e : Forall) = {
		val Forall(name,lblop,set,block) = e
		val interior =
		"FORALL_SET_TM(%s,%s)".format(name.name,tVar(set)) ::: {
				forall_depth += 1
				val stmts = indent(tForallBodyPrefix(e) ::: tStmt(block))
				forall_depth -= 1
				stmts
			}
		tForallPrefix(e) :::
		{ lblop match {
			case Some(lbl) => {
				"{" ::: indent( "int %s = 0;".format(lbl.name) ::: 
								interior ::: 
								indent("%s++;".format(lbl.name)) :::
								"ENDSET") :::
				"}"
			}
			case None => interior ::: "ENDSET"
		} } ::: tForallSuffix(e)
	}
	def create(build_dir : String) : Unit = {
		val toplevel = normalizeGlobals(top.code)
		
		val linear_systems = { 
			for(ls <- top.linear_systems; line <- tStmt(ls)) 
				yield line
		}.mkString("\n")
		val linear_system_inits = {
			for(ls <- top.linear_systems)
				yield "LINEAR_SYSTEM_INIT(%s);".format(ls.name)
		}.mkString("\n")
		val var_decls = tStmt(toplevel.var_decls).mkString("\n")
		val func_decls = tStmt(toplevel.func_decls).mkString("\n")
		
		val var_defs = indent(tStmt(toplevel.var_defs)).mkString("\n")
		val func_defs = tStmt(toplevel.func_defs).mkString("\n")
		
		val log_str = if(Config.getOrElse("redirect-to-log",false).asInstanceOf[Boolean]) "" else "//"
		val helper_fns = tHelperFunctions().mkString("\n")
		val str = """
//#include "LisztClient.h"
//MyLiszt __liszt;
//FileMesh::FluentFile __fluent;
//kernel_ptr __kernel;
#define LisztScalar %s
#include "%s"
//Helper Functions
%s
//Linear Systems
%s
//Variable declarations
#define __liszt_nested false
%s
#undef __liszt_nested
template<bool __liszt_nested>
struct LisztCode {
//Function declarations
%s
};
//Function definitions
%s
template<bool __liszt_nested>
void liszt_initialize_globals_and_run() {
//Linear System initializations
%s
//Global variable definitions
%s
LisztCode<false>::_liszt_start();
}
int main(int argc, char** argv) {
    predescriptor(argc,argv);
    prepare();
	liszt_initialize_globals_and_run<false>();
    finalize();
}
		""".format(tType(FloatT),runtime_header,helper_fns,linear_systems,var_decls,func_decls,func_defs,linear_system_inits,var_defs,log_str)
		Util.WriteFile("%s/%s/main.cpp".format(build_dir, output_dir)) {
			out => out println str
		}
		val (graph,_,field_info) = AbstractInterpretation.create(top)
		new MPIStencil(graph,field_info,"%s/%s/stencil.cpp".format(build_dir,output_dir))
	} 
}
