package Liszt
import Liszt._
import Liszt.Types._
import Liszt.Trees._
import scala.collection.immutable.{Set => SET}


class InlineFunctions( blk_raw : Block ) {
	object Log extends Liszt.Log("InlineFunctions")
	def remove_nested_blocks(t : Tree) : Tree = t match {
		case Block(stmts,exp) => 
			val new_stmts = stmts.flatMap {
				case Block(stmts,_ : EmptyExp) => (stmts map remove_nested_blocks).asInstanceOf[List[Stmt]]
				case stmt => List(remove_nested_blocks(stmt).asInstanceOf[Stmt])
			}
			Block(new_stmts,exp)
			
		case t => ShapeAny.trees(t,t.children map remove_nested_blocks)
	}
	val blk = remove_nested_blocks(blk_raw).asInstanceOf[Block]
	def can_inline(df : DefFunc) : Boolean = {
		def num_returns(tree : Tree) : Int = tree match {
			case Return(e) => 1
			case x => x.children.foldRight(0){ (a,b) => num_returns(a) + b }
		}
		num_returns(df) match {
			case 0 => true
			case 1 => 
				val stmts = ASTUtil.asStmtList(df.body)
				val ends_return = stmts.last.isInstanceOf[Return]
				if(!ends_return)
					Log.log("...doesn't end in return: %s".format(stmts.last))
				ends_return
			case n => { Log.log("...too many returns"); false }
		}
	}
	//in the future we can put more complicated logic here
	val should_inline : DefFunc => Boolean = Util.Memoize[DefFunc,Boolean] {
		(df : DefFunc) => 
		val ci = can_inline(df)
		Log.log("inlining decision: %s %s".format(df.name,ci))
		ci
	}.apply
	
	def should_inline_call(c : Call) = {
		Log.log("inline call %s?".format(c))
		!c.isExternal && should_inline(blk.function_info(c.func.name).impl)
	}
	val inlined_functions : String => DefFunc = Util.Memoize[String,DefFunc] {
		(name : String) => 
		val df = blk.function_info(name).impl
		
		def generate_inlined_block(c : Call) : List[Stmt] = {
			val call_target = inlined_functions(c.func.name)
			val dup_call_target = ASTUtil.duplicate(call_target).asInstanceOf[DefFunc]
			val formal_parameter_names = dup_call_target.args map (x => x.name)
			val actual_parameter_names = c.args map (x => x.name)
			val formal_to_actual = Map.empty.withDefault( (x:String) => x ) ++ (formal_parameter_names zip actual_parameter_names)
			val call_target_stmts = ASTUtil.asStmtList(dup_call_target.body)
			def replace(t : Tree) : Tree = t match {
				case ShapeAny(tpe,names,trees) => ShapeAny(t,tpe,names map formal_to_actual,trees map replace)
			}
			val call_target_replaced = for(stmt <- call_target_stmts) 
				yield replace(stmt).asInstanceOf[Stmt]
			call_target_replaced
		}
		val inlined = ASTUtil.flatMapBlocks(df) {
			case DefVar(ret_name,c : Call) if should_inline_call(c) => {
				val call_target_replaced = generate_inlined_block(c)
				val return_stmt = call_target_replaced.last.asInstanceOf[Return]
				call_target_replaced.init ::: List(DefVar(ret_name,return_stmt.e))
			}
			case ExpStmt(c @ Call(t,name,args)) if should_inline_call(c) => {
				val call_target_replaced = generate_inlined_block(c)
				if(!call_target_replaced.isEmpty && 
				   call_target_replaced.last.isInstanceOf[Return]) {
					//ignore the return statement (if we return Unit then we don't care),
					//otherwise we want to through it away
					call_target_replaced.init
				} else {
					call_target_replaced
				}
			}
			case x => List(x)
		}
		inlined.asInstanceOf[DefFunc]
	}.apply
	val newstmts = for(stmt <- blk.stmts) yield stmt match {
		case df : DefFunc =>
			Log.log("Generating (possible) inlined %s".format(df.name))
			inlined_functions(df.name)
		case x => x
	}
	val new_block = Block(newstmts,blk.exp)
	//inlining potentially made a bunch of the functions dead, one more pass to eliminated
	//dead functions
	
	def live_stmt(stmt : Stmt) = stmt match {
		case df : DefFunc => !new_block.function_info(df.name).calledBy.isEmpty || df.name == "_liszt_start"
		case x => true
	}
	val new_live_stmts = for(stmt <- new_block.stmts if live_stmt(stmt)) 
		yield stmt
	val result = Block(new_live_stmts,blk.exp)
}