package Liszt

class Module(base_trait : String) {
	object Log extends Liszt.Log("Module")
	Log.log("Loading module %s".format(base_trait))
	
	private def load_module(name : String) = Liszt.Trees.Trait.load(name) match {
		case Some(x) => x
		case None => Log.error("link error: %s module not found.",name)
	}
	
	val (all_traits,linear_solver_op) = {
		var loaded_parents : Set[String] = Set.empty
		var parents : List[Liszt.Trees.Trait] = List.empty
		var linear_solver : Option[String] = None
		def add(instance : Liszt.Trees.TraitApply) {
			if(!loaded_parents.contains(instance.name)) {
				loaded_parents += instance.name
				if(!instance.isLinearSystem) {
					val trt = load_module(instance.name).instanciate(instance.args)
					for(p <- trt.parents)
						add(p)
					parents ::= trt
				} else {
					if(linear_solver == None)
						linear_solver = Some(instance.name)
				}
			}
		}
		add( new Liszt.Trees.TraitApply(base_trait,List(),false) )
		(parents,linear_solver)
	}
	
	def dependencies = for(t <- all_traits; d <- t.module_dependencies) yield d
	
	
	private def packageLinearSystem(solver_name : String, stmts : List[Liszt.Trees.Stmt]) : List[Liszt.Trees.Stmt] = {
		def isIndex( x : Liszt.Trees.Stmt) : Boolean = x.isInstanceOf[Liszt.Trees.Index]
		val (idxs, other) = stmts partition isIndex
		
		
		//check that index functions have the right type
		for(v <- idxs) {
			val Liszt.Trees.Index(_,arg_types,index_type) = v
			for(a <- arg_types) {
				if(!ASTUtil.isMeshElement(a))
					Log.error("All arguments to an index must be mesh elements: %s",v)
			}
			index_type match {
				case Liszt.Types.VectorT(_,_ : Liszt.Types.IndexType) =>
				case _ : Liszt.Types.IndexType =>
				case _ => Log.error("index functions can only return indicies or vectors of indicies: %s",v)
			}
		}
		
		Liszt.Trees.DefLinearSystem(name,solver_name,idxs.asInstanceOf[List[Liszt.Trees.Index]]) :: other
	}
	def implementation : List[Liszt.Trees.Stmt] = {
		import Liszt.Trees._
		
		val stmts = for(t <- all_traits.reverse; stmt <- t.implementation.stmts)
			yield stmt
		
		val flattened_stmts = flattenModule(stmts)
		
		linear_solver_op match {
			case Some(linear_solver) =>
				packageLinearSystem(linear_solver,flattened_stmts)
			case None =>
				flattened_stmts
		}
	}
	
	def name = base_trait
	//flatten the module into a list of statements,
	//we prefix definitions with the module name, making the implementation module-free
	private def flattenModule(stmts : List[Liszt.Trees.Stmt]) : List[Liszt.Trees.Stmt] = {
		import Liszt.Trees._
		val funcs = for(df @ DefFunc(t,nme,args,exp) <- stmts if exp != ExternStmt) 
			yield df -> ASTUtil.moduleName(name,nme)
		val vals = for(vf @ DefVar(nme,args) <- stmts)
			yield vf -> ASTUtil.moduleName(name,nme)
		val index_names = Map.empty ++ {
			for(idx @ Index(nme,_,_) <- stmts)
				yield nme -> ASTUtil.moduleName(name,nme)
		}
		val defToNewName = Map.empty[Tree,String] ++ funcs ++ vals
		val varToDef = ASTUtil.varToDefMap(stmts)
		
		def rename(tree : Tree) : Tree = {
			tree match {
				case DefFunc(t,nme,args,exp) if defToNewName.contains(tree) => 
					DefFunc(t,defToNewName(tree),args,rename(exp).asInstanceOf[Stmt]) copyAttr tree
				case DefVar(name,exp) if defToNewName.contains(tree) =>
					DefVar(defToNewName(tree),rename(exp).asInstanceOf[Exp]) copyAttr tree
				case Var(t,name) => varToDef(tree) match {
					case Some(x) => defToNewName.get(x) match {
						case Some(x) => Var(t,x) copyAttr tree
						case None => tree
					}
					case None => index_names.get(name) match {
						case Some(new_name) => Var(t,new_name) copyAttr tree
						case None => tree
					}
				}
				case _ => ShapeAny.trees(tree,tree.children map rename)
			}
		}
		val result = for(i <- stmts) 
			yield rename(i).asInstanceOf[Stmt]
		result
	}
}