/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */
package Liszt
import scala.collection.immutable.{Set => SSet}
import Liszt.Types._
import Liszt.Trees._

class SMPCodeGen(val top : CompilationUnit, globals : SSet[String] ) extends CLikeCodeGen {
	import ASTUtil._
	import CodeLineImplicits._
	val Log = new Log("CodeGen")
	
	var inTemplate = false
	var inForall = false
	var workgroup_triviality : Map[Int,Boolean] = Map() //workgroup_id to isTrivial
	
	def tFunctionHeader(e : DefFunc, isDecl : Boolean) = {
		val fnt = getCType(e).asInstanceOf[Func]
		val name = e.name
		val args = for(a <- e.args) yield a.name
		val template_header = if(fnt.args.exists(_.isInstanceOf[SpecialType])) {
			val template_args = for( SpecialType(_,typ) <- fnt.args) yield "typename %s".format(typ) 
			inTemplate = true
			"template< %s > ".format(template_args.mkString(","))
		} else {
			""
		}
		val argwt = for( (t,a) <- fnt.args zip args) yield "%s %s".format(tType(t),a)
		val return_type = tType(fnt.range)
		"%s %s %s(%s)".format(template_header,return_type,name,argwt.mkString(","))
	}
	
	def tIndex(e : Index) = Log.ierror("unexpected tIndex")
	
	def isPointerType(t : Type) = false
	override def tEnsureFieldState(e : EnsureFieldState) = if(inForall) List()
	else if(isFieldT(e.field.t))
		"%s.ensureState(%s_TYPE,&System::barrier);".format(tVar(e.field), tReductionOp(e.state))
	else
		"SCALAR_REDUCTION_ENSURE_STATE(%s,%s);".format(e.field.name,tReductionOp(e.state))
	def tFieldInit(e : FieldInit) = Log.ierror("field init unexpected.")
	def tSetInit(e : SetInit) = Log.ierror("set init unexpected.")
	def tIfExp(e : IfExp) = Log.ierror("unexpected ifexp")
	
	override def tDefVar(e : DefVar) = {
		val DefVar(name,init) = e
		def declaration = super.tDefVar(DefVar(e.name,EmptyExp(init.t)) copyAttr e)
		def v = tVar(Var(init.t,name)) //variable refering to this declaration
		val result : List[CodeLine] = init match {
			case fi @ FieldInit(_,init) => {
				val real_name = ASTUtil.nameFromTempHack(e.name)
				val FieldT(mo,value) = e.init.t
				def letter = mo match {
					case Vertex => "V"
					case Edge => "E"
					case Face => "F"
					case Cell => "C"
					case _ => Log.ierror("unexpected type: %s",e.t)
				}
				init match {
					case ConstSource(exp) => "System::loadFieldConst(&%s,\"%s\",%s);".format(real_name,real_name,tVar(exp))
					case StringSource(url) => if(url == "position") {
						"System::loadPositions(&%s,\"%s\");".format(real_name,real_name)
					} else {
						Log.error(fi.report("url not supported"))
					}
				}
			}
			case SetInit(_,set_name) => {
				val real_name = ASTUtil.nameFromTempHack(e.name)
				val Set(v) = e.init.t
				"System::loadBoundary<%sType>(this,%s,&%s,NULL);".format(tType(v),tStringL(StringL(set_name)),real_name)
				//TODO(zach): initialize set wrapper
			}
			case _ => super.tDefVar(e)
		}
		result
	}
	
	
	//next two overrides handle scalar reductions for SMP
	override def tTypeFor(e : Tree) = e match {
		case dv : DefVar if globals.contains(dv.name) => 
			"Global< %s , %s >".format(super.tTypeFor(e),inForall)
		case _ => super.tTypeFor(e)
	}
	override def tVar(e : Var) = if(globals.contains(e.name)) {
		
		val str = if(e.isWrite) {
			if(inForall) "NESTED_WRITE"
			else "UNNESTED_WRITE"
		} else "READ"
		
		"SCALAR_REDUCTION_%s(%s)".format(str,super.tVar(e))
	
	} else super.tVar(e)
	
	override def tDefFunc(e : DefFunc) = {
		
		inForall = false 
		val unnested = 	super.tDefFunc(e)
		inForall = true
		val nested = super.tDefFunc(e.copy(name = e.name + "_nested") copyAttr e)
		inForall = false
		inTemplate = false
		unnested ::: nested
	}
	
	override def tFieldWrite(e : FieldWrite) = {
		"%s.lock(%s);".format(tVar(e.field),tVar(e.arg)) ::: {
		e match {
			case FieldWrite(field,arg,op @ Ops.Reduce(_ : Ops.PickOp),rhs) => tAssign(Assign(Var(rhs.t,"%s[%s]".format(tVar(field),tVar(arg))),op,rhs))
			case _ => super.tFieldWrite(e)
		} } ::: "%s.unlock(%s);".format(tVar(e.field),tVar(e.arg))
	}
	def tForall(e : Forall) = {
		val Forall(name,lblop,set,block) = e
		val suffix = if(inTemplate) "_TM" else ""
		
		val old_forall = inForall
		inForall = true
		val interior =
		"FORALL_SET%s(%s,%s)".format(suffix,name.name,tVar(set)) ::: {
				val stmts = indent(tForallBodyPrefix(e) ::: tStmt(block))
				stmts
			}
		inForall = old_forall
		

		tForallPrefix(e) :::
		{ lblop match {
			case Some(lbl) => {
				"{" ::: indent( "int %s = 0;".format(lbl.name) ::: 
								interior ::: 
								indent("%s++;".format(lbl.name)) :::
								"ENDSET") :::
				"}"
			}
			case None => interior ::: "ENDSET"
		} } ::: tForallSuffix(e)
	}
	
	override def tWorkgroup(e : Workgroup) = if(inForall) super.tWorkgroup(e) else {
		//top-level for-all comprehensions take a different path than nested ones:
		Log.log("workgroup!: %s".format(e))
		val Forall(v,_,set_var,block) = e.stmts.head
		val wid = e.id
		val set_value_type = tType(set_var.t.asInstanceOf[Set].t)
		val args = "%d,%s,%s,%s,%s".format(wid,v.name,set_value_type,tVar(set_var),workgroup_triviality(wid))
		inForall = true
		val result = NewStmt.
		add("WORKGROUP_FORALL_BEGIN(%s)".format(args)).
		begin.
			add(tStmt(block)).
		end.
		add("WORKGROUP_FORALL_END(%s)".format(args))
		inForall = false
		result
	}
	
	def boundaryType(t : Type) = t match {
		case Vertex => "vertex_set"
		case Edge => "edge_set"
		case Face => "face_set"
		case Cell => "cell_set"
	}
	override def tCall(e : Call) = {
		val name = e.func.name + { if (isExternal(e.func.name) || !inForall) "" else "_nested" }
		super.tCall(e.copy(func = e.func.copy(name = name)))
	}
	def globalDeclarations() : StmtResult = {
		//get declarations for all things declared outside of thread specific memory
		//fields, set base representations
		//TODO(zach) global variables not implemented correctly yet
		top.code.stmts flatMap { 
			case DefVar(v,SetInit(t,set_name)) => line { "BoundarySet< %s > %s;".format(boundaryType(t.asInstanceOf[Set].t),v) }
			case dv @ DefVar(v,FieldInit(_,init)) => line { "%s %s;".format(tTypeFor(dv),v) }
			case _ => List()
		}
	}
	override def runtimeTypeFor(ctx : CtxT, exptoctype : ExpT, tree : Tree) = {
		def boundary(t : String) = "BoundarySet< %s >".format(t)
		tree match {
			case SetInit(t,_) => SpecialType(t,t match {
				case Set(t) => boundary(boundaryType(t))
				case _ => Log.ierror("not a set  type %s",t)
			})
			case _ => super.runtimeTypeFor(ctx,exptoctype,tree)
		}
	}
	override def tPrintContext() = "print_context"
	override def tPrintStream() = "(*log_out)"
	def create(build_dir : String) : Unit = {
		top.code = ASTUtil.wrapForallInWorkgroup(top.code).asInstanceOf[Block]
//		Log.log("code is: %s".format(top.code))
		val (tgraph,workgroup_info,_) = AbstractInterpretation.create(top)
		for( (k,v) <- workgroup_info)
			Log.log("%s: %s".format(k.toShortString,v))
		workgroup_triviality = Map.empty.withDefault( (x : Int) => true) ++ {
			for ((w,l) <- workgroup_info) 
				yield (w.id, l.forall( w => w.isTrivial))
		}
		val coloring_generator = new ColoringGenerator(tgraph,workgroup_info,workgroup_triviality,tType)
		
		val toplevel = normalizeGlobals(top.code)
		
		val var_decls = { toplevel.var_decls.stmts flatMap {
			//fields are already declared globally
			case stmt if ASTUtil.isFieldT(getCType(stmt)) => List()
			case DefVar(_,_ : SetInit) => List() //generated globally
			case stmt => tStmt(stmt)
		} }.mkString("\n")
		
		val filtered_var_defs = Util.removeAssignmentsToFieldsAndSets(toplevel.var_defs)
		val var_defs = indent(tStmt(filtered_var_defs)).mkString("\n")
		val func_defs = tStmt(toplevel.func_defs).mkString("\n")
		
		val log_str = if(Config.getOrElse("redirect-to-log",false).asInstanceOf[Boolean]) "" else "//"
		val helper_fns = tHelperFunctions().mkString("\n")
		val global_declarations = globalDeclarations().mkString("\n")
		val str = """
#include "ScalaSMPRuntime.h"
//Helper Functions
%s
//Global Variable declarations
%s

class LisztClient : public LisztThread {
public:
//Local Variable declarations
%s
//Function definitions
%s
void toplevel() {
//variable definitions
%s
}
void main() {
    _liszt_start();
}

}; //LisztClient
LisztThread * LisztClient_create() {
	return new LisztClient;
}
#include "coloring.cpp"
""".format(helper_fns,global_declarations,var_decls,func_defs,var_defs)
		val coloring = coloring_generator.write()
		Util.WriteFile("%s/smp/main.cpp".format(build_dir)) {
			out => out println str
		}
		Util.WriteFile("%s/smp/coloring.cpp".format(build_dir)) {
			out => out println coloring
		}

	} 

}
