/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */
package Liszt
import scala.collection.immutable.{Set => SSet}
import scala.collection.mutable.{Set => MSet}
import scala.collection.mutable.{Stack => SStack}
import scala.collection.mutable.{Map => SMap}
import Liszt.Types._
import Liszt.Trees._


class SyrahCodeGen(val top : CompilationUnit, globals : SSet[String] ) extends CLikeCodeGen {
  import ASTUtil._
  import CodeLineImplicits._
  val Log = new Log("CodeGen")
  val MaskStack = new SStack[String];
  var inTemplate = false
  var inForall = false
  var singleReturnStmt = false
  var BoundarySets : MSet[String] = MSet.empty;

  var LocalSequentialSets : MSet[String] = MSet.empty;

  def NumReturnStmts(body : Tree) : Int = {
    //Log.log("tree %s".format(body.toString()));
    body match {
      case Return(e) => 1;
      case If(predE, thenE, elseE) => NumReturnStmts(thenE) + NumReturnStmts(elseE);
      case IfExp(t, predE, thenE, elseE) => NumReturnStmts(thenE) + NumReturnStmts(elseE);
      case While(pred,block) => NumReturnStmts(block);
      case Forall(name,_,set,block) => NumReturnStmts(block);
      case Block(stmts, exp) => NumReturnStmts(exp) + (for (e <- stmts) yield NumReturnStmts(e)).foldLeft(0)(_+_)
      case _ => 0;
    }
  }

  def isMainFunc(name : String) = (name.endsWith("___main"));

  def tFunctionHeader(e : DefFunc, isDecl : Boolean) = {
    "ERROR";
  }

  def SyrahFunctionHeader(e : DefFunc, isDecl : Boolean, addMask : Boolean) = {
    val fnt = getCType(e).asInstanceOf[Func]
    val name = e.name
    val args = for(a <- e.args) yield a.name
    val template_header = if(fnt.args.exists(_.isInstanceOf[SpecialType])) {
      val template_args = for( SpecialType(_,typ) <- fnt.args) yield "typename %s".format(typ)
      inTemplate = true
      "template< %s > ".format(template_args.mkString(","))
    } else {
      ""
    }

    //val inline_flag = { if (isMainFunc(name)) { "" } else { "SYRAH_FORCEINLINE" } }

    val inline_flag = { if (isMainFunc(name)) { "" } else { "inline" } }

    val argwt = for( (t,a) <- fnt.args zip args) yield "%s %s".format(tArgType(t),a)
    val extra = if (addMask) { "%sconst FixedVectorMask<VecWidth>& __inputMask".format(if (e.args.size > 0) { ", " } else { "" }) } else { "" }
    val return_type = tType(fnt.range)
    "%s %s %s %s(%s%s)".format(inline_flag, template_header,return_type,name,argwt.mkString(","), extra)
  }

  def tArgType(t : Type) = {
    "const %s&".format(tType(t));
  }

  override def tType(t : Type) = {
    //println("tType for '%s'".format(t.toString()));
    t match {
      case x : SimdT => x.t match {
        case FloatT => "FixedVector<float, VecWidth>"
        case DoubleT => "FixedVector<double, VecWidth>"
        case IntT => "FixedVector<int, VecWidth>"
        case BoolT => "FixedVectorMask<VecWidth>"
        case VectorT(MetaIntT(n),t) => "SimdVec<%d, %s, VecWidth>".format(n, tType(t));
        case MatrixT(MetaIntT(n),MetaIntT(m), t) => "SimdMatrix<%d, %d, %s, VecWidth>".format(n, m, tType(t));
        case Vertex => "SIMDObject<VecWidth, %s>".format(tType(Vertex));
        case Face => "SIMDObject<VecWidth, %s>".format(tType(Face));
        case Edge => "SIMDObject<VecWidth, %s>".format(tType(Edge));
        case Cell => "SIMDObject<VecWidth, %s>".format(tType(Cell));
        case y @ Set(_) => "SIMDObject<VecWidth, %s>".format(tType(y));
        case _ => super.tType(t);
      }
      case _ => super.tType(t);
    }
  }

  override def tBinary(e: Binary) = e match {
    case Binary(t, lhs, op : Ops.PickOp, rhs) => t match {
      case SimdT(_) => "%s(%s, %s)".format(tPickOp(op), tVar(lhs), tVar(rhs));
      case _ => super.tBinary(e);
    }
    case _ => super.tBinary(e);
  }

  // TODO(boulos): Deal with the special type junk in CTypes for
  // _return_type. .... Actually, currently dodge it by overriding
  // tDefVar for TopoCall ;).

  override def tAssign(e : Assign) = {
    val vec_assign = isSimdT(e.r.t) || isSimdT(e.lv.t);
    val vec_reduce = !isSimdT(e.lv.t) && isSimdT(e.r.t);
    if (vec_assign) {
      if (vec_reduce) {
        // The RHS is a SimdT and the lhs is a non-SimdT.
        val default_val = e.op match {
          case Ops.Reduce(Ops.Plus) => e.r.t match {
            case SimdT(_) => "%s::Zero()".format(tType(e.r.t));
            case _ => "0";
          }
          case Ops.Reduce(Ops.Minus) => e.r.t match {
            case SimdT(_) => "%s::Zero()".format(tType(e.r.t));
            case _ => "0";
          }
          case Ops.Reduce(Ops.Times) => "1";
          case Ops.Reduce(Ops.Max) => e.r.t match {
            case SimdT(FloatT) => "-std::numeric_limits<float>::infinity()";
            case SimdT(VectorT(_, FloatT)) => "-std::numeric_limits<float>::infinity()";
            case SimdT(MatrixT(_, _, FloatT)) => "-std::numeric_limits<float>::infinity()";
            case SimdT(DoubleT) => "-std::numeric_limits<double>::infinity()";
            case SimdT(VectorT(_, DoubleT)) => "-std::numeric_limits<double>::infinity()";
            case SimdT(MatrixT(_, _, DoubleT)) => "-std::numeric_limits<double>::infinity()";
            case SimdT(IntT) => "std::numeric_limits<int>::min()";
            case _ => Log.ierror("vec_reduce with max and type = '%s'".format(tType(e.r.t))); "ERROR";
          }
          case Ops.Reduce(Ops.Min) => e.r.t match {
            case SimdT(FloatT) => "std::numeric_limits<float>::infinity()";
            case SimdT(VectorT(_, FloatT)) => "std::numeric_limits<float>::infinity()";
            case SimdT(MatrixT(_, _, FloatT)) => "std::numeric_limits<float>::infinity()";
            case SimdT(DoubleT) => "std::numeric_limits<double>::infinity()";
            case SimdT(VectorT(_, DoubleT)) => "std::numeric_limits<double>::infinity()";
            case SimdT(MatrixT(_, _, DoubleT)) => "std::numeric_limits<double>::infinity()";
            case SimdT(IntT) => "std::numeric_limits<int>::max()";
            case _ => Log.ierror("vec_reduce with min and type = '%s'".format(tType(e.r.t))); "ERROR";
          }
          case _ => Log.ierror("some sort of non-reduce... '%s'".format(e.toString)); "ERROR";
        }

        val simd_ctor = "%s".format(tType(e.r.t));

        val temp_name = Names.create("reduce_temp");
        val rhs_alloc = if (MaskStack.size > 0) {
          "%s %s = select(%s, %s(%s), %s);".format(tType(e.r.t), temp_name, tVar(e.r), simd_ctor, default_val, MaskStack.top);
        } else { "" };

        val rhs_vec = if (MaskStack.size > 0) { temp_name; } else { e.r }

        val fold_type = e.op match {
          case Ops.Reduce(Ops.Plus) => "foldSum";
          case Ops.Reduce(Ops.Minus) => "foldSum";
          case Ops.Reduce(Ops.Times) => "foldProd";
          case Ops.Reduce(Ops.Max) => "foldMax";
          case Ops.Reduce(Ops.Min) => "foldMin";
          case _ => Log.ierror("unsupported reduction '%s'".format(e)); "ERROR";
        }

        val folded_name = Names.create("folded_result");
        val base_type = e.r.t match {
          case SimdT(x) => x;
          case _ => e.r.t;
        }

        // I don't want to add foldSub
        val adjustment = e.op match {
          case Ops.Reduce(Ops.Minus) => "-";
          case _ => "";
        }
        val rhs_folded = "%s %s = %s(%s.%s());".format(tType(base_type), folded_name, adjustment, rhs_vec, fold_type);
        val final_assign = Assign(e.lv, e.op, Var(base_type, folded_name));
        rhs_alloc ::: rhs_folded ::: super.tAssign(final_assign);
      } else {
        val lhs_val = e.lv match {
          case v : Var => tVar(v);
          case ve: VecExtract => tVecExtract(ve);
          case me: MatExtract => tMatExtract(me);
          case _ => Log.ierror("Unsupported lvalue: %s", e); "ERROR";
        }
        if (MaskStack.size > 0) {
          e.op match {
            case Ops.Assign => "%s.merge(%s, %s);".format(lhs_val, tVar(e.r), MaskStack.top);
            case Ops.Reduce(Ops.And) => "%s.merge(%s & %s, %s);".format(lhs_val, lhs_val, tVar(e.r), MaskStack.top);
            case Ops.Reduce(Ops.Or) => "%s.merge(%s | %s, %s);".format(lhs_val, lhs_val, tVar(e.r), MaskStack.top);
            case Ops.Reduce(op: Ops.PickOp) => "%s.merge(%s(%s,%s), %s);".format(lhs_val, tPickOp(op), lhs_val, tVar(e.r), MaskStack.top);
            case Ops.Reduce(op: Ops.BinOp) => "%s.merge(%s %s %s, %s);".format(lhs_val, lhs_val, tBinOp(op), tVar(e.r), MaskStack.top);
            case _ => Log.ierror("unsupported assignment!");
          }
        } else {
          e.op match {
            case Ops.Reduce(op: Ops.PickOp) => "%s = %s(%s, %s);".format(lhs_val, tPickOp(op), lhs_val, tVar(e.r));
            case _ => super.tAssign(e);
          }
        }
      }
    } else {
      super.tAssign(e);
    }
  }

  override def tForall(e : Forall) = {
    // handle a forall loop. Different if we're inside already or not.
    val Forall(name,label,set,block) = e;
    if (isSimdT(set.t) /*inForall*/) {
      // The set is actually a vector of different sets.

      val loop_mask = Names.create("loop_mask");
      val input_mask = if (MaskStack.size > 0) MaskStack.top else "FixedVectorMask<VecWidth>::True()";
      val result = "FORALL_SET_SYRAH_VECTORIZED(%s, %s, VecWidth, %s, %s)".format(name.name, tVar(set), loop_mask, input_mask) :::
      { MaskStack.push(loop_mask); val code_block = indent(tStmt(block)); MaskStack.pop(); code_block } ::: "ENDSET_SYRAH_NESTED(%s, VecWidth, %s)".format(name.name, loop_mask);
      result;
    } else {
      // Get ourselves a vector full of the elements.
      inForall = true;
      val forallLoopVar = name.name;
      val loop_mask = Names.create("loop_mask");

      // If the set is a global variable, I think it's inherently a boundary set definition.
      //println("Checking to see if set %s is a BoundarySet".format(set.name));
      val is_boundary_set = BoundarySets.contains(set.name);
      val iter_type = if (is_boundary_set) "_BOUNDARY_ITER" else "";

      val result = "FORALL_SET_SYRAH_BEGIN(%s,%s, VecWidth)".format(forallLoopVar, tVar(set)) :::
      indent("FORALL_SET_SYRAH%s_NOMASK(%s,VecWidth)".format(iter_type, forallLoopVar, tVar(set)) ::: indent(tStmt(block)) ::: "ENDSET" ::: "FORALL_SET_SYRAH%s_CLEANUP(%s,VecWidth,%s)".format(iter_type, forallLoopVar, loop_mask) ::: { MaskStack.push(loop_mask); val code_block = indent(tStmt(block)); MaskStack.pop(); code_block } ::: "ENDSET") ::: "FORALL_SET_SYRAH_END";
      inForall = false;
      result;
    }
  }

  override def tIf(e: If) = {
    val If(blk,thenE,elseE) = e
    val vec_cond = isSimdT(blk.exp.t);

    if (!vec_cond) {
      super.tIf(e);
    } else {
      val pred_mask = Names.create("if_pred");
      // Compute the mask (itself possibly masked via the
      // current mask).
      val stmts = blk.stmts.flatMap(tStmt);
      val cond = tStmt(DefVar(pred_mask, blk.exp) copyAttr blk.exp);

      val then_mask = Names.create("then_mask");
      val else_mask = Names.create("else_mask");

      var then_mask_real = "";
      if (MaskStack.size > 0) {
        then_mask_real = "FixedVectorMask<VecWidth> %s = (%s & %s);".format(then_mask, MaskStack.top, pred_mask);
        MaskStack.push(then_mask);
      } else {
        MaskStack.push(pred_mask);
      }

      val cond_var = MaskStack.top;

      // Okay, before we proceed. See if the code is actually a conditional assignment to a single variable
      val then_is_assign = if (isSingleStatment(thenE)) {
        thenE.stmts(0) match {
          case Assign(lhs, op, rhs) => true;
          case _ => false;
        }
      } else { false };
      val else_is_assign = if (isSingleStatment(elseE)) {
        elseE.stmts(0) match {
          case Assign(lhs, op, rhs) => true;
          case _ => false;
        }
      } else { false; }

      //println("then_is_assign = %s, else_is_assign = %s".format(then_is_assign, else_is_assign));

      val use_select = {
        if (then_is_assign && else_is_assign) {
          // Alright, figure out if they assign to the same lhs
          val then_assign : Assign = thenE.stmts(0).asInstanceOf[Assign];
          val else_assign : Assign = elseE.stmts(0).asInstanceOf[Assign];
          val same_lhs = {
            val then_is_var = then_assign.lv match {
              case Var(t, name) => true;
              case _ => false;
            };
            val else_is_var = else_assign.lv match {
              case Var(t, name) => true;
              case _ => false;
            };
            if (then_is_var && else_is_var) {
              val then_lhs = then_assign.lv.asInstanceOf[Var];
              val else_lhs = else_assign.lv.asInstanceOf[Var];
              then_lhs.name == else_lhs.name;
            } else { false }
          }

          val same_op  = then_assign.op == else_assign.op;
          val simple_assign = (then_assign.op == Liszt.Ops.Assign);
          //println("same_lhs = %s, same_op = %s, simple_assign = %s".format(same_lhs, same_op, simple_assign));
          same_lhs && same_op && simple_assign;
        } else {
          false;
        }
      };

      if (use_select) {
        MaskStack.pop();

        val then_assign : Assign = thenE.stmts(0).asInstanceOf[Assign];
        val else_assign : Assign = elseE.stmts(0).asInstanceOf[Assign];
        val then_val = tVar(then_assign.r);
        val else_val = tVar(else_assign.r);
        val lhs = then_assign.lv.asInstanceOf[Var];

        stmts ::: cond ::: then_mask_real ::: "%s = select(%s, %s, %s);".format(tVar(lhs), then_val, else_val, cond_var);
      } else {
        // General case
        val then_code = indent(tStmt(thenE));
        MaskStack.pop();
        var else_mask_real = "";
        if (MaskStack.size > 0) {
          else_mask_real = "FixedVectorMask<VecWidth> %s = (andNot(%s,%s));".format(else_mask, MaskStack.top, pred_mask);
        } else {
          else_mask_real = "FixedVectorMask<VecWidth> %s = (!%s);".format(else_mask, pred_mask);
        }
        MaskStack.push(else_mask);
        val else_code = indent(tStmt(elseE));
        MaskStack.pop();
        val if_check    = "if (Any(%s)) {".format(cond_var);
        val else_branch = "}\nif (Any(%s)) {".format(else_mask);
        val end_if = "}";
        // TODO(boulos): Determine when to add an if-any check.
        /*
         val if_check = ";";
         val else_branch = ";";
         val end_if = ";";
         */
        stmts ::: cond ::: then_mask_real ::: else_mask_real ::: if_check ::: then_code ::: else_branch ::: else_code ::: end_if;
      }
    }
  }

  override def tReturn(e: Return) = {
    if (inForall) {
      // We're a vectorized function, update the done mask and branch on it
      val need_mask = (MaskStack.size > 1 ||
                       (MaskStack.size == 1 && !(MaskStack.top.startsWith("input_mask_copy") && singleReturnStmt)));

      val updateResult = e.e match {
        case EmptyExp(_) => ";"
        // TODO(boulos): This nested mask function thing kinda
        // sucks. We are okay with returnValue ignoring the
        // "input mask" probably and we don't want it to be
        // anded everywhere necessarily. It's more like "if
        // there's a non-function visible load/store, respect
        // the input mask".. But we also need to deal with "the
        // function had some control flow, updated the
        // __returnValue previously and now we've popped up with
        // a possibly different mask"
        case exp => { if (need_mask) { "__returnValue.merge(%s, %s);".format(tExp(exp), MaskStack.top) } else { "__returnValue = %s;".format(tExp(exp)); }}
      }
      val returnResult = e.e match {
        case EmptyExp(_) => "return;"
        case exp => "return __returnValue;";
      }
      if (need_mask) {
        // NOTE(boulos): Input mask is passed in and then copied by
        // value to allow update here...

        // We're in control flow within the function

        // We'll still be on if we're the opposite of people who are exiting
        val stillOn = Names.create("stillOn");
        val stillOnCode  = "FixedVectorMask<VecWidth> %s = !%s;".format(stillOn, MaskStack.top);
        var updateAllMasks : StmtResult = List();
        //val updateAllMasks = MaskStack.flatMap { s => "%s &= %s;".format(s, stillOn) }
        for(mask <- MaskStack.elems) {
          updateAllMasks = updateAllMasks ::: "%s &= %s;".format(mask, stillOn);
        }
        // Check to see if we really do need to jump out (bottom mask is all off)
        val jumpOut = "if (None(%s)) %s".format(MaskStack.last, returnResult);

        stillOnCode ::: updateResult ::: updateAllMasks ::: jumpOut;
      } else {
        // Not in control flow within the function (passed in mask is given, but doesn't matter)
        updateResult ::: returnResult;
      }
    } else {
      super.tReturn(e);
    }
  }

  override def tWhile(e: While) = {
    // If the condition (and all escapes, such as break/return
    // which don't exist in Liszt) depends on the induction
    // variable for the FORALL loop, we need a vectorized
    // while. Otherwise, a scalar while loop suffices.
    val is_vec = isSimdT(e.pred.exp.t);

    if (is_vec) {
      val While(pred,block) = e
      val cond = Names.create("while_pred");
      val combined_mask = if (MaskStack.size > 0) { Names.create("new_mask") } else { cond };
      val combine_code = if (MaskStack.size > 0) { "FixedVectorMask<VecWidth> %s = (%s & %s);".format(combined_mask, cond, MaskStack.top); } else { "" };

      val bexp = pred.exp
      val exps = pred.stmts

      // NOTE(boulos): I'm doing this with a do { stmts; if
      // (None(cond)) break; stmts; } while(true); instead to
      // avoid confusing the compiler with goto. It's
      // semantically equivalent but no explicit goto.
      "do {" :::
      (indent {
	exps.flatMap(tStmt) :::
        tStmt(DefVar(cond, bexp) copyAttr bexp) :::
        combine_code :::
        "if (None(%s)) break;".format(combined_mask) :::
	{ MaskStack.push(combined_mask); val block_code = tStmt(block); MaskStack.pop(); block_code; }
      }):::
      "} while(true);"
    } else {
      super.tWhile(e);
    }
  }

  def tIndex(e : Index) = Log.ierror("unexpected tIndex")

  def isPointerType(t : Type) = false
  override def tEnsureFieldState(e : EnsureFieldState) = "";
  //if(inForall || !isFieldT(e.field.t)) List() else "%s.ensureState(%s_TYPE,&System::barrier);".format(tVar(e.field), tReductionOp(e.state))


  def tFieldInit(e : FieldInit) = Log.ierror("field init unexpected.")
  def tSetInit(e : SetInit) = Log.ierror("set init unexpected.")
  def tIfExp(e : IfExp) = Log.ierror("unexpected ifexp")

  override def tDefVar(e : DefVar) = {
    val DefVar(name,init) = e
    //def declaration = super.tDefVar(DefVar(e.name,EmptyExp) copyAttr e)


    val result : List[CodeLine] = init match {
      case fi @ FieldInit(_,init) => {
	val real_name = ASTUtil.nameFromTempHack(e.name)
	val FieldT(mo,value) = e.init.t
	def letter = mo match {
	  case Vertex => "V"
	  case Edge => "E"
	  case Face => "F"
	  case Cell => "C"
	  case _ => Log.ierror("unexpected type: %s",e.t)
	}
	init match {
	  case ConstSource(exp) => "System::loadFieldConst(&%s,\"%s\",%s);".format(real_name,real_name,tVar(exp))
	  case StringSource(url) => if(url == "position") {
	    "System::loadPositions(&%s,\"%s\");".format(real_name,real_name)
	  } else {
	    Log.error(fi.report("url not supported"))
	  }
	}
      }
      case TopoCall(t, op, args) => {
        // If result is a SimdT we should use a for-loop.
        t match {
          case SimdT(base_type) => {
            val decl = if (isSetT(base_type)) {
              "SIMDObject<VecWidth, %s> %s;".format(specialReturnType(op.asFunction), name);
            } else if (isMeshT(base_type)) {
              "SIMDObject<VecWidth, %s> %s;".format(tType(base_type), name);
            } else {
              "FixedVector<%s, VecWidth> %s;".format(tType(base_type), name);
            }

            val indexed_args = for (a <- args) yield "%s%s".format(tVar(a), if (isSimdT(a.t)) "[i]" else "");

            val for_init = "for (int i = 0; i < VecWidth; i++) { %s[i] = %s(%s); }".format(name, op.asFunction, indexed_args.mkString(", "));
            decl ::: for_init;
          }
          case _ => super.tDefVar(e);
        }
      }
      case SetInit(_,set_name) => {
	val real_name = ASTUtil.nameFromTempHack(e.name)
        //def v = tVar(Var(init.t,name)) //variable refering to this declaration

	val Set(v) = e.init.t
	"System::loadBoundary<%sType>(this,%s,&%s_global,&%s);".format(tType(v),tStringL(StringL(set_name)),real_name,real_name)
	//TODO(zach): initialize set wrapper
      }
      // FIXME(boulos): What a pain in the ass.
      case Var(t, "Constants___float3_zero") => {
        if (inForall) {
          val simd_type = SimdT(VectorT(MetaIntT(3), FloatT));
          "%s %s = %s;".format(tType(simd_type), name, tExp(init));
        } else {
          super.tDefVar(e);
        }
      }
      case Var(t, "Constants___float3x3_zero") => {
        if (inForall) {
          val simd_type = SimdT(MatrixT(MetaIntT(3), MetaIntT(3), FloatT));
          "%s %s = %s;".format(tType(simd_type), name, tExp(init));
        } else {
          super.tDefVar(e);
        }
      }
      case _ => super.tDefVar(e)
    }
    result
  }

  def tNestedFunc(e : DefFunc, withMask : Boolean) = {
    val DefFunc(ftype,name,args,body) = e;
    val fnt = getCType(e).asInstanceOf[Func]
    val return_type = tType(fnt.range);
    val declare_return = if (return_type == "void") { ";" } else { "%s __returnValue;".format(return_type); }
    val end_return = if (return_type == "void") { "return;" } else { "return __returnValue;".format(return_type); }

    val translated = if (withMask) {
      val local_mask = Names.create("input_mask_copy");
      MaskStack.push(local_mask);
      val with_mask = indent(tStmt(body));
      MaskStack.pop();
      "FixedVectorMask<VecWidth> %s = __inputMask;".format(local_mask) ::: with_mask;
    } else {
      indent(tStmt(body));
    }

    "%s {".format(SyrahFunctionHeader(e, false, withMask)) ::: indent(declare_return) ::: translated ::: indent(end_return) ::: "}";
  }

  def ComputeSequentialSets(e : DefFunc) : Unit = {
    LocalSequentialSets = MSet.empty;
    for (t <- e.body) {
      t match {
        case DefVar(name, TopoCall(_, func, topo_args)) => {
          //println("Considering initialization of %s for sequential-ness".format(name));
          // If it's a topocall with mesh as the args, it's sequential.
          val just_one_arg = topo_args.length == 1;
          if (just_one_arg) {
            val arg = topo_args(0);
            if (arg.t == Mesh) {
              //println("%s IS sequential!".format(name));
              LocalSequentialSets.add(name);
            }
          }
        }
        case Forall(defvar, _, set, _) => {
          //println("Considering initialization of forall var %s for sequential-ness".format(defvar.name));
          if (LocalSequentialSets.contains(tVar(set))) {
            //println("%s IS sequential via %s!\n".format(defvar.name, tVar(set)));
            LocalSequentialSets.add(defvar.name);
          }
        }
        case _ => ;
      }
    }
  }

  override def tDefFunc(e : DefFunc) = e match {
    case DefFunc(_,_,_,ExternStmt) => "%s;".format(SyrahFunctionHeader(e, true, false));
    case _ => {
      //println("number of return stmts in function '%s' = %d".format(e.name, NumReturnStmts(e.body)));
      singleReturnStmt = (NumReturnStmts(e.body) == 1);
      ComputeSequentialSets(e);

      // Assume we had the nested functions generated before hand.
      if (e.name.endsWith("_nested")) {
        inForall = true;
        val result_with_mask = tNestedFunc(e, true);
        val result_no_mask = tNestedFunc(e, false);
        inForall = false;
        result_with_mask ::: result_no_mask;
      } else {
        inForall = false;
        val result = super.tDefFunc(e);
        inForall = false;
        "%s {".format(SyrahFunctionHeader(e, false, false)) :::
          indent(tStmt(e.body)) :::
        "}"
      }
    }
  }


  override def tLiteral(e : Literal) = e match {
    case SimdL(l : Literal) => super.tLiteral(l);
    case _ => super.tLiteral(e);
  }

/*
  override def tSimdL(l : SimdL) = l match {
    case SimdL(v : VectorL) => {
      // So we have a vector constructor, not necessarily a "literal".

    }
    case SimdL(scalar : _) => super.tLiteral(scalar);
  }
*/

  override def tFieldRead(e: FieldRead) = {
    // NOTE(boulos): The field can't be vectorized since you can't
    // take reference to fields...
    if (isSimdT(e.arg.t) || inForall) {
      // TODO(boulos): the inForall is crap.
      val field_type = (e.field.t.asInstanceOf[FieldT]);
      val field_result = field_type.range;

      val sequential_read = LocalSequentialSets.contains(tVar(e.arg));
      val sequential_suffix = if (sequential_read) "Sequential" else "";

      if (isVec(field_result)) {
        //val vec_type = field_result.asInstanceOf[VectorT];
        //val dim = (vec_type.n.asInstanceOf[MetaIntT]).a;
        "%s.simdVecRead%s(%s)".format(tVar(e.field), sequential_suffix, tVar(e.arg));
      } else if (isMat(field_result)) {
        //val mat_type = field_result.asInstanceOf[MatrixT];
        //val n = (mat_type.n.asInstanceOf[MetaIntT]).a;
        //val m = (mat_type.m.asInstanceOf[MetaIntT]).a;
        "%s.simdMatRead%s(%s)".format(tVar(e.field), sequential_suffix, tVar(e.arg));
      } else {
        "%s.simdRead%s(%s)".format(tVar(e.field), sequential_suffix, tVar(e.arg));
      }
    } else {
      super.tFieldRead(e);
    }
  }

  override def tFieldWrite(e : FieldWrite) = {
    if (inForall) {
      // We're writing to multiple entries at once. Because of
      // possible aliasing (e.g. write to head of one edge and
      // tail of another which conflict) we scalarize this loop
      // for now.

      def rhs_value(rhs : Var) = {
        if (isSimdT(rhs.t)) {
          if (isSimdVec(rhs.t)) {
            // SimdT(Vector(N))
            "%s.getVec(__i)".format(tVar(rhs));
          } else if (isSimdMat(rhs.t)) {
            "%s.getMat(__i)".format(tVar(rhs));
          } else {
            // Scalar SimdT
            "%s[__i]".format(tVar(rhs));
          }
        } else {
          "%s".format(tVar(rhs));
        }
      }

      // TODO(boulos): The vec writes are slower for some
      // reason... Disabling for now.
      val simple_write = e.op match {
        case Ops.Assign => true;
        case Ops.Reduce(Ops.Plus) => if (isSimdScalar(e.rhs.t)) true else false;
        case Ops.Reduce(Ops.Minus) => if (isSimdScalar(e.rhs.t)) true else false;
        case _ => false;
      };

      if (simple_write) {
        val sequential_write = LocalSequentialSets.contains(tVar(e.arg));
        val sequential_suffix = if (sequential_write) "Sequential" else "";

        val mask_string = if (MaskStack.size > 0) ", %s".format(MaskStack.top) else "";

        val write_suffix = e.op match {
          case Ops.Assign => "%s".format(sequential_suffix);
          case Ops.Reduce(Ops.Plus) => "PlusEq"; // No seq yet
          case Ops.Reduce(Ops.Minus) => "MinusEq"; // No seq yet
          case _ => "ERROR";
        };

        "%s.simdWrite%s(%s, %s%s);".format(tVar(e.field), write_suffix, tVar(e.arg), tVar(e.rhs), mask_string);
      } else {
        val check_mask = if (MaskStack.size > 0) "if (!%s.get(__i)) continue;".format(MaskStack.top) else ";";
        "for (int __i = 0; __i < VecWidth; ++__i) {" :::
        indent( { check_mask :::
                 "%s.lock(%s[__i]);".format(tVar(e.field), tVar(e.arg)) :::
                 { e match {
                   // NOTE(boulos): I think the reduce op is fine
                   // actually. Since I'm scalarized anyway.
                   case FieldWrite(field, arg,  Ops.Reduce(op : Ops.PickOp), rhs) => {
                     "%s[%s[__i]].%sWith(%s);".format(tVar(field), tVar(arg), super.tPickOp(op), rhs_value(rhs));
                   }
                   case FieldWrite(field, arg, op : Ops.AssignOp, rhs) => {
                     "%s[%s[__i]] %s %s;".format(tVar(field), tVar(arg), super.tAssignOp(op), rhs_value(rhs));
                   }
                 }
                } ::: "%s.unlock(%s[__i]);".format(tVar(e.field), tVar(e.arg)) } ) ::: "}"
      }
    } else {
      "%s.lock(%s);".format(tVar(e.field),tVar(e.arg)) ::: {
	e match {
	  case FieldWrite(field,arg,op @ Ops.Reduce(_ : Ops.PickOp),rhs) => tAssign(Assign(Var(rhs.t,"%s[%s]".format(tVar(field),tVar(arg))),op,rhs))
	  case _ => super.tFieldWrite(e)
	} } ::: "%s.unlock(%s);".format(tVar(e.field),tVar(e.arg))
    }
  }
  def boundaryType(t : Type) = t match {
    case Vertex => "vertex_set"
    case Edge => "edge_set"
    case Face => "face_set"
    case Cell => "cell_set"
  }
  override def tCall(e : Call) = {
    var arg_string = e.args.map(tVar).mkString(", ");
    if (inForall && !isExternal(e.func.name)) {
      if (MaskStack.size > 0) {
        // Calling with a mask
        arg_string += ", %s".format(MaskStack.top);
      } else {
        // Call with true (alternatively we can also make another variant that takes no mask...)

        // Calling with a "nomask" version is about another 8% faster with Joe.

        //arg_string += ", FixedVectorMask<VecWidth>::True()";
      }
    }
    "%s(%s)".format(tVar(e.func), arg_string);
  }

  def markBoundarySets() = {
    for (stmt <- top.code.stmts) {
      stmt match {
        case DefVar(v, SetInit(t, set_name)) => BoundarySets.add(v);
        case _ => ;
      }
    }
  }

  def globalDeclarations() : StmtResult = {
    //get declarations for all things declared outside of thread specific memory
    //fields, set base representations
    //TODO(zach) global variables not implemented correctly yet
    top.code.stmts flatMap {
      case DefVar(v,SetInit(t,set_name)) => line { "BoundarySet< %s > %s_global;".format(boundaryType(t.asInstanceOf[Set].t),v) }
      case dv @ DefVar(v,FieldInit(_,init)) => line { "%s %s;".format(tTypeFor(dv),v) }
      case _ => List()
    }
  }
  override def runtimeTypeFor(ctx : CtxT, exptoctype : ExpT, tree : Tree) = {
    def boundary(t : String) = "BoundarySetWrapper< BoundarySet< %s > >".format(t)
    tree match {
      case SetInit(t,_) => SpecialType(t,t match {
	case Set(t) => boundary(boundaryType(t))
	case _ => Log.ierror("not a set  type %s",t)
      })
      case DefVar(name, init) => init match {
        case x @ SimdL(l) => x.t;
        case _ => /*println("skipping runtime type for tree '%s'".format(tree.toString));*/ super.runtimeTypeFor(ctx, exptoctype, tree);
      }
      case x @ SimdL(l) => x.t;
      case _ => /*println("skipping runtime type for tree '%s'".format(tree.toString));*/super.runtimeTypeFor(ctx,exptoctype,tree)
    }
  }
  override def tPrintContext() = "print_context"
  override def tPrintStream() = "(*log_out)"
  def create(build_dir : String) : Unit = {
    
    val toplevel = normalizeGlobals(top.code)
    markBoundarySets();
    
    val var_decls = { toplevel.var_decls.stmts flatMap {
      //fields are already declared globally
      case stmt if ASTUtil.isFieldT(getCType(stmt)) => List()
      case DefVar(_,_ : SetInit) => List() //generated globally
      case stmt => tStmt(stmt)
    } }.mkString("\n")

    val filtered_var_defs = Util.removeAssignmentsToFieldsAndSets(toplevel.var_defs)
    val var_defs = indent(tStmt(filtered_var_defs)).mkString("\n")
    val func_defs = tStmt(toplevel.func_defs).mkString("\n")

    val log_str = if(Config.getOrElse("redirect-to-log",false).asInstanceOf[Boolean]) "" else "//"
    val helper_fns = tHelperFunctions().mkString("\n")
    val global_declarations = globalDeclarations().mkString("\n")
    val str = """
    #include "/Users/boulos/projects/syrah/src/include/syrah/FixedVector.h"
    using namespace syrah;
    static const int VecWidth = 16;
    #include "ScalaSMPRuntime.h"
    // include after ScalaSMPRuntime to get Vector.h first.
    #include "SyrahVector.h"
    #include "SyrahMatrix.h"
    #include "SyrahObjectWrapper.h"
    #define PRINT_VEC(vec) (*log_out) << "vec = " << vec << "\n";
    template<size_t N, typename ElemType>
    bool isSequential(const SIMDObject<N, ElemType>& vec) {
      size_t first_id = vec[0].ID();
      for (size_t i = 0; i < N; i++) {
        if (first_id + i != vec[i].ID()) return false;
      }
      return true;
    }
    #define PRINT_NOT_SEQUENTIAL(vec) if (!isSequential(vec)) { PRINT_VEC(vec); }

    //Helper Functions
    %s
    //Global Variable declarations
    %s

    class LisztClient : public LisztThread {
      public:
    //Local Variable declarations
    %s
    //Function definitions
    %s
    void toplevel() {
      //variable definitions
      %s
    }
    void main() {
      _liszt_start();
    }

    }; //LisztClient
    LisztThread * LisztClient_create() {
      return new LisztClient;
    }

    void createColoringMetaData(const ColoringMesh & cpu_mesh, ::Synchronization::InterferenceBuilder * igraph) {}
    """.format(helper_fns,global_declarations,var_decls,func_defs,var_defs)

    Util.WriteFile("%s/syrah/main.cpp".format(build_dir)) {
      out => out println str
    }

  } 

}
