/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */
package Liszt
import Liszt._
import Liszt.Types._
import Liszt.Trees._
import scala.collection.immutable.{Set => SET}

class UpdatePlacement(topLevel : Block, globals : SET[String]) {
  
/*
  sealed abstract class FieldState
  case object Top extends FieldState
  case object Unknown extends FieldState
  case class Delayable(val field : String, val op : ReductionOp) extends FieldState
  case class CondDelayable(val field : String, val op : ReductionOp) extends FieldState
  case object Clean extends */
  //lattice is:
  //  Top (top)
  //  Unknown the state of the field is unknown b/c it came from a function entry, in the lattice this connects to the specific cond delayables for each reduction
  //  Delayable x ... (for all reductions x) field is dirty with x reduction but we can wait to do the update
  //  CondDelayable x ... (for all reduction x)  field may be dirty with x reduction if the field came into the function with x reduction, or it will be clean because of a conditionall update that was inserted
  //  Clean (bottom)
  //
  sealed abstract class FieldInfo
  case object UnknownState extends FieldInfo
  case object MultipleStates extends FieldInfo
  case object Top extends FieldInfo
  case class FieldState(field : Var, state : Ops.ReductionOp) extends FieldInfo {
     def asTree = EnsureFieldState(field,state)
     override def equals(rhs : Any) = {
     	rhs match {
     		case FieldState(v,s) => (v.name == field.name && s == state)
     		case _ => false
     	}
     }
     override def hashCode() = field.name.hashCode + state.hashCode
  }
  
  sealed abstract class WriteStatus
  case object Clean extends WriteStatus
  case object Dirty extends WriteStatus
  case object UnknownStatus extends WriteStatus

  private def union[A](a : List[SET[A]]) : SET[A] = a.foldLeft(SET.empty[A])(_ ++ _)

  val varToResult : Map[String,FunctionInfo] = Map.empty ++ {
    for(e @ DefFunc(t,name,args,blk) <- topLevel.stmts if blk != ExternStmt)
      yield (name, new FunctionInfo(e))
  }
  
  def withUpdates : Block = {
  	val stmts = for(s <- topLevel.stmts) yield s match {
  		case DefFunc(_,name,_,x) if x != ExternStmt => varToResult(name).cfg.toAST
  		case x => x 
  	}
  	Block(stmts, topLevel.exp) copyAttr topLevel
  }
  //TODO(zach): We need to unify the check for reads/writes that occurs inside the Forall block
  //with the whole function check to avoid massive code duplication.  
  //Right now some Binary functions are missed in getreads/getwrites leading to weird compiler crashes
  private def getreads(es : List[Tree]) : SET[FieldState] = {
      val fr = es flatMap {
      	case FieldRead(t,f,a) => List(FieldState(f,Ops.RNone))
      	case lv @ SparseVectorAccess(_,v,_) if !lv.isWrite => List(FieldState(v,Ops.RNone))
      	case lv @ SparseMatrixAccess(_,m,_,_) if !lv.isWrite => List(FieldState(m,Ops.RNone))
      	case _ => List()
      }
      val calls = for(Call(t,func,args) <- es if varToResult.contains(func.name))
        yield varToResult(func.name).reads
      union(calls) ++ fr
    }
  private def getwrites(es : List[Tree]) : SET[FieldState] = {
    val fw = es flatMap {
    	case FieldWrite(f,args,op,a) => List(FieldState(f,op.asReduction))
    	case Assign(SparseVectorAccess(_,v,_),op,_) => List(FieldState(v,op.asReduction))
    	case Assign(SparseMatrixAccess(_,m,_,_),op,_) => List(FieldState(m,op.asReduction))
    	case _ => List()
    }   
    val calls = for(Call(t,func,args) <- es if varToResult.contains(func.name))
      yield varToResult(func.name).writes
    union(calls) ++ fw
  }
  
  private def getGlobalWrites(blk : Block) : SET[FieldState] = {
  	 val gw = for(Assign(g @ Var(_,name),op,a) <- blk if globals.contains(name)) yield FieldState(g,op.asReduction)
  	 val gwv = for(Assign(VecExtract(_,g @ Var(_,name),_),op,a) <- blk if globals.contains(name)) yield FieldState(g,op.asReduction)
  	 val gwm = for(Assign(MatExtract(_,g @ Var(_,name),_,_),op,a) <- blk if globals.contains(name)) yield FieldState(g,op.asReduction)
  	 SET.empty ++ gw ++ gwv ++ gwm
  }
  
  private def getGlobalReads(blk : Block) : SET[FieldState] = {
      SET.empty ++ {
      	for(v @ Var(_,name) <- blk if !v.isWrite && globals.contains(name))
      		yield FieldState(v, Ops.RNone)
      }
  }
  //type ValueType = Map[String,FieldState]
  class FunctionInfo(func : DefFunc) {
    lazy val cfg = {
       val cfg_ =  new CFGs.CFG(func)
       cfg_.modify {
        (e:Stmt) =>
         val efs = e match {
          case DefVar(_,exp) => exp match {
            case FieldRead(t,field,arg) => List(EnsureFieldState(field,Ops.RNone))
            case SparseVectorAccess(_,v,_) => List(EnsureFieldState(v,Ops.RNone))
            case SparseMatrixAccess(_,m,_,_) => List(EnsureFieldState(m,Ops.RNone))
            
            case Binary(_ : SparseVectorType,m,_,v) => {
            	List(EnsureFieldState(m,Ops.RNone),EnsureFieldState(v,Ops.RNone))
            }
            case _ => List()
          }
          case FieldWrite(field,arg,op,rhs) => List(EnsureFieldState(field,op.asReduction))
          case Assign(lhs,op,rhs) => {
          	lhs match {
          		case SparseVectorAccess(_,v,_) => List(EnsureFieldState(v,op.asReduction))
          		case SparseMatrixAccess(_,m,_,_) => List(EnsureFieldState(m,op.asReduction))
          		case _ => List()
          	}
          }
          case fa @ Forall(DefVar(name,_),_,set,block) => {
          	  
          	  val es = for(e <- block) yield e
              val freads = getreads(es)
              val fwrites = getwrites(es)

              val globalDefs = SET.empty ++ { for(DefVar(name,_) <- block if globals.contains(name)) yield name }
              val greads = getGlobalReads(block) filter (x => !globalDefs.contains(x.field.name))
              val gwrites = getGlobalWrites(block) filter (x => !globalDefs.contains(x.field.name))
              
              val reads = freads ++ greads
              val writes = fwrites ++ gwrites
              
              val all = (reads ++ writes)
              val allgrouped = Map.empty ++ all.groupBy( (a:FieldState) => a.field.name )
              val allupdates =
              for( set <- allgrouped.valuesIterator ) yield {
                  if(set.size == 1) {
                    set.head
                  } else {
                    if(globals.contains(set.head.field.name)) {
                        throw new Exception("Non-disjoint reader/write set: " + set)
                    } else {
                        fa.mustBeTrivial = true //we allow read/write conflicts as long as it turns out that the loop is trivial (i.e. writes only to its iteration variable)
                        FieldState(set.head.field,Ops.RModify)
                    }
                  }
              }
              allupdates.toList.map(_.asTree)
          }
          //TODO(zach): handle LinearSystemNonzeroes
          case _ => List()
        }
        val global_rw = e match {
            case _ : Forall => List() //ignore forall blocks, we only care about unnested accesses
            case stmt => for( g @ Var(_,name) <- stmt if globals.contains(name))
        	    yield EnsureFieldState(g,Ops.RNone)
        }
        val update_stmts = efs ::: global_rw
        val supdate_stmts = update_stmts.sortWith( (a,b) => {
               (a.state,b.state) match {
                  case (Ops.RNone, Ops.RNone) => a.field.name > b.field.name
                  case (_,Ops.RNone) => true
                  case (Ops.RNone,_) => false
                  case (_,_) => a.field.name > b.field.name
               }
             })
        supdate_stmts ::: List(e)
       }
       cfg_
    }
    lazy val reads : SET[FieldState] = getreads(cfg.alltrees)
    lazy val writes : SET[FieldState] = getwrites(cfg.alltrees)
    lazy val removeRedundant = new RemoveRedundant(cfg,reads,writes)
    lazy val exitInfo = removeRedundant.outmap(cfg.exit)
    def run() { removeRedundant.remove() }
  }
  type ValueType = Map[String,(FieldInfo,WriteStatus)]
  class RemoveRedundant(val cfg : CFGs.CFG, reads : SET[FieldState], writes : SET[FieldState]) extends CFGs.OptimizationPass {
    type V = ValueType
    private val allfields = SET.empty ++ { for( EnsureFieldState(f,_) <- cfg.alltrees ) yield f.name }
    val bottom : V = Map.empty ++ allfields.map( (_,(MultipleStates,Dirty)) )
    val top : V = Map.empty ++ allfields.map( (_,(Top,Clean)) )
    val isForward = true
    val inputValue : V = Map.empty ++ allfields.map( (_,(UnknownState,UnknownStatus)) )
    def transferFn(e : Tree)(v : V) : V = e match {
      case EnsureFieldState(field,Ops.RNone) => {
          val (fs,ws) = v(field.name)
          v + (field.name -> (fs,Clean))
      }
      case EnsureFieldState(field,state) => {
          val (fs,ws) = v(field.name)
          v + (field.name -> (FieldState(field,state),Dirty))
      }
      case Call(t,f,args) => {
          for( (name,(fs,ws)) <- varToResult(f.name).exitInfo) yield name ->
          (fs match {
            case UnknownState => v.getOrElse(name,(UnknownState,UnknownStatus))._1
            case _ => fs
          },
          ws match {
            case UnknownStatus => v.getOrElse(name,(UnknownState,UnknownStatus))._2
            case _ => ws
          })
      }
      case _ => v
    }
    private def mapintersect[A,B,C,D](a : Map[A,B], b : Map[A,C], fn : (B,C) => D) : Map[A,D] = Map.empty ++ {
      for( i <- a.keySet ++ b.keySet) yield (i,fn(a(i),b(i)))
    }
    
    def fieldmeet(a : FieldInfo, b : FieldInfo) : FieldInfo = (a,b) match {
      case (MultipleStates,_) => MultipleStates
      case (_,MultipleStates) => MultipleStates
      case (Top,b) => b
      case (b,Top) => b
      case (FieldState(f1,s1),FieldState(f2,s2)) =>
        if(s1 == s2) a else UnknownState
      case (UnknownState,UnknownState) => UnknownState
      case (FieldState(f1,s1),UnknownState) => MultipleStates
      case (UnknownState,FieldState(f1,s1)) => MultipleStates
    }
    def writemeet(a : WriteStatus, b : WriteStatus) = (a,b) match {
      case (Dirty,_) => Dirty
      case (_,Dirty) => Dirty
      case (Clean,a) => a
      case (a,Clean) => a
      case (UnknownStatus,UnknownStatus) => UnknownStatus
    }
    def pairmeet( a : (FieldInfo,WriteStatus), b : (FieldInfo,WriteStatus)) : (FieldInfo,WriteStatus) =
      (fieldmeet(a._1,b._1), writemeet(a._2,b._2))
    def meet(a : V, b : V) : V = mapintersect(a,b,pairmeet)
    
    def remove() {
      def removeRedundant(eo : Option[Stmt], input : V, output : V) = eo match {
         case Some(e) => {
             e match {
                case EnsureFieldState(f,Ops.RNone) =>
                  input(f.name)._2 match {
                    case Clean => List()
                    case _ => List(e)
                  }
                case EnsureFieldState(f,state) =>
                  input(f.name)._1 match {
                    case FieldState(f2,state2) if state == state2 => List()
                    case _ => List(e)
                  }
                case _ => List(e)
             }
         }
         case None => List()
      }
      modifyBlocks(removeRedundant)
    }
  }
  for((name,info) <- varToResult) {
    info.run()
    //info.cfg.draw()
  }
}





class UpdatePlacement2(topLevel : Block, globals : SET[String]) {
 
  object Log extends Liszt.Log("UpdatePlacement")
  
  sealed abstract class FieldState
  case object EntryState extends FieldState //state of the object on entry to the function,
                                           //whatever it may be, if we find that exit == EntryState,
                                           //then this indicates that the function does not modify that field's state 
  case class ConstantState(state : Ops.ReductionOp) extends FieldState //field must be in a specific state
  case object MultipleStates extends FieldState //field may be in many states, (bottom)
  case object Top extends FieldState //no knowledge has reach the node yet
  
  //            Top
  //  /          | ............... \
  // EntryState ConstantState0 ... ConstantStateN 
  
  def withUpdates : Block = {
  	val with_updates = insertUpdates(topLevel)
  	val removed_redundant = removeRedundantUpdates(with_updates)
  	removed_redundant
  }
	
	def insertUpdates(toplevel : Block) : Block = {
		object ClassScope { //getStates includes a forward reference, which requires it be defined at class scope
			val getStates : Tree => List[(Var,Ops.ReductionOp)] = Util.Memoize[Tree,List[(Var,Ops.ReductionOp)]] {
				case FieldRead(t,f,a) => List( f -> Ops.RNone )
				case lv @ SparseVectorAccess(_,v,_) if !lv.isWrite => List( v -> Ops.RNone ) 
				case lv @ SparseMatrixAccess(_,m,_,_) if !lv.isWrite => List( m -> Ops.RNone )
				case c @ Call(t,func,args) if !c.isExternal => (args flatMap getStates) ::: getStates(toplevel.function_info(func.name).impl)
				case FieldWrite(f,args,op,a) => List(f -> op.asReduction)
				case Assign(SparseVectorAccess(_,v,_),op,_) => List(v -> op.asReduction)
				case Assign(SparseMatrixAccess(_,m,_,_),op,_) => List(m -> op.asReduction)
				case Assign(g @ Var(_,name),op,a) if globals.contains(name) => List(g -> op.asReduction)
				case Assign(VecExtract(_,g @ Var(_,name),_),op,a) if globals.contains(name) => List(g -> op.asReduction)
				case Assign(MatExtract(_,g @ Var(_,name),_,_),op,a) if globals.contains(name) => List(g -> op.asReduction)
				case v @ Var(_,name) if !v.isWrite && globals.contains(name) => List(v -> Ops.RNone)
				case x => x match {
					case ShapeAny(tree,names,trees) => {
						val defines = for(DefVar(name,_) <- trees if globals.contains(name)) yield name
						val states = trees flatMap getStates
						states.filter { case (v,_) => !defines.contains(v.name) } //remove states that do not escape the scope of this statment
					}
				}
			}.apply
		
			def sortEnsures(es : List[EnsureFieldState]) : List[EnsureFieldState] = {
				def compare_state(a : EnsureFieldState, b : EnsureFieldState) = (a.state,b.state) match {
					case (Ops.RNone, Ops.RNone) => a.field.name > b.field.name
					case (_,Ops.RNone) => true
					case (Ops.RNone,_) => false
					case (_,_) => a.field.name > b.field.name
				}
				es.sortWith(compare_state)
			}
			def getUniqueStates(tree : Tree) : List[EnsureFieldState] = {
				val states = getStates(tree)
				val state_map = scala.collection.mutable.Map[String,EnsureFieldState]()
				
				for( (v,op) <- states) {
					val current_state = state_map.getOrElseUpdate(v.name,EnsureFieldState(v,op))
					if(current_state.state != op) {
						Log.error("Field %s is in two states (%s,%s) at one time in a for-comprehension.",v.name,current_state,op)
					}
				}
				//we sort the new statements so they come out in the same order every time.
				val sorted_states = sortEnsures(state_map.values.toList)
				List.empty ++ sorted_states
			}
			
			def globalWritesToReads(es : List[EnsureFieldState]) : List[EnsureFieldState] = {
				val fixed = for(e <- es) yield {
					if(globals.contains(e.field.name))
						EnsureFieldState(e.field,Ops.RNone)
					else e
				}
				sortEnsures(fixed)
			}
			
			def in_forall(tree : Tree) : Boolean = tree.context match {
				case Some(_ : Forall) => true
				case Some(x) => in_forall(x)
				case None => false
			}
			def add_updates(tree : Tree) = ASTUtil.flatMapBlocks(tree) {
				case x if in_forall(x) => List(x) //statement only appears in a nested context
				case fa : Forall => getUniqueStates(fa) ::: List(fa)
				
				case dv @ DefVar(_,c : Call) => (c.children flatMap getUniqueStates) ::: List(dv)
				case es @ ExpStmt(c : Call) => (c.children flatMap getUniqueStates) ::: List(es)
				//ignore control flow and call statments, we only want to annotate statements
				case non_leaf if !non_leaf.isLeaf => List(non_leaf)
				case stmt => globalWritesToReads(getUniqueStates(stmt)) ::: List(stmt) //writes to globals at the unnested level are treated as reads because
				                                                                       //we want the field clean when writing to it at the top level
			}
			val with_updates = for(stmt <- toplevel.stmts) yield {
				stmt match {
					case df : DefFunc => df.copy(body = add_updates(df.body).asInstanceOf[Stmt])
					case x => x
				}
			}
			val ret = Block(with_updates,toplevel.exp)
		}
		ClassScope.ret
	}
	def removeRedundantUpdates(toplevel : Block) : Block = {
		object ClassScope { //function_info_map has a forward reference to FunctionInfo, requiring it be defined at class scope
			val function_info_map = Map.empty ++ { 
				for(func <- toplevel.functions) yield
					func.name -> new FunctionInfo(func)
			}
			def all_ensure_state_objs(t : Tree) : SET[String] = t match {
				case EnsureFieldState(field,_) => SET(field.name)
				case x => SET.empty ++ { x.children flatMap all_ensure_state_objs }
			}
			val id_to_ensure_state_obj = all_ensure_state_objs(toplevel).toArray
			val ensure_state_obj_to_id = Map.empty ++ (id_to_ensure_state_obj.zipWithIndex)
			
			val new_stmts = for(stmt <- toplevel.stmts) yield {
				stmt match {
					case df : DefFunc => { 
						function_info_map(df.name).toTree.asInstanceOf[Stmt]
					}
					case _ => stmt
				}
			}
			class FunctionInfo(func : DefFunc) extends CFGs.OptimizationPass2 {
				override lazy val cfg = new CFGs.CFG2(func)
				type V = Map[String,FieldState]
				
				override lazy val top = Map.empty ++ { for( (k,v) <- ensure_state_obj_to_id) yield k -> Top }
				override def isForward = true
				override lazy val inputValue = Map.empty ++ { for( (k,v) <- ensure_state_obj_to_id) yield k -> EntryState }
				override def transferFn(e : Tree)(v : V) : V = e match {
					case EnsureFieldState(f,state) => v + (f.name -> ConstantState(state))
					case c @ Call(t,fname,args) if !c.isExternal => {
						val exit_state = function_info_map(fname.name).exitState
						val new_state = exit_state ++ {
							for((k,v) <- v if exit_state(k) == EntryState )
								yield k -> v
						}
						new_state
					}
					case _ => v
				}
				def fieldStateMeet(a : FieldState, b : FieldState) : FieldState = (a,b) match {
					case (Top,x) => x
					case (x,Top) => x
					case (x,y) if x == y => x
					case _ => MultipleStates
				}
				override def meet(a : V, b : V) : V = {
					val result = Map.empty ++ {
						for( (k,v) <- a) yield {
							(k -> fieldStateMeet(v,b(k)))
						}
					}
					result
				}
				override def equal(a : V, b : V) : Boolean = {
					a.equals(b)
				}
				
				def exitState = result(cfg.exit).out
				
				def toTree = cfg.toTree {
					(n : Node) =>
					cfg.treeForNode(n) match {
						case Some(es @ EnsureFieldState(f,op)) => result(n).in(f.name) match {
							case ConstantState(op2) if op == op2 => List()
							case _ => List(es)
						}
						case _ => cfg.nodeIdentity(n)
					}
				}
				
			}
			val ret = Block(new_stmts,toplevel.exp)
		}
		ClassScope.ret
	}
}