package Liszt

object Util {
	def WriteFile(filename : String) (blk : (java.io.PrintWriter) => Any) = {
		//Uses the Loan pattern to give access to a java file writer inside the passed 'blk' function value.
		val file = new java.io.FileOutputStream(filename)
		val out = new java.io.PrintWriter(file)
		blk(out)
		out.close()
	}
	
	//creates a memoized version of fn. Fn should be side-effect free.
	//If fn needs to call itself it should itself indirectly through the memoize object
	case class Memoize[A,B](fn : A => B) {
		private val data = new scala.collection.mutable.HashMap[A,B]()
		def apply(a : A) : B = data.getOrElseUpdate(a,fn(a))
	}
	
	def merge[A,B,C,D](small : Map[A,B], large : Map[A,B])(fn : (B,B) => B) : Map[A,B] = {
		if(small.size > large.size)
			merge(large,small) { (a : B,b : B) => fn(b,a) }
		else {
			large ++ {
				for( (k,v) <- small) yield large.get(k) match {
					case Some(lhs) => k -> fn(lhs,v)
					case None => k -> v
				}
			}
		}
	}
	
	def removeAssignmentsToFieldsAndSets(b : Trees.Block) : Trees.Block = {
		val nstmts = b.stmts.filter {
			case Trees.Assign(a,op,b) if ASTUtil.isFieldT(a.t) || ASTUtil.isSetT(a.t) => false
			case _ => true
		}
		Trees.Block(nstmts,b.exp)
	}
}

//helper classes for code generation
trait CodeLines {
	
	class Line(val i : Int, str : String) {
		override def toString = ("    " * i) + str
		def indent = new Line(i+1,str)
	}
	def indent(l : Line) : Line = l.indent
	def indent(ls : List[Line]) : List[Line] = ls map indent
	def line(str : String) : List[Line] = List(new Line(0,str))
	def line(strs : List[String]) : List[Line] = for(s <- strs) yield new Line(0,s)
	class RichStringLine(str : String) {
		def l = line(str)
	}
	implicit def toRichStringLine(str : String) : RichStringLine = new RichStringLine(str)
}