/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */
package Liszt
import Liszt.Types._
import Liszt.Trees._

class VecCodeGen(val top : CompilationUnit) extends CLikeCodeGen {
	import ASTUtil._
	import CodeLineImplicits._
	val Log = new Log("CodeGen")
	
	object typedef {
		private val map = new scala.collection.mutable.HashMap[String,String]
		var typedefs : List[String] = List()
		def forType(ctype : String) = map.getOrElseUpdate(ctype, {
			if(ctype.contains(",")) {
				val n = Names.create("type_ref")
				typedefs = "typedef %s %s;".format(ctype,n) :: typedefs
				n
			} else {
				ctype
			}
		})
	}
	override def tTypeFor(t : Tree) = typedef.forType(super.tTypeFor(t))
	
	override def tType(t : Type) = t match {
		case BoolT => "int" //avoid bit_refernces in vectors by compiling bools as integers
		case UnitT => "int"
		case _ => typedef.forType(super.tType(t))
	}
	
	def wrappedType(s : String) = "Value< %s >".format(s)
	
	val liszt_obj = "__liszt"
	val fluent_file = "__fluent"
	
	
	def testWatchpoint(position : Position) : String = {
	  "__liszt.testWatchPoints(%s,%s);".format("\"%s\"" format position.filename,position.lineno)
	}
	
	var isTemplated = false //hack to keep track of if we are inside a template (need _TM on forall macro)
	def tFunctionHeader(e : DefFunc, isDecl : Boolean) = {
		val fnt = getCType(e).asInstanceOf[Func]
		val name = e.name
		val args = for(a <- e.args) yield a.name
		val template_header = if(fnt.args.exists(_.isInstanceOf[SpecialType])) {
			val template_args = for( SpecialType(_,typ) <- fnt.args) yield "typename %s".format(typ) 
			if(!isDecl) 
				isTemplated = true
			"template< %s > ".format(template_args.mkString(","))
		} else {
			""
		}
		val argwt = for( (t,a) <- fnt.args zip args) yield "%s & %s".format(wrappedType(tType(t)),a)
		val return_type = wrappedType(tType(fnt.range))
		"%s%s %s(%s)".format(template_header,return_type,name,argwt.mkString(","))
	}
	
	override def tVar(v : Var) = v.t match {
		case _ : Func => super.tVar(v)
		case _ => "VGET(%s)".format(super.tVar(v))
	}
	
	override def tFieldWrite(e : FieldWrite) =  { for(line <- super.tFieldWrite(e)) yield "VEXEC %s".format(line) }  ::: testWatchpoint(e.position.get)
	
	override def tIf(e : If) = {
		val If(blk,thenE,elseE) = e
		val newVar = Names.create("if_pred")
		blk.stmts.flatMap(tStmt) :::
		tStmt(DefVar(newVar,blk.exp) copyAttr blk.exp) :::
		"VIF(%s)".format(newVar) :::
			indent(tStmt(thenE)) :::
		"VELSE(%s)".format(newVar) :::
			indent(tStmt(elseE)) :::
		"VEND_IF"
	}
	
	override def tWhile(e : While) = {
		val While(pred,block) = e
		val lbl = Names.create("while_label")
		val bexp = pred.exp
		val exps = pred.stmts
		val newVar = Names.create("while_pred")
		"VWHILE(%s)".format(lbl) :::
			indent(exps.flatMap(tStmt) :::
				   tStmt(DefVar(newVar, bexp) copyAttr bexp)) :::
		"VWHILE_EXP(%s,%s)".format(lbl,newVar) :::
			indent(tStmt(block)) :::
		"VEND_WHILE(%s)".format(lbl)
	}
	
	override def tDefFunc(e : DefFunc) = e match {
		case DefFunc(_,_,_,ExternStmt) => super.tDefFunc(e)
		case DefFunc(t,name,args,body) => {
			val range = tType(getCType(e).asInstanceOf[Func].range)
			val result = "%s {".format(tFunctionHeader(e,false)) :::
				"VFUNCTION(%s);".format(range) :::
				indent(tStmt(body)) ::: 
				"VEND_FUNCTION" :::
			"}"
			isTemplated = false
			result
		}
	}
	
	def isPointerType(t : Type) = false
	def tFieldInit(e : FieldInit) = Log.ierror("field init unexpected.")
	def tSetInit(e : SetInit) = Log.ierror("set init unexpected.")
	def tIfExp(e : IfExp) = Log.ierror("unexpected ifexp")
	
	override def tDefVar(e : DefVar) = {
		e.init match {
			case _ : EmptyExp => "%s %s;".format(wrappedType(tTypeFor(e)),e.name) //variable decl
			case fi @ FieldInit(_,exp) => {
				val real_name = ASTUtil.nameFromTempHack(e.name)
				val init : List[CodeLine] = exp match {
					case ConstSource(exp) => "VEXEC %s.buildFieldWithConst(%d,\"%s\",%s,&VGET(%s));".format(liszt_obj,fi.id,real_name,tVar(exp),real_name)
					case StringSource("position") => "VEXEC %s.buildFieldWithPositions(%d,\"%s\",&VGET(%s));".format(liszt_obj,fi.id,real_name,real_name)
					case _ => Log.error(e.report("field initialization symbol not supported"))
				}
				init ::: testWatchpoint(e.position.get)
			}
			case SetInit(t,set_name) => {
				"Value< %s > %s;".format(tTypeFor(e),e.name) :::
				"VEXEC %s.buildSet<%sType>(\"%s\",&VGET(%s));".format(liszt_obj,tType(t.asInstanceOf[Set].t),set_name,e.name)
			}
			case Call(_,fn,args) if !isExternal(fn.name) => {
				"%s %s = %s(%s);".format(wrappedType(tTypeFor(e)),e.name,fn.name, {for(a <- args) yield a.name}.mkString(","))
			}
			case _ : IfExp => super.tDefVar(e)
			case exp => "VDEFINE(%s,%s) = %s;".format(tTypeFor(e),e.name,tExp(exp))
		}
	}
	
	override def tReturn(e : Return) = e.e match {
		case _ : EmptyExp => "VRETURN(0);"
		case exp => "VRETURN(%s);".format(tExp(exp)) 
	}
	
	override def tExpStmt(e : ExpStmt) = e.e match {
		case Call(_,fn,args) if !isExternal(fn.name) => "%s(%s);".format(fn.name,{for(a <- args) yield a.name}.mkString(","))
		case _ : EmptyExp => List()
		case _ => "VEXEC %s;".format(tExp(e.e))
	}
	
	override def tEnsureFieldState(e: EnsureFieldState) = List()
	
	override def tForall(e : Forall) = {
		val Forall(name,lblOp,set,block) = e
		val suffix = if(isTemplated) "_TM" else ""
		lblOp match {
			case None => "VFORALL%s(%s,%s)".format(suffix,name.name,set.name) ::: indent(tStmt(block)) ::: "VEND_FORALL"
			case Some(lbl) => "VFORALL_LABELED%s(%s,%s,%s)".format(suffix,name.name,lbl.name,set.name) ::: indent(tStmt(block)) ::: "VEND_FORALL"
		}
	}
	
	override def tAssign(e : Assign) = e match {
		case Assign(name,op,rhs) if isFieldT(rhs.t) => List()
		case Assign(name : Var,Ops.Assign,rhs) => "VASSIGN(%s) = %s;".format(name,tExp(rhs))
		case Assign(name : Var,Ops.Reduce(op),rhs) => "VASSIGN(%s) = %s;".format(name,tExp(Binary(name.t,name,op,rhs)))
		case Assign(VecExtract(_,v,i),Ops.Assign,rhs) => "VASSIGN(%s)[%s] = %s;".format(v,tExp(i),tExp(rhs))
		case Assign(ve @ VecExtract(_,v,i),Ops.Reduce(op),rhs) => "VASSIGN(%s)[%s] = %s;".format(v,tExp(i),tExp(Binary(rhs.t,Var(ve.t,tVecExtract(ve)),op,rhs)))
		case Assign(MatExtract(_,v,r,c),Ops.Assign,rhs) => "VASSIGN(%s)(%s,%s) = %s;".format(v,tVar(r),tVar(c),tExp(rhs))
		case Assign(ve @ MatExtract(_,v,r,c),Ops.Reduce(op),rhs) => "VASSIGN(%s)(%s,%s) = %s;".format(v,tVar(r),tVar(c),tExp(Binary(rhs.t,Var(ve.t,tMatExtract(ve)),op,rhs)))
	}
	override def tPrint(e : Print) = "VEXEC std::cout << \"liszt: \" << %s std::endl;".format(e.es.map(tVar).mkString(""," << "," << "))
	
	def create(build_dir : String) : Unit = { 
		val toplevel = normalizeGlobals(top.code)
		
		val deflines = tStmt(toplevel.var_decls) ::: tStmt(toplevel.func_decls) ::: tStmt(toplevel.func_defs)
		val initlines = tStmt(toplevel.var_defs)
		val helper_fns = tHelperFunctions().mkString("\n")
		val str = """
#define ID __ID
#include "ScalaVecRuntime.h"
namespace __liszt_client {
//typedefs
%s
//helper functions
%s
//definitions
%s
void LisztClient_main() {
//initializations
%s
	_liszt_start();
}

} //namespace __liszt_client
void LisztClient_main() {
    __liszt_client::LisztClient_main();
}
::LisztSinglePrivate::LisztSingle * LisztClient_liszt() {
	return &__liszt;
}
void LisztClient_predescriptor(int argc, char ** argv) {
	__liszt.init(argc,argv);
}
#undef ID
		""".format(typedef.typedefs.reverse.mkString("\n"),helper_fns,deflines.mkString("\n"),indent(initlines).mkString("\n"))
		
		Util.WriteFile("%s/vis/main.cpp".format(build_dir)) {
			out => out println str
		}
	} 
}
