import Liszt.Language._
import Liszt.MetaInteger._


@lisztcode 
trait LinearHexLS extends LinearSystem {
  def row(v : Vertex) : RowIndex = AutoIndex
  def col(v : Vertex) : ColIndex = AutoIndex
  def rows(c : Cell) = Vec(row(vertex(c,0)),
			   row(vertex(c,1)),
			   row(vertex(c,2)),
			   row(vertex(c,3)),
			   row(vertex(c,4)),
			   row(vertex(c,5)),
			   row(vertex(c,6)),
			   row(vertex(c,7)))
  def cols(c : Cell) = Vec(col(vertex(c,0)),
			   col(vertex(c,1)),
			   col(vertex(c,2)),
			   col(vertex(c,3)),
			   col(vertex(c,4)),
			   col(vertex(c,5)),
			   col(vertex(c,6)),
			   col(vertex(c,7)))
}

/*
@lisztcode 
trait QuadraticTriangleLS[N <: IntM] extends LinearSystem {
  def row(v : Vertex) : Vec[N,RowIndex] = AutoIndex
  def row(e : Edge) : Vec[N,RowIndex] = AutoIndex
  def row(f : Face) : Vec[N,RowIndex] = AutoIndex
  def row(c : Cell) : Vec[N,RowIndex] = AutoIndex
  def col(v : Vertex) : Vec[N,ColIndex] = AutoIndex
  def col(e : Edge) : Vec[N,ColIndex] = AutoIndex
  def col(f : Face) : Vec[N,ColIndex] = AutoIndex
  def col(c : Cell) : Vec[N,ColIndex] = AutoIndex
}*/

@lisztcode
object MyLS extends LinearHexLS with SuperLU


@lisztcode
object HexFem {
  val position = FieldWithLabel[Vertex,Vec[_3,Double]]("position")

  // Map the face based boundary onto the vertices.
  val wall = BoundarySet[Face]("wall")
  val isBoundary  = FieldWithConst[Vertex,Double](0.0)

  def SetBoundary() : Unit = {
    for (f <- wall) {
      for (v <- vertices(f)) {
	isBoundary(v) = 1.0
      }
    }
  }

  // Weights and points for the 3D 8-point Gauss quadrature.
  // Accurate up to degree two in each coordinate direction.
  val nint = 8
  val weights = Vec(1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0)
  val qp = pow(1.0/sqrt(3.0), 3)
  val r = Vec(qp, qp, qp, qp,-qp,-qp,-qp,-qp)
  val s = Vec(qp, qp,-qp,-qp, qp, qp,-qp,-qp)
  val t = Vec(qp,-qp, qp,-qp, qp,-qp, qp,-qp)

  def det(A : Mat[_2,_2,Double]) : Double = {
    return A(1,1)*A(2,2) - A(1,2)*A(2,1);
  }

  def trans(A: Mat[_3,_8,Double]) : Mat[_8,_3,Double] = {
    var r = 0
    var c = 0
    var Atrans = Mat(Vec(0.0,0.0,0.0),
		     Vec(0.0,0.0,0.0),
		     Vec(0.0,0.0,0.0),
		     Vec(0.0,0.0,0.0),
		     Vec(0.0,0.0,0.0),
		     Vec(0.0,0.0,0.0),
		     Vec(0.0,0.0,0.0),
		     Vec(0.0,0.0,0.0))
    while (r < 3) {
      c = 0
      while (c < 8) {
	Atrans(c,r) = A(r,c)
	c += 1
      }
      r += 1
    }
    return Atrans
  }

  //def ludcmp[N <: IntM](A : Mat[N,N,Double], n : Int) : Int = {
  def ludcmp(A : Mat[_3,_3,Double], indx : Vec[_3,Int], n : Int)
  : Mat[_3,_3,Double] = {
    var d = 0
    var i = 0
    var j = 0
    var k = 0
    var imax = 0
    var big = 0.0
    var dum = 0.0
    var sum = 0.0
    var temp = 0.0
    var vv = Vec(0.0,0.0,0.0)
    var LU = A

    d = 1
    i = 0
    while (i < n) {
      big = 0.0
      j = 0
      while (j < n) {
	temp = fabs(LU(i,j))
	if (temp > big) {
	  big = temp
	}
	j += 1
      }
      if (big == 0.0) {
	Print("Singular matrix in routine ludcmp")
      }
      vv(i) = 1.0 / big
      i += 1
    }

    j = 0
    while (j < n) {
      i = 0
      while (i < j) {
	sum = LU(i,j)
	k = 0
	while (k < i) {
	  sum -= LU(i,k)*LU(k,j)
	  k += 1
	}
	LU(i,j) = sum
	i += 1
      }

      big = 0.0
      i = j
      while(i < n) {
	sum = LU(i,j)
	k = 0
	while (k < j) {
	  sum -= LU(i,k)*LU(k,j)
	  k += 1
	}
	LU(i,j) = sum
	dum = vv(i)*fabs(sum)
	if (dum >= big) {
	  big = dum;
	  imax = i
	}
	i += 1
      }
      if (j != imax) {
	k = 0
	while (k < n) {
	  dum = LU(imax,k)
	  LU(imax,k) = LU(j,k)
	  LU(j,k) = dum
	  k += 1
	}
	d = -d
	vv(imax) = vv(j)
      }
      indx(j) = imax
      if (LU(j,j) == 0.0) {
	LU(j,j) = 1e-20
      }
      if (j != n-1) {
	dum = 1.0 / LU(j,j)
	i = j+1
	while (i < n) {
	  // Note to Zach and Niels that this does not work...
	  // LU(i,j) *= dum
	  LU(i,j) = dum*LU(i,j)
	  i += 1
	}
      }
      j += 1
    }
    Print("ludcmp: The LU ", LU, " and the pivots ", indx)
    return LU
  }

  def lupivot(A : Mat[_3,_3,Double], n : Int)
  : Vec[_3,Int] = {
    var d = 0
    var i = 0
    var j = 0
    var k = 0
    var imax = 0
    var big = 0.0
    var dum = 0.0
    var sum = 0.0
    var temp = 0.0
    var vv = Vec(0.0,0.0,0.0)
    var LU = A
    var indx = Vec(0,0,0)

    d = 1
    i = 0
    while (i < n) {
      big = 0.0
      j = 0
      while (j < n) {
	temp = fabs(LU(i,j))
	if (temp > big) {
	  big = temp
	}
	j += 1
      }
      if (big == 0.0) {
	Print("Singular matrix in routine ludcmp")
      }
      vv(i) = 1.0 / big
      i += 1
    }

    j = 0
    while (j < n) {
      i = 0
      while (i < j) {
	sum = LU(i,j)
	k = 0
	while (k < i) {
	  sum -= LU(i,k)*LU(k,j)
	  k += 1
	}
	LU(i,j) = sum
	i += 1
      }

      big = 0.0
      i = j
      while(i < n) {
	sum = LU(i,j)
	k = 0
	while (k < j) {
	  sum -= LU(i,k)*LU(k,j)
	  k += 1
	}
	LU(i,j) = sum
	dum = vv(i)*fabs(sum)
	if (dum >= big) {
	  big = dum;
	  imax = i
	}
	i += 1
      }
      if (j != imax) {
	k = 0
	while (k < n) {
	  dum = LU(imax,k)
	  LU(imax,k) = LU(j,k)
	  LU(j,k) = dum
	  k += 1
	}
	d = -d
	vv(imax) = vv(j)
      }
      indx(j) = imax
      if (LU(j,j) == 0.0) {
	LU(j,j) = 1e-20
      }
      if (j != n-1) {
	dum = 1.0 / LU(j,j)
	i = j+1
	while (i < n) {
	  // Note to Zach and Niels that this does not work...
	  // LU(i,j) *= dum
	  LU(i,j) = dum*LU(i,j)
	  i += 1
	}
      }
      j += 1
    }
    return indx
  }

  def lubksb(A : Mat[_3,_3,Double], indx : Vec[_3,Int],
	     b : Vec[_3,Double], n : Int) : Vec[_3,Double] = {
    var i = 0
    var ii = -1
    var ip = 0
    var j = 0
    var sum = 0.0
    var x = b

    // The forward substitution.
    i = 0
    while (i < n) {
      ip = indx(i)
      sum = x(ip)
      x(ip) = x(i)
      if (ii >= 0) {
	j = ii
	while (j <= i-1) {
	  sum -= A(i,j)*x(j)
	  j += 1
	}
      } else if (sum != 0.0) {
	ii = i
      }
      x(i)= sum
      i += 1
    }
    // The backward substitution.
    i = n-1
    while (i >= 0) {
      sum = x(i)
      j = i+1
      while (j < n) {
	sum -= A(i,j)*x(j)
	j += 1
      }
      x(i) = sum / A(i,i)
      i -= 1
    }
    return x
  }

  def lubksbm(A : Mat[_3,_3,Double], indx : Vec[_3,Int],
	      B : Mat[_3,_8,Double], n : Int, m : Int)
  : Mat[_3,_8,Double] = {
    var k = 0
    var i = 0
    var ii = -1
    var ip = 0
    var j = 0
    var sum = 0.0
    var x = B

    // Perform the solve on each column of B.
    k = 0
    while (k < m) {
      // The forward substitution.
      i = 0
      while (i < n) {
	ip = indx(i)
	sum = x(ip,k)
	x(ip,k) = x(i,k)
	if (ii >= 0) {
	  j = ii
	  while (j <= i-1) {
	    sum -= A(i,j)*x(j,k)
	    j += 1
	  }
	} else if (sum != 0.0) {
	  ii = i
	}
	x(i,k)= sum
      i += 1
      }
      // The backward substitution.
      i = n-1
      while (i >= 0) {
	sum = x(i,k)
	j = i+1
	while (j < n) {
	  sum -= A(i,j)*x(j,k)
	  j += 1
      }
	x(i,k) = sum / A(i,i)
	i -= 1
      }
      k += 1
    }
    return x
  }

  def CellNodes(c : Cell) : Mat[_8,_3,Double] = {
    var nodes = Mat(Vec(0.0,0.0,0.0),
		    Vec(0.0,0.0,0.0),
		    Vec(0.0,0.0,0.0),
		    Vec(0.0,0.0,0.0),
		    Vec(0.0,0.0,0.0),
		    Vec(0.0,0.0,0.0),
		    Vec(0.0,0.0,0.0),
		    Vec(0.0,0.0,0.0))
    var i = 0
    while (i < 8) {
      var pos = position(vertex(c,i))
      nodes(i,0) = pos(_0)
      nodes(i,1) = pos(_1)
      nodes(i,2) = pos(_2)
      i = i + 1
    }
    return nodes
  }

  def LinearBasis(c : Cell, r : Double, s: Double, t: Double)
  : Vec[_8,Double] = {
    val a = 1.0/8.0
    var Na = Vec(-a*(r-1.0)*(s-1.0)*(t-1.0),
		 a*(r+1.0)*(s-1.0)*(t-1.0),
		 -a*(r+1.0)*(s+1.0)*(t-1.0),
		 a*(r-1.0)*(s+1.0)*(t-1.0),
		 a*(r-1.0)*(s-1.0)*(t+1.0),
		 -a*(r+1.0)*(s-1.0)*(t+1.0),
		 a*(r+1.0)*(s+1.0)*(t+1.0),
		 -a*(r-1.0)*(s+1.0)*(t+1.0))
    return Na
  }

  def LinearDerivatives(c : Cell, r : Double, s: Double, t: Double)
  : Mat[_3,_8,Double] = {
    val a = 1.0/8.0
    var Be = Mat(
      Vec(-a*(s-1.0)*(t-1.0),
	  a*(s-1.0)*(t-1.0),
	  -a*(s+1.0)*(t-1.0),
	  a*(s+1.0)*(t-1.0),
	  a*(s-1.0)*(t+1.0),
	  -a*(s-1.0)*(t+1.0),
	  a*(s+1.0)*(t+1.0),
	  -a*(s+1.0)*(t+1.0)),
      Vec(-a*(r-1.0)*(t-1.0),
	  a*(r+1.0)*(t-1.0),
	  -a*(r+1.0)*(t-1.0),
	  a*(r-1.0)*(t-1.0),
	  a*(r-1.0)*(t+1.0),
	  -a*(r+1.0)*(t+1.0),
	  a*(r+1.0)*(t+1.0),
	  -a*(r-1.0)*(t+1.0)),
      Vec(-a*(r-1.0)*(s-1.0),
	  a*(r+1.0)*(s-1.0),
	  -a*(r+1.0)*(s+1.0),
	  a*(r-1.0)*(s+1.0),
	  a*(r-1.0)*(s-1.0),
	  -a*(r+1.0)*(s-1.0),
	  a*(r+1.0)*(s+1.0),
	  -a*(r-1.0)*(s+1.0)))
    var JeT = LinearJacobian(c, Be)
    var indx = Vec(0,0,0)
    var JeTLU = ludcmp(JeT, indx, 3)
    indx = lupivot(JeT, 3)
    Be = lubksbm(JeTLU, indx, Be, 3, 8)
    return Be
  }

  def LinearJacobian(c : Cell, Be : Mat[_3,_8,Double])
  : Mat[_3,_3,Double] = {
    val nodes = CellNodes(c)
    return Be*nodes
  }

  def LinearDeterminant(c : Cell, r : Double, s: Double, t: Double)
  : Double = {
    val a = 1.0/8.0
    var Be = Mat(
      Vec(-a*(s-1.0)*(t-1.0),
	  a*(s-1.0)*(t-1.0),
	  -a*(s+1.0)*(t-1.0),
	  a*(s+1.0)*(t-1.0),
	  a*(s-1.0)*(t+1.0),
	  -a*(s-1.0)*(t+1.0),
	  a*(s+1.0)*(t+1.0),
	  -a*(s+1.0)*(t+1.0)),
      Vec(-a*(r-1.0)*(t-1.0),
	  a*(r+1.0)*(t-1.0),
	  -a*(r+1.0)*(t-1.0),
	  a*(r-1.0)*(t-1.0),
	  a*(r-1.0)*(t+1.0),
	  -a*(r+1.0)*(t+1.0),
	  a*(r+1.0)*(t+1.0),
	  -a*(r-1.0)*(t+1.0)),
      Vec(-a*(r-1.0)*(s-1.0),
	  a*(r+1.0)*(s-1.0),
	  -a*(r+1.0)*(s+1.0),
	  a*(r-1.0)*(s+1.0),
	  a*(r-1.0)*(s-1.0),
	  -a*(r+1.0)*(s-1.0),
	  a*(r+1.0)*(s+1.0),
	  -a*(r-1.0)*(s+1.0)))
    var JeT = LinearJacobian(c, Be)
    var indx = Vec(0,0,0)
    var JeTLU = ludcmp(JeT, indx, 3)
    indx = lupivot(JeT, 3)

    var je = 1.0
    var i = 0
    while (i < 3) {
      je *= JeTLU(i,i)
      if (indx(i) != i) {
	je *= -1.0
      }
      i += 1
    }
    return je
  }

  def LinearElementStiffness(c : Cell) : Mat[_8,_8,Double] = {
    var i = 0
    var Ke = Mat(Vec(0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0),
		 Vec(0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0),
		 Vec(0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0),
		 Vec(0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0),
		 Vec(0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0),
		 Vec(0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0),
		 Vec(0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0),
		 Vec(0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0))

    while (i < nint) {
      val Be = LinearDerivatives(c, r(i), s(i), t(i))
      val je = LinearDeterminant(c, r(i), s(i), t(i))
      Ke += weights(i)*je*trans(Be)*Be
      i += 1
    }
    return Ke
  }

  def LinearElementForce(c : Cell, Ke : Mat[_8,_8,Double]) : Vec[_8,Double] = {
    var i = 0
    var Fe = Vec(0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0)
    var fe = F(c)

    while (i < nint) {
      val Na = LinearBasis(c, r(i), s(i), t(i))
      val je = LinearDeterminant(c, r(i), s(i), t(i))
      Fe += weights(i)*je*Na*fe
      i += 1
    }
    return Fe
  }

  def LinearElementBoundary(c : Cell, Ke : Mat[_8,_8,Double])
  : Vec[_8,Double] = {
    var i = 0
    var ge = G(CellNodes(c))
    var Fg = -Ke*ge
    var Fh = Vec(0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0)
    return Fg + Fh
  }

  def F(c : Cell) : Double = {
    return 0.1
  }

  /*def F(nodes : Mat[_8,_3,Double]) : Vec[_8,Double] = {
    val f = Vec(0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0)
    var i = 0
    while (i < 8) {
      f(i) = 0.5*(nodes(i,0)*nodes(i,0) + nodes(i,1)*nodes(i,1))
      - nodes(i,2)*nodes(i,2)
      i += 1
    }
    return f
  }*/

  def G(nodes : Mat[_8,_3,Double]) : Vec[_8,Double] = {
    val g = Vec(0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0)
    return g
  }

  def H(nodes : Mat[_8,_3,Double]) : Vec[_8,Double] = {
    val h = Vec(0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0)
    return h
  }

  def main() {
    /*val Amat = Mat(Vec(1.0,0.0,1.0),Vec(0.0,1.0,1.0))
    val Gmat = Mat(Vec(2.0,0.0),Vec(0.0,3.0),Vec(0.0,1.0))
    val Cmat = Amat*Gmat
    val Dmat = Mat(Vec(8.0,1.0,6.0),Vec(3.0,5.0,7.0),Vec(4.0,9.0,2.0))
    val Bvec = Vec(1.0,2.0,3.0)
    val Bmat = Mat(Vec(1.0,2.0,3.0,4.0,5.0,6.0,7.0,8.0),
		   Vec(2.0,3.0,4.0,5.0,6.0,7.0,8.0,9.0),
		   Vec(3.0,4.0,5.0,6.0,7.0,8.0,9.0,10.0))
    val Be = Mat(Vec(1.0,2.0,3.0,4.0,5.0,6.0,7.0,8.0),
		 Vec(9.0,8.0,7.0,6.0,5.0,4.0,3.0,2.0),
		 Vec(10.0,11.0,12.0,13.0,14.0,15.0,16.0,17.0))
    val perm = Vec(0,0,0)

    Print("The product! ", Cmat)

    Print("The matrix ", Be, " and its tanspose ", trans(Be))

    val LU = ludcmp(Dmat, perm, 3)
    val indx = lupivot(Dmat, 3)
    Print("The LU ", LU, " and the pivots ", indx)
    val Xvec = lubksb(LU, indx, Bvec, 3)
    val Xmat = lubksbm(LU, indx, Bmat, 3, 8)
    Print("The X vector ", Xvec)
    Print("The X matrix ", Xmat)*/

    // Set up the boundaries.
    SetBoundary()

    MyLS.nonzeroes {
      nz =>
	for (v1 <- vertices(mesh)) {
	  if (isBoundary(v1) == 0.0) {
	    val r = MyLS.row(v1)
	    val c = MyLS.col(v1)
	    nz.b(r)
	    nz.x(c)
	    nz.A(r,c)
	    for (v2 <- vertices(v1)) {
	      if (isBoundary(v2) == 0.0) {
		nz.A(r, MyLS.col(v2))
	      }
	    }
	  }
	}
    }
    
    val A = MyLS.A()
    val x = MyLS.x()
    val b = MyLS.b()
    
    for (c <- cells(mesh)) {
      val numNodes = size(vertices(c))
      val Ke = LinearElementStiffness(c)
      val Fe = LinearElementForce(c, Ke)
      val Fb = LinearElementBoundary(c, Ke)

      val rows = MyLS.rows(c)
      val cols = MyLS.cols(c)
      A(rows,cols) += Ke
      b(rows) += Fe + Fb
    }

    MyLS.solve(A,x,b)
    for(v <- vertices(mesh)) {
      val col = MyLS.col(v)
      if(isBoundary(v) == 0.0)
	Print("ID ", ID(v), " value ", x(col))
    }
  }
}
