import Liszt.Language._
import Liszt.MetaInteger._

@lisztcode
object JoeWithModels {
	val rhs_rho = FieldWithConst[Cell,Float](0.f) ;
	val rhs_rhou = FieldWithConst[Cell,Vec[_3,Float]](Constants.float3_zero) ;
	val rhs_rhoE = FieldWithConst[Cell,Float](0.f) ;

	val face_rho  = FieldWithConst[Face,Float](0.f) ;
	val face_rhou = FieldWithConst[Face,Vec[_3,Float]](Constants.float3_zero) ;
	val face_rhoE = FieldWithConst[Face,Float](0.f) ;
	
	// notice that this would originally be called in JoeWithModels.initialHook()
	val rho = FieldWithConst[Cell,Float](IC.rho_init) ;
	val rhou = FieldWithConst[Cell,Vec[_3,Float]](IC.rho_init * IC.u_init) ;
	val rhoE = FieldWithConst[Cell,Float](IC.p_init / (IC.GAMMA - 1.f)  +  0.5f * IC.rho_init * dot( IC.u_init, IC.u_init )) ;

	val zone_interior = BoundarySet[Face]("zone_interior")
	val zone_boundary = BoundarySet[Face]("zone_boundary")
	val symmetry = BoundarySet[Face]("symmetry")
	val wall = BoundarySet[Face]("wall")
	val other_boundaries = BoundarySet[Face]("other_boundaries")
	val cbc = BoundarySet[Face]("cbc")
	val hook = BoundarySet[Face]("hook")
	val cbc_subsonic_outlet = BoundarySet[Face]("cbc_subsonic_outlet")
	
	def init() { Print("JoeWithModels()") ;}
	def initialHook() {Print("setInitialConditions()") ;}

	// TODO(Montse):: HACK!!!! NEED TO MODIFY THE REFERENCE OUTSIDE CELL!!!! MIGHT NOT BE ZERO!!!!
	def getInsideBoundaryCell( f : Face ) : Cell = {
		val c = if ( ID(inside(f)) == 0 ) outside(f) else inside(f) ;
		return c ;
	}	

	def run() {
		runExplicitBackwardEuler() ;
	}

	def runExplicitBackwardEuler() {

		val zeroFiveVec = Vec(0.f, 0.f, 0.f, 0.f, 0.f) ;
		val dummyA = Vec( zeroFiveVec, zeroFiveVec, zeroFiveVec, zeroFiveVec, zeroFiveVec) ;		
 
		setNavierStokesBC() ;
		
		var start_time = 0.0;
		var step = 0 ;
		while (step < Constants.nsteps ) {
			if (step == 1) {
				start_time = wall_time();
			}
			step += 1 ;
			calcRhs(dummyA) ; // note that this function originally would take rhs_rho, rhs_rhou, rhs_rhoE, rho, rhou, rhoE fields as arguments			
		}

		 Print( "TIME_FOR_LOOP: ", wall_time() - start_time ) ;
			
	}

	def calcRhs( A : Vec[_5, Vec[_5,Float]] ) {
		for (f <- zone_interior) { calcRhsInterior(f) }
		for (c <- cells(mesh)) {
			var trhs_rho  = 0.f  
			var trhs_rhou = Vec(0.f,0.f,0.f)
			var trhs_rhoE = 0.f
			for (f <- faces(c)) {
				var tfrhs_rho  = face_rho(f) 
				var tfrhs_rhou = face_rhou(f)
				var tfrhs_rhoE = face_rhoE(f)
				trhs_rho  += tfrhs_rho
				trhs_rhou += tfrhs_rhou
				trhs_rhoE += tfrhs_rhoE
				
			}
			rhs_rho(c)  = trhs_rho 
			rhs_rhou(c) = trhs_rhou
			rhs_rhoE(c) = trhs_rhoE
		}
	}

	def setNavierStokesBC() { }
	def showResidue( rhsResid: Vec[_5,Float], step: Int ) { }
	def CBCBoundary( f: Face, T_bc : Float, p_bc : Float, u_bc : Vec[_3,Float] ) { }
	def CBCSubsonicOutletBoundary( f: Face, p_bc : Float ) { }
	def SymmetryBoundary( f: Face ) { }
	def setRhoFa( f: Face ) { }

	def calcRhsInterior( f: Face ) {	
		val icv0 = outside(f) ;
		val icv1 = inside(f) ;

		val rho0 = rho(icv0) ;
		val u0 = UgpWithCvCompFlow.vel(icv0) ;
		val p0 = UgpWithCvCompFlow.press(icv0) ;
		val h0 = UgpWithCvCompFlow.enthalpy(icv0) ;

		val rho1 = rho(icv1) ;
		val u1 = UgpWithCvCompFlow.vel(icv1) ;
		val p1 = UgpWithCvCompFlow.press(icv1) ;
		val h1 = UgpWithCvCompFlow.enthalpy(icv1) ;

		// face unit normal and area...
		var nVec = MeshGeometryCalc.fa_normal(f) ;
		val area = sqrt(dot(nVec,nVec)) ;
		nVec = normalize(nVec) ;
	
		val kine0 = UgpWithCvCompFlow.kine(icv0) ;
		val kine1 = UgpWithCvCompFlow.kine(icv1) ;

		//Frho5[_0] = Frho, Frho5[_1] = Frhou.x , Frho5[_2] = Frhou.y, Frho5[_3] = Frhou.y, Frho5[_4] = FrhoE
		val Frho5 = UgpWithCvCompFlow.calcEulerFlux_HLLC(rho0, u0, p0, h0, UgpWithCvCompFlow.gamma(icv0), rho1, u1, p1, h1, UgpWithCvCompFlow.gamma(icv1), area, nVec, 0.f, kine0, kine1) ;
		
		face_rho(f)  = Frho5(_0) ;
		face_rhou(f) = Vec(Frho5(_1), Frho5(_2), Frho5(_3)) ;
		face_rhoE(f) = Frho5(_4) ;
		
		// rhs_rho(icv0) -=  Frho5(_0) ;
		// rhs_rhou(icv0) -= Vec(Frho5(_1), Frho5(_2), Frho5(_3)) ;
		// rhs_rhoE(icv0) -= Frho5(_4) ;
		// 
		// rhs_rho(icv1) += Frho5(_0) ;
		// rhs_rhou(icv1) += Vec(Frho5(_1), Frho5(_2), Frho5(_3)) ;
		// rhs_rhoE(icv1) += Frho5(_4) ;
	}

	def calcRhsSymmetry( f: Face) { }
	def calcRhsOtherBoundaries( f: Face ) { }

}
