import Liszt.Language._
import Liszt.MetaInteger._

@lisztcode
object UgpWithCvCompFlow {
	
	val vel = FieldWithConst[Cell,Vec[_3,Float]](Constants.float3_zero) ;
	val press = FieldWithConst[Cell,Float](0.f) ;
	val temp = FieldWithConst[Cell,Float](0.f) ;
	val enthalpy = FieldWithConst[Cell,Float](0.f) ;
	val LambdaOverCp = FieldWithConst[Cell,Float](0.f) ;
	val sos = FieldWithConst[Cell,Float](0.f) ;
	val kine = FieldWithConst[Cell,Float](0.f) ;
	val local_dt = FieldWithConst[Cell,Float](0.f) ;

	// these fields should be build in "initialHookScalarRansCombModel(), so we got rid of the whole function
	val RoM = FieldWithConst[Cell,Float](IC.R_gas) ;
	val gamma = FieldWithConst[Cell,Float](IC.GAMMA) ;
	val muLam = FieldWithConst[Cell,Float](0.f) ;
	
	// this should be called in runExplicitBackwardEuler
	val gam_fa = FieldWithConst[Face,Float](0.f) ;
	val RoM_fa = FieldWithConst[Face,Float](0.f) ;
	val h_fa = FieldWithConst[Face,Float](0.f) ;
	val T_fa = FieldWithConst[Face,Float](0.f) ;
	val vel_fa = FieldWithConst[Face,Vec[_3,Float]](Constants.float3_zero) ;
	val p_fa = FieldWithConst[Face,Float](0.f) ;
	val mu_fa = FieldWithConst[Face,Float](0.f) ;
	val lamOcp_fa = FieldWithConst[Face,Float](0.f) ;
	val rho_fa = FieldWithConst[Face,Float](0.f) ;

	def calcRansStateVarAndMaterialProperties() { } 

	def ComputeBCProperties_T( f: Face ) {
		gam_fa(f) = IC.GAMMA ;
		RoM_fa(f) = IC.R_gas ;
		h_fa(f) = IC.GAMMA * IC.R_gas / ( IC.GAMMA - 1.f ) * T_fa(f) ;
	}

	def ComputeGeometry(){ }
	def init() { }

	def calcEulerFlux_HLLC( rhoL : Float, uL : Vec[_3,Float], pL : Float, h0 : Float, gammaL : Float, rhoR : Float, uR : Vec[_3,Float], pR : Float, h1 : Float, gammaR : Float, area : Float, nVec : Vec[_3,Float], surfVeloc : Float, kL : Float, kR : Float ) : Vec[_5,Float] = {
		
		var Frho = 0.f ;
		var Frhou = Constants.float3_zero ;
		var FrhoE = 0.f ;		
		
		val unL = dot(uL,nVec) ;
		val uLuL = dot(uL,uL) ;
		val cL = sqrt( gammaL * pL / rhoL ) ;
		val hL = gammaL / ( gammaL - 1.f ) * pL / rhoL + 0.5f * uLuL + kL ;
		val eL = hL * rhoL - pL ;

		val unR = dot(uR,nVec) ;
		val uRuR = dot(uR,uR) ;
		val cR = sqrt( gammaR * pR / rhoR ) ;
		val hR = gammaR / ( gammaR - 1.f ) * pR / rhoR + 0.5f * uRuR + kR ;
		val eR = hR * rhoR - pR ;

		// Roe's averaging
		val Rrho = sqrt( rhoR / rhoL ) ;
		val tmp = 1.f / ( 1.f + Rrho ) ;
		val velRoe = tmp * ( uL + uR *Rrho ) ;
		val uRoe = dot( velRoe, nVec ) ;
		//val hRoe = tmp * ( hL + hR * Rrho ) ;
		
		//val cRoe = sqrt( (gammaL - 1.f) * ( hRoe - 0.5f * dot( velRoe, velRoe ) ) ) ;
		val gamPdivRho = tmp * ( (gammaL * pL / rhoL + 0.5f * (gammaL - 1.f) * uLuL ) + (gammaR * pR / rhoR + 0.5f * (gammaR - 1.f) * uRuR ) * Rrho ) ;
		val cRoe = sqrt( gamPdivRho - ((gammaL + gammaR) * 0.5f - 1.f ) * 0.5f * dot( velRoe, velRoe ) );

		// speed of sound at L and R 
		val sL = (uRoe - cRoe).min(unL - cL) ;
		val sR = (uRoe + cRoe).max(unR + cR) ;

		// speed of contact surface
		val sM = (pL - pR - rhoL * unL * (sL - unL) + rhoR * unR * (sR - unR) ) / (rhoR * (sR - unR) - rhoL * (sL - unL)) ;

		// pressure at right and left (pR =pL ) side contact surfaca
		val pStar = rhoR * (unR - sR) * (unR - sM) + pR ;


/*
		if (sM >= 0.f) {
			if (sL > 0.f) {
				Frho = rhoL * unL ;
				Frhou = rhoL * unL * uL + pL * nVec ;
				FrhoE = (eL + pL) * unL ;
			} else {
				val invSLmSs = 1.f / (sL - sM ) ;
				val sLmuL = sL - unL ;
				val rhoSL = rhoL * sLmuL * invSLmSs ;
				val rhouSL = (( rhoL * sLmuL) * uL + (pStar - pL) * nVec) * invSLmSs ;
				val eSL = ( sLmuL * eL - pL * unL + pStar * sM ) * invSLmSs ;

				Frho = rhoSL * sM ;
				Frhou = rhouSL * sM + pStar * nVec ;
				FrhoE = ( eSL + pStar ) * sM ; 
			}
		} else {
			if ( sR >= 0.f ) {
				val invSRmSs = 1.f / (sR - sM ) ;
				val sRmuR = sR - unR ;
				val rhoSR = rhoR * sRmuR * invSRmSs ;
				val rhouSR = (( rhoR * sRmuR) * uR + (pStar - pR) * nVec) * invSRmSs ;
				val eSR = ( sRmuR * eR - pR * unR + pStar * sM ) * invSRmSs ;

				Frho = rhoSR * sM ;
				Frhou = rhouSR * sM + pStar * nVec ;
				FrhoE = ( eSR + pStar ) * sM ;
			} else {
				Frho = rhoR * unR ;
				Frhou = rhoR * unR * uR + pR * nVec ;
				FrhoE = (eR + pR) * unR ;
			}
		}
*/


                // Calculate the nested code first.

                // if (sM >= 0.f && sL <= 0.f) || (sM < 0.f && sR >= 0.f)
                val sL_gt_0 = sL > 0.f;
                val sR_ge_0 = sR >= 0.f;
                val sM_ge_0 = sM >= 0.f;

                val top_nest = (sM_ge_0 & !sL_gt_0);
                val btm_nest = (!sM_ge_0 & sR_ge_0);
                val either_nest = (top_nest | btm_nest);

                val unSide = if (sM_ge_0) unL else unR;
                val pSide = if (sM_ge_0) pL else pR;
                val eSide = if (sM_ge_0) eL else eR;
                val rhoSide = if (sM_ge_0) rhoL else rhoR;
                val uSide = if (sM_ge_0) uL else uR;

                if (either_nest) {
                  val sSide = if (sM_ge_0) sL else sR;

                  val inv_diff = 1.f / (sSide - sM);
                  val sSide_minus_other = sSide - unSide;
                  val scaled_rho = rhoSide * sSide_minus_other * inv_diff;
                  val scaled_more = (rhoSide * sSide_minus_other * uSide + (pStar - pSide) * nVec) * inv_diff;
                  val eS_side = (sSide_minus_other * eSide - pSide * unSide + pStar * sM) * inv_diff;

                  Frho = scaled_rho * sM;
                  Frhou = scaled_more * sM + pStar * nVec;
                  FrhoE = (eS_side + pStar) * sM;
                } else {

                  Frho = rhoSide * unSide;
                  Frhou = rhoSide * unSide * uSide + pSide * nVec;
                  FrhoE = (eSide + pSide) * unSide;
                }

		Frho *= area ;
		Frhou *= area ;
		FrhoE *= area ;
		
		return Vec(Frho, Frhou.x, Frhou.y, Frhou.z, FrhoE) ;		
	}

	def calcDt( cfl_target : Float ) : Float = { 0.1f; }

}


