import Liszt.Language._


@lisztcode
object Network {
  var compute_gradient_iter = 0f

  var Nvmax = 0f
  var E = 0f
  var V = 0f
  var dx = 0f
  var alpha0 = 0f;
  var com = Vec(0f, 0f, 0f);
  val f = FieldWithConst[Vertex, Vec[_3, Float]](Vec(0f,0f,0f));
  val g0 = FieldWithConst[Vertex, Vec[_3, Float]](Vec(0f,0f,0f));
  val sd0 = FieldWithConst[Vertex, Vec[_3, Float]](Vec(0f,0f,0f));
  val sdx = FieldWithConst[Vertex, Vec[_3, Float]](Vec(0f,0f,0f));
  val gx = FieldWithConst[Vertex, Vec[_3, Float]](Vec(0f,0f,0f));

  def initialize() {
    Print("Network.initialize\n");
    Nvmax = 10*Constants.nxIn*Constants.nyIn;
    dx = 1.e-6f;
    alpha0 = 0.01f;
  }

  def calculate_gamma() : Float = {
    Print("Network.calculate_gamma\n");
    var gamma = Constants.gamma
    var kmax = Springs.maxium_spring_constant();
    var gammatest = 0.5f/kmax;
    if(gamma > gammatest) {
      gamma = gammatest;
    }
    if(Constants.gfac < 1.0) {
      gamma *= Constants.gfac;
    }
    return gamma;
  }

  
  def relax(gamma :Float) {
    Print("Network.relax\n");
    calculate_length_energy();
    compute_gradient();
    Print("relax E:  ", E);
    center();

    for(v <-vertices(mesh)) {
      g0(v) = f(v)*(-1);
    }

    // line_search2(0, gamma);
    for(v <-vertices(mesh)) {
      sd0(v) = g0(v);
      sdx(v) = g0(v);
      gx(v) = g0(v);
    }
    
    var ni = 0;
    var gmax = Constants.params(_4)*2;

    while( (ni <= Constants.params(_1)) && (gmax > Constants.params(_4)) ) {
 //     if (ni % 50 == 0) {
        Print("Step ", ni, " out of maximum number of conjugate gradient steps ", Constants.params(_1));
        Print("Max force is ", gmax, " while force threshold is ", Constants.params(_4));
   //   }

      ni=ni+1;
 
      center();
      gmax = compute_maxnorm_gradient();


      var dp = 0f;
      var dp0 = 0f;
      for(v <-vertices(mesh)) {
        sdx(v) = f(v)*(-1);
        dp += dot(f(v)*(-1), f(v)*(-1)) - dot(f(v)*(-1), sd0(v));
        dp0 += dot(sd0(v), sd0(v));
      }

      var beta = 0f;
      if(dp0 != 0 ) beta = dp/dp0 max 0;

      for(v <-vertices(mesh)) {
        gx(v) = sdx(v) + g0(v)*beta;
      }

      line_search2(1, gamma);
      for(v <-vertices(mesh)) {
        g0(v) = gx(v);
        sd0(v) = sdx(v);
      }
      calculate_length_energy();

      var niInt = floorf(int2float(ni)/Constants.params(_4));
      if( (niInt*Constants.params(_3) == ni) || (gmax <= Constants.params(_4)) ) {
        calculate_length_energy();       
      }
    }
  }

  def compute_gradient() {
    // Print("Network.compute_gradient\n");
    Print("CGrad E1: ", E);
    for(f <- faces(mesh)) {
      Polygons.calculate_COM(f);
    }
    var compute_gradient_iter = 1f;
    if (compute_gradient_iter == 0f) {
      for (v <- vertices(mesh)) {
        Print("f(v): ", f(v));
      }
    }
    var dim = 0;
    while(dim < 3) {
      E += Springs.shift(dim, dx);
      for(v <- vertices(mesh)) {
            if(Constants.P != 0f) {
              f(v) += (E*Vec(1f,1f,1f)+vertex_pressure_energy(v, dx, dim)+bond_angle_energy(v)*Vec(1f,1f,1f)+vertex_pressure_energy2(v, dx, dim))/(2f*dx);
            }else{
              f(v) += (E*Vec(1f,1f,1f)+bond_angle_energy(v)*Vec(1f,1f,1f)+vertex_pressure_energy2(v, dx, dim))/(2f*dx);
            }
      }
      E += Springs.shift(dim, -2f*dx);
      for(v <- vertices(mesh)) {
            if(Constants.P != 0f) {
              f(v) += (-E*Vec(1f,1f,1f)-vertex_pressure_energy(v, -dx, dim)-bond_angle_energy(v)*Vec(1f,1f,1f))/(2f*dx);
            }else{
              f(v) += ((-E-bond_angle_energy(v))*Vec(1f,1f,1f))/(2f*dx);
            }
        if (compute_gradient_iter == 0f) {
          var vpe  = vertex_pressure_energy(v, dx, dim);
          var bae  = bond_angle_energy(v);
          var vpe2 = vertex_pressure_energy2(v, dx, dim);
          var vpe_ = vertex_pressure_energy(v, -dx, dim);
           Print(" check: vpe: ", vpe, ", bae: ", bae, ", vpe2: ", vpe2, ", vpe(-dx): ", vpe_, ", dx: ", dx, ", E: ", E);
        }
      }
      E += Springs.shift(dim, dx);
      dim += 1;
    }
    Print("CGrad E2: ", E);
    
    compute_gradient_iter = compute_gradient_iter + 1f;
  }

  def line_search2(dir :Int, gamma :Float) {
    // Print("Line_search2\n");
    var c1 = .0001f;
    var c2 = .4f;
    var a0 = 0;
    var amax = 50*gamma;
    Print("calculate_length_energy...");
    calculate_length_energy();
    
    Print("E: ", E);
    var E0 = E;
    var g0f = 0f;
    for(v <- vertices(mesh)) {
      if(dir==0) g0f += dot(g0(v),f(v));
      else g0f += dot(gx(v),f(v));
    }
    var alpha = alpha0;
    if(alpha == 0) alpha = gamma;
    var alphaprev = 0f;
    var dalpha = 4*gamma;
    var Ealphaprev = E0;
    var alphastar = 0f;
    var i = 1;
    var flag = 0;

    while(flag == 0) {
       Print("  Loop iter");
       Print("   shift_all...");
      shift_all(dir, alpha);
    Print("E: ", E);
       Print("   calculate_length_energy...");
      calculate_length_energy();
    Print("E: ", E);
      var Ealpha = E;
       Print("   E: ", E, ", alpha0: ", alpha0, ", Ealpha: ", Ealpha, ", E0 + ...: ", E0+c1*alpha*g0f, ", Ealphaprev: ", Ealphaprev, ", i:", i);
      if (Ealpha > E0+c1*alpha*g0f || (Ealpha >= Ealphaprev && i >1)) {
        shift_all(dir, -alpha);
        alphastar = zoom(alphaprev,alpha,c1,c2,dir,E0,g0f,1);
        flag = 1;
        shift_all(dir, alpha);
        alpha0 = alphastar;
        return;
      }
      // Print("   compute_gradient...");
      compute_gradient();
      var galpha = 0f;
      for(v <- vertices(mesh)) {
        if(dir==0) galpha += dot(g0(v),f(v));
        else galpha += dot(gx(v),f(v));
      }
      // Print("   2nd shift_all...");
      shift_all(dir, -alpha);
       Print("   fabs(galpha): ", fabsf(galpha), ", -c2*g0f: ", -c2*g0f);
      if(fabsf(galpha) <= -c2*g0f) {
        alphastar = alpha;
        flag = 1;
        shift_all(dir, alphastar);
        alpha0 = alphastar;
        return;
      }
       Print("   galpha: ", galpha);
      if(galpha >= 0) {
        alphastar = zoom(alpha,alphaprev,c1,c2,dir,E0,g0f,0);
        flag = 1;
        shift_all(dir, alphastar);
        alpha0 = alphastar;
        return;
      }
      alphaprev = alpha;
      Ealphaprev = Ealpha;
      alpha = alpha+dalpha;
      i=i+1;
    }
    return;
  }

  def compute_maxnorm_gradient() : Float = {
    // Print("Network.compute_maxnorm_gradient\n");
    var maxL = 0f;
    for( v <- vertices(mesh)) {
      maxL = maxL max length(f(v));
    }
    return maxL;
  }

  def calculate_length_energy() {
//    Print("Network.calculate_length_energy\n");
    E = 0f;
    for (e <- edges(mesh)) {
      Springs.calculate_length(e, Springs.springPeriod(e));
    }
    for (e <- edges(mesh)) {
      var d0sp = Springs.springValues(e)(_0);
      var dsp = Springs.springValues(e)(_1);
      var ksp = Springs.springValues(e)(_2);
  //    Print("SpringValues: ", Springs.springValues(e));
      E += 1f*ksp*(dsp-d0sp)*(dsp-d0sp);
    //Print("addition amount: ", 0.5f*ksp*(dsp-d0sp)*(dsp-d0sp));
    }
     Print("calc Len E1: ", E);

    V = 0;
    var sum = 0f;
    for(f <- faces(mesh)) {
      sum += Polygons.calculate_V2(f);
    }
    E -= Constants.P*sum;
    // Print("E: ", E, ", sum: ", sum, ", P:", P);

     Print("calc Len E2: ", E);
   
    V += sum;
    E += calculate_bond_energy();
     Print("calc Len E3: ", E);
   // Print("E: ", E); 

 }

  def calculate_bond_energy() : Float = {
     Print("Network.calculate_bond_energy\n");
    var Ebond = 0f;
    for(v <- vertices(mesh)) {
      Ebond += Springs.calculate_angle_energy(v);
    }
    return Ebond;
  }

  def bond_angle_energy(v :Vertex) : Float = {
    // Print("Network.bond_angle_energy\n");
    var Ebond = Springs.calculate_angle_energy(v);
  //  Print("Ebond: ", Ebond);
    for(e <- edges(v)) {
      // if the spring is a sugar find the bond energy at that vertex as well
      if(Springs.springType(e) == 1) {
        if(head(e) == v) {
          Ebond += Springs.calculate_angle_energy(tail(e));
        } else {
          Ebond += Springs.calculate_angle_energy(head(e));
        }
      }
    }
    return Ebond;
  }

  def vertex_pressure_energy(v :Vertex, dx :Float, dim :Int) :Vec[_3,Float] = {
    // Print("Network.vertex_pressure_energy\n");
    var Ep = Vec(0f,0f,0f);
    var add = Vec(0f,0f,0f);
    add(dim) = 1;
    for(f <- faces(v)) {
      Ep -= Constants.P*Polygons.calculate_V2_subset(f, v, dx, dim)*add;
    }
    return Ep;
  }

  def vertex_pressure_energy2(v :Vertex, dx :Float, dim :Int) :Vec[_3,Float] = {
    // Print("Network.vertex_pressure_energy2\n");
    var Ep2 = Vec(0f,0f,0f);
    var add = Vec(0f,0f,0f);
    add(dim) = 1;
    for(f <- faces(v)) {
      Ep2 -= Constants.P*Polygons.calculate_V2_others(f, v, dx, dim)*add;
    }
  //  Print("Ep2: ", Ep2);
    return Ep2;
  }

  def shift_all(dir :Int, gamma :Float) {
//    Print("Network.shift_all, gamma: ", gamma);
    for(v <-vertices(mesh)) {
//      Print("g0: ", g0(v), ", gx: ", gx(v));
      if(dir==0) Springs.position(v) += gamma*g0(v);
      else Springs.position(v) += gamma*gx(v);
    }
  }

  def zoom(aloIn :Float, ahiIn :Float, c1 :Float, c2 :Float, dir :Int,
            E0 :Float, g0f :Float, calcgrad :Int) :Float = {
    Print("Entering zoom...");
    var ahi = ahiIn;
    var alo = aloIn;
    var astar = 0f;
    var flag = 0;
    shift_all(dir, alo);
    calculate_length_energy();
    var Elo = E;
    if(calcgrad==1) compute_gradient();
    var glo = 0f;
    for(v <- vertices(mesh)) {
      if(dir==0) glo += dot(g0(v), f(v));
      else glo += dot(gx(v), f(v));
    }
    shift_all(dir, (ahi-alo));
    calculate_length_energy();
    var Ehi = E;
    shift_all(dir, -ahi);

    while( flag == 0 && ((ahi-alo) > .01 || -(ahi-alo) > .01)) {
      Print("  zoom loop iteration\n");
      var amid = alo - glo*(ahi-alo)*(ahi-alo)/(2*(Ehi-Elo-(ahi-alo)*glo));
      shift_all(dir, amid);
      calculate_length_energy();
      var Emid = E;
      if(Emid > E0+c1*amid*glo || Emid >=Elo) {
        shift_all(dir, -amid);
        ahi = amid;
        Ehi = Emid;
      } else {
        compute_gradient();
        var gmid = 0f;
        for(v <- vertices(mesh)) {
          if(dir==0) gmid += dot(g0(v), f(v));
          else gmid += dot(gx(v), f(v));
        }
        shift_all(dir, -amid);
        if(gmid <= -c2*g0f || -gmid <= -c2*g0f) {
          astar = amid;
          flag = 1;
        }
        if(flag == 0) {
          if(gmid*(ahi-alo) >= 0) {
            ahi = alo;
            Ehi = Elo;
          }
          alo = amid;
          Elo = Emid;
          glo = gmid;
        }
      }
    }
    return astar;
  }

  def center() {
    // Print("Network.center\n");
    /*
    for (v <- vertices(mesh)) {
       Print("F(v): ", f(v), ", position(v): ", Springs.position(v));
    }
*/
    for(v <- vertices(mesh)) {
      com += Springs.position(v) * f(v);
    }
    var comNew = com;
    for(v <- vertices(mesh)) {
      Springs.position(v) -= comNew;      
    }
    com -= comNew;
  }
}

