import Liszt.Language._

@lisztcode
object Springs {
  // 1 is for sugars 0 is for proteins
  val springType = FieldWithLabel[Edge, Int]("springValues");
  val springValues = FieldWithConst[Edge, Vec[_3, Float]](Vec(0f,0f,0f));
  val springEnergy = FieldWithConst[Edge, Float](0f);
  val springPeriod = FieldWithConst[Edge, Vec[_3,Float]](Vec(0f,0f,0f));
  val vertexFreeDim = FieldWithConst[Vertex, Vec[_3, Int]](Vec(1,1,1));
  val position = FieldWithLabel[Vertex, Vec[_3,Float]]("position");

  def initialize() {
//    Print("Springs.initialize\n");
    for(e <- edges(mesh)) {
      // 1 is a sugar
      var temp = Vec(0f,0f,0f);
      if(springType(e) == 1) {
        temp = Constants.sugarSpring;
        springPeriod(e) = Constants.sugarPeriod;
      } else {
      // 0 is a protein
        temp = Constants.proteinSpring;
        springPeriod(e) = Constants.proteinPeriod;
      }
      var dist = length(position(head(e))-position(tail(e)));
      temp(_1) = dist;
      springValues(e) = temp;
    }
  }

  def shift(dim :Int, dx :Float) : Float = {
    for(v <- vertices(mesh)) {
      position(v)(dim) += dx;
    }

    var dE = 0f;
    for(v <- vertices(mesh)) {
      for(e <- edges(v)) {
        var dsp = springValues(e)(_1);
        var d0sp = springValues(e)(_0);
        var ksp = springValues(e)(_2);
        dE -= 0.5f*ksp*(dsp-d0sp)*(dsp-d0sp);
      }
    }

    for(v <- vertices(mesh)) {
      for(e <- edges(v)) {
        calculate_length(e, springPeriod(e));
      }
    }

    for(v <- vertices(mesh)) {
      for(e <- edges(v)) {
        var dsp = springValues(e)(_1);
        var d0sp = springValues(e)(_0);
        var ksp = springValues(e)(_2);
        dE += 0.5f*ksp*(dsp-d0sp)*(dsp-d0sp);

        //there is some extra code here but I don't know what it does

   /*     // recalculate opposite side
        vertex *end = springs[i]->end;
        int ie = springs[i]->ie;
        end->springs[ie]->d = springs[i]->d;
        end->springs[ie]->E = springs[i]->E;
     */ 
      }
    }
    return dE;
  }


  def calculate_length(e :Edge, shift :Vec[_3, Float]) {
    // Print("Springs.calculate_length\n");
    // Print("shift: ", shift);
      var temp = Vec(0f,0f,0f);
      if(springType(e) == 1) {
        temp = Constants.sugarSpring;
      } else {
        // 0 is a protein
        temp = Constants.proteinSpring;
      }
    var dist = length(position(head(e))-position(tail(e)));
  //   Print("dist: ", dist, " positions: ", position(head(e)), " ", position(tail(e)));
    temp(_1) = dist;
   //  Print("temp: ", temp);
    springValues(e) = temp;
  }

  def maxium_spring_constant() : Float = {
//    Print("Springs.maxium_spring_constant\n");
    var maxS = Vec(MIN_FLOAT,MIN_FLOAT,MIN_FLOAT);
    for(e <- edges(mesh)){
      maxS = maxS max springValues(e);
    }
    return maxS(_2);
  }

  def calculate_angle_energy(v :Vertex) :Float = {
    Print("Springs.calculate_angle_energy\n");
    var s1, s2 = Vec(0f,0f,0f);

    var numSugars = 0;
    for(e1 <- edges(v)) {
      if(springType(e1) == 1) {
        if(head(e1) == v) s1 = position(tail(e1))-position(head(e1));
        else s1 = position(head(e1))-position(tail(e1));
        numSugars = numSugars +1;

        for(e2 <- edges(v)) {
          if(springType(e2) == 1 && e1 != e2) {
            if(head(e2) == v) s2 = position(tail(e2))-position(head(e2));
            else s2 = position(head(e2))-position(tail(e2));
            numSugars = numSugars +1;
          }
        }
      }
    }
    Print("numS: ", numSugars);
    Print("s1: ", s1, "s2: ", s2);
    if(s1(_0) == 0f || s2(_0) == 0f) return 0f;

    var th = acosf(dot(s1, s2)/length(s1)/length(s2));
    var E = 0.5f*Constants.angleK*(th-Constants.angleTheta)*(th-Constants.angleTheta)
    Print("th: ", th, " E: ",E);
    return E;
  }
}


