#ifndef _PARALLEL_LOAD_UTIL_H
#define _PARALLEL_LOAD_UTIL_H

#include<stdio.h>
#include<stdlib.h>

#define LOG_PRINT(format, ...) do { \
	int rank,size; \
	MPI_Comm_rank(MPI_COMM_WORLD,&rank); \
	MPI_Comm_size(MPI_COMM_WORLD,&size); \
	fprintf(stderr,"[%d/%d] %s:%d: log: " format "\n", rank,size, __FILE__, __LINE__, ## __VA_ARGS__); \
} while(0)

#define LOG_LOG(format, ...) LOG_PRINT(format, ## __VA_ARGS__)

#if 0
#define LOG_DBG(format, ...) LOG_PRINT(format, ## __VA_ARGS__) 
#else
#define LOG_DBG(format, ...)
#endif

#define LOG_WARN(format, ...) LOG_PRINT(format, ## __VA_ARGS__)

#define LOG_ERROR(format, ...) do { LOG_PRINT(format, ## __VA_ARGS__); abort(); } while(0)


//MPI Utility functions
inline int MPIU_GetDispFromSize(const int * send_sizes, size_t n, int * send_disp) {
    send_disp[0] = 0;
    for(size_t i = 1; i < n; i++)
        send_disp[i] = send_sizes[i-1] + send_disp[i-1];
    return send_disp[n-1] + send_sizes[n-1];
}

typedef unsigned char byte;

#endif