/*
 * FieldInfo.h
 *
 *  Created on: Jul 27, 2009
 *      Author: zdevito
 */

#ifndef FIELDINFO_H_
#define FIELDINFO_H_
#include "Fields/CommInfo.h"
#include "Liszt/ElemTypes.h"
#include "liszt_memory.h"
#include <ext/hash_set>
#include <iostream>
#include <mpi.h>

// TODO(mbarrien): Figure out a way around friending an MPI-specific class
namespace LisztMPIPrivate {
    class FieldInit;
}
using LisztPrivate::ElemTypes::id_type;

namespace FieldCommon {
class FieldInfo {
public:
    FieldInfo() {}
    typedef __gnu_cxx::hash_set<id_type> GhostSet;
    void addGhostWrite(id_type id) {
        ghost_writes.insert(id);
    }
    void addGhostRead(id_type id) {
        ghost_reads.insert(id);
    }
    bool readsGhosts() const {
        return ghost_reads.size() > 0;
    }
    bool writesGhosts() const {
        return ghost_writes.size() > 0;
    }
    ::std::tr1::shared_ptr<CommPair> getReaderComm() {
        return registered_readers;
    }
    ::std::tr1::shared_ptr<CommPair> getWriterComm() {
        return registered_writers;
    }
    void print(const char * name, bool has_outside) {
        using namespace std;
        cout << "Registered Readers: " << endl;
        registered_readers->recv.print();
        cout << "Registered Writers: " << endl;
        registered_writers->recv.print();
        cout << endl << endl;
        printRow(name,registered_readers->recv,true,has_outside);
        printRow(name,registered_writers->recv,false,has_outside);
    }
private:
    void printRow(const char * name, const CommInfo & info, bool is_reader, bool has_outside) {
        int size;
        int from;
        using namespace std;
        MPI_Comm_size(MPI_COMM_WORLD,&size);
        MPI_Comm_rank(MPI_COMM_WORLD,&from);
        for(size_t i = 0; i < info.messages.size(); i++) {
          size_t to = info.messages[i].rank;
          cout << "COMM_INFO_ROW" <<"\t" << name << "\t" << size << "\t" << from << "\t" << to << "\t" << info.messages[i].size << "\t" << is_reader << "\t" << has_outside << endl;
        }
    }
    friend class LisztMPIPrivate::FieldInit;
    GhostSet ghost_reads;
    GhostSet ghost_writes;
    ::std::tr1::shared_ptr<CommPair> registered_readers;
    ::std::tr1::shared_ptr<CommPair> registered_writers;
};
}  // namespace FieldCommon
#endif /* FIELDINFO_H_ */
