/*
 * FieldValueReference.h
 *
 *  Created on: Aug 18, 2009
 *      Author: zdevito
 */

#ifndef FIELDVALUEREFERENCE_H_
#define FIELDVALUEREFERENCE_H_
#include "Fields/ReductionOps.h"
#include "Liszt/ElemTypes.h"
namespace FieldCommon {

//building block for Field implementations that provides
//a level of indirection for updating field values

//this will need to be modified to support OpenMP
//with some locking protocol
using ::LisztPrivate::id_type;
template<typename Field>
class FieldValueReference {
public:
    typedef typename Field::elem_type elem_type;
    typedef typename Field::value_type value_type;

    typedef Field field_type;
    typedef FieldValueReference<Field> this_type;

    FieldValueReference(field_type & f, id_type i)
    : field(f), idx(i) {}
#define FORWARD_OP(op) \
    template<typename T> \
    this_type & operator op(const T & rhs) { \
        *(field.getWriteEntry(idx)) op rhs; \
        return *this;\
    } \
    this_type & operator op(const this_type & r) { \
        return *this op (const value_type &) r; \
    }
    FORWARD_OP(=)
    FORWARD_OP(+=)
    FORWARD_OP(-=)
    FORWARD_OP(*=)
    FORWARD_OP(/=)
    FORWARD_OP(&=)
    FORWARD_OP(|=)
    FORWARD_OP(^=)

#undef FORWARD_OP
    this_type & minWith(const value_type & rhs) {
        value_type * e = field.getWriteEntry(idx);
        *e = ReductionImpl<value_type>::min(*e,rhs);
        return *this;
    }
    this_type & minWith(const this_type & r) {
        return minWith( (const value_type &) r);
    }

    this_type & maxWith(const value_type & rhs) {
        value_type * e = field.getWriteEntry(idx);
        *e = ReductionImpl<value_type>::max(*e,rhs);
        return *this;
    }
    this_type & maxWith(const this_type & r) {
        return maxWith( (const value_type &) r);
    }

    this_type & operator++(int) {
        (*field.getWriteEntry(idx))++;
        return *this;
    }
    this_type & operator--(int) {
        (*field.getWriteEntry(idx))--;
        return *this;
    }
    this_type & operator++() {
        ++(*field.getWriteEntry(idx));
        return *this;
    }
    this_type & operator--() {
        --(*field.getWriteEntry(idx));
        return *this;
    }
    const value_type operator-() const {
        return - (const value_type &) (*this);
    }
private:
    field_type & field;
    id_type idx;
};

}  // namespace FieldCommon

#endif /* FIELDVALUEREFERENCE_H_ */
