/*
 * ReductionOps.h
 *
 *  Created on: Aug 3, 2009
 *      Author: zdevito
 */

#ifndef REDUCTIONOPS_H_
#define REDUCTIONOPS_H_
#include "Fields/Identities.h"
#include <algorithm>
#include "Fields/ReductionImpl.h"

#define LISZT_REDUCTION_OPS(_) \
	_(LISZT_MAX)  \
	_(LISZT_MIN)  \
	_(LISZT_PROD) \
	_(LISZT_SUM)  \
	_(LISZT_LAND) \
	_(LISZT_LOR)  \
	_(LISZT_LXOR) \
	_(LISZT_BAND) \
	_(LISZT_BOR) \
	_(LISZT_BXOR) \
	_(LISZT_ASSIGN) \
	_(LISZT_READ)

#define AS_TYPE(t) t##_TYPE,
enum LisztOpType {
	LISZT_REDUCTION_OPS(AS_TYPE)
	LISZT_OP_SIZE
};
#undef AS_TYPE

#define AS_STRING(t) #t,

inline static const char * ReductionOpToString(LisztOpType typ) {
static const char * ops[] = {
LISZT_REDUCTION_OPS(AS_STRING)
NULL };
	return ops[typ];
}

#undef AS_STRING
/*
template<typename T>
struct ComparisonFunctions {
    static const T & min(const T & a, const T & b) {
        return std::min(a,b);
    }
    static const T & max(const T & a, const T & b) {
        return std::max(a,b);
    }
};

template<size_t N, typename T>
struct ComparisonFunctions< vec<N,T> > {
    static const vec<N,T>  min(const vec<N,T> & a, const vec<N,T> & b) {
        vec<N,T> ret(Identities<T>::min_identity());
        for(size_t i = 0; i < N; i++)
            ret[i] = ComparisonFunctions<T>::min(a[i],b[i]);
        return ret;
    }
    static const vec<N,T>  max(const vec<N,T> & a, const vec<N,T> & b) {
        vec<N,T> ret(Identities<T>::max_identity());
        for(size_t i = 0; i < N; i++)
            ret[i] = ComparisonFunctions<T>::max(a[i],b[i]);
        return ret;
    }
};

template<size_t R, size_t C, typename T>
struct ComparisonFunctions< matrix<R,C,T> > {
    static const matrix<R,C,T>  min(const matrix<R,C,T> & a, const matrix<R,C,T> & b) {
        matrix<R,C,T> ret(Identities<T>::min_identity());
        for(size_t i = 0; i < R; i++)
            for(size_t j = 0; j < C; j++)
                ret.v[i][j] = ComparisonFunctions<T>::min(a.v[i][j],b.v[i][j]);
        return ret;
    }
    static const matrix<R,C,T>  max(const matrix<R,C,T> & a, const matrix<R,C,T> & b) {
        matrix<R,C,T> ret(Identities<T>::max_identity());
        for(size_t i = 0; i < R; i++)
            for(size_t j = 0; j < C; j++)
                ret.v[i][j] = ComparisonFunctions<T>::max(a.v[i][j],b.v[i][j]);
        return ret;
    }
}; */

struct LISZT_SUM {
    template<typename T>
    static T identity() {
        return ReductionImpl<T>::additive();
    }
    template<typename T>
    static void apply(const T & d,T * v) {
         *v = ReductionImpl<T>::plus(*v,d);
    }
    const static LisztOpType type = LISZT_SUM_TYPE;
};

struct LISZT_SUB {
    template<typename T>
    static T identity() {
        return ReductionImpl<T>::additive();
    }
    template<typename T>
    static void apply(const T & d,T * v) {
         *v -= d;
    }
    const static LisztOpType type = LISZT_SUM_TYPE;
};

struct LISZT_ASSIGN {
    template<typename T>
    static T identity() {
        //assign's identity is really "don't care", which is what we get when we leave of the return statement here
        //this is going to result in a warning, so we should probably figure out a better solution for this, however
        //there is no guarentee that a type will have any particular identity defined
        return T();
    }
    template<typename T>
    static void apply(const T & d,T * v) {
         *v = d;
    }
    const static LisztOpType type = LISZT_ASSIGN_TYPE;
};

struct LISZT_MAX {
    template<typename T>
    static T identity() {
        return ReductionImpl<T>::max_identity();
    }
    template<typename T>
    static void apply(const T & d,T * v) {
        *v = ReductionImpl<T>::max(d,*v);
    }
    const static LisztOpType type = LISZT_MAX_TYPE;
};

struct LISZT_MIN {
    template<typename T>
    static T identity() {
        return ReductionImpl<T>::min_identity();
    }
    template<typename T>
    static void apply(const T & d,T * v) {
        *v = ReductionImpl<T>::min(d,*v);
    }
    const static LisztOpType type = LISZT_MIN_TYPE;
};

struct LISZT_PROD {
    template<typename T>
    static T identity() {
        return ReductionImpl<T>::multiplicative();
    }
    template<typename T>
    static void apply(const T & d,T * v) {
        //*v = d * (*v); // used to be *v *= d ; has to be changes because matrix does not implement *= operator
        *v = ReductionImpl<T>::times(*v,d);
    }

    const static LisztOpType type = LISZT_PROD_TYPE;
};
struct LISZT_DIV {
    template<typename T>
    static T identity() {
        return ReductionImpl<T>::multiplicative();
    }
    template<typename T>
    static void apply(const T & d,T * v) {
        //*v = d * (*v); // used to be *v *= d ; has to be changes because matrix does not implement *= operator
        *v /= d;
    }
    const static LisztOpType type = LISZT_PROD_TYPE;
};


struct LISZT_LAND {
    template<typename T>
    static T identity() {
        return ReductionImpl<T>::land_identity();
    }
    template<typename T>
    static void apply(const T & d,T * v) {
        *v = ReductionImpl<T>::land(*v,d);
    }
    const static LisztOpType type = LISZT_LAND_TYPE;
};

struct LISZT_LOR {
    template<typename T>
    static T identity() {
        return ReductionImpl<T>::lor_identity();
    }
    template<typename T>
    static void apply(const T & d,T * v) {
        *v = ReductionImpl<T>::lor(*v, d);
    }
    const static LisztOpType type = LISZT_LOR_TYPE;
};

struct LISZT_LXOR {
    template<typename T>
    static T identity() {
        return ReductionImpl<T>::lxor_identity();
    }
    template<typename T>
    static void apply(const T & d,T * v) {
        *v =  ReductionImpl<T>::lxor(*v,d);
    }
    const static LisztOpType type = LISZT_LXOR_TYPE;
};

struct LISZT_BAND {
    template<typename T>
    static T identity() {
        return ReductionImpl<T>::band_identity();
    }
    template<typename T>
    static void apply(const T & d,T * v) {
        *v = ReductionImpl<T>::band(*v,d);
    }
    const static LisztOpType type = LISZT_BAND_TYPE;
};

struct LISZT_BOR {
    template<typename T>
    static T identity() {
        return ReductionImpl<T>::bor_identity();
    }
    template<typename T>
    static void apply(const T & d,T * v) {
        *v = ReductionImpl<T>::bor(*v,d);
    }
    const static LisztOpType type = LISZT_BOR_TYPE;
};

struct LISZT_BXOR {
    template<typename T>
    static T identity() {
        return ReductionImpl<T>::bxor_identity();
    }
    template<typename T>
    static void apply(const T & d,T * v) {
        *v =  ReductionImpl<T>::bxor(*v,d);
    }
    const static LisztOpType type = LISZT_BXOR_TYPE;
};

struct LISZT_READ {
	template<typename T>
	static T identity() {
		return Identities<T>::additive();
	}
	template<typename T>
    static void apply(const T & d,T * v) {}
	const static LisztOpType type = LISZT_READ_TYPE;
};

#endif /* REDUCTIONOPS_H_ */
