#ifndef _GPU_FIELD_H
#define _GPU_FIELD_H

#include "Gpu/Profiling.h"

namespace LisztGPU {

template<typename MT, typename VT>
class Field {
public:
	__host__ void init(size_t nelems, const VT & v) {
		//TODO(zach): Don't be so dumb about this
		VT * hValue = new VT[nelems];
		std::fill_n(hValue,nelems,v);
		init(nelems,hValue);
		delete [] hValue;
	}
	__host__ void init(size_t nelems, const VT * data) {
		DEVICEALLOC(&value, sizeof(VT) * nelems);
		cudaMemcpy(value,data,nelems * sizeof(VT),cudaMemcpyHostToDevice);
	}
	template<typename MO>
	__device__ VT & operator[](const MO & mo) {
		GRDM_INCREMENT_BYTES_ACCESSED(VT)
		return value[mo.ID()];
	}
	
	template<typename MO>
	__device__ VT operator()(const MO & mo) const {
		GRDM_INCREMENT_BYTES_ACCESSED(VT)
		return value[mo.ID()];
	}

private:
	VT * value; //pointer to GPU Device memory.
};

}
#endif