/*
 * GPUField.cpp
 *
 *  Created on: Mar 2, 2010
 *      Author: njoubert
 */

#include "GPUField.h"


__global__ void GPUMRField_reduceKernelADD_float(
		idx_type tyS,
		idx_type thS,
		idx_type * dtype_indices,
		idx_type * dthread_indices,
		float * dthread_values,
		float * dresult) {

	idx_type myId = threadIdx.x + blockDim.x * blockIdx.x;

	if (myId >= tyS)
		return;

	//look up your section of the values array.
	idx_type th_idx_s;
	idx_type th_idx_e;

	//get the start and end of map keys in th_indx array
	th_idx_s = dtype_indices[myId];
	if (myId == tyS-1)
		th_idx_e = thS;
	else
		th_idx_e = dtype_indices[myId+1];

	//reduce them!
	float res = 0;
	for (idx_type i = th_idx_s; i < th_idx_e; i++) {
		res += dthread_values[i];
	}

	//write result to dresult
	dresult[myId] += res;

}

__global__ void GPUMRField_reduceKernelADD_float3(
		idx_type tyS,
		idx_type thS,
		idx_type * dtype_indices,
		idx_type * dthread_indices,
		float3 * dthread_values,
		float3 * dresult) {

	idx_type myId = threadIdx.x + blockDim.x * blockIdx.x;

	if (myId >= tyS)
		return;

	//look up your section of the values array.
	idx_type th_idx_s;
	idx_type th_idx_e;

	//get the start and end of map keys in th_indx array
	th_idx_s = dtype_indices[myId];
	if (myId == tyS-1)
		th_idx_e = thS;
	else
		th_idx_e = dtype_indices[myId+1];

	//reduce them!
	float3 res = make_float3(0.f,0.f,0.f);
	for (idx_type i = th_idx_s; i < th_idx_e; i++) {
		float3 tmp = dthread_values[i];
		res.x += tmp.x;
		res.y += tmp.y;
		res.z += tmp.z;
	}

	//write result to dresult
	dresult[myId].x += res.x;
	dresult[myId].y += res.y;
	dresult[myId].z += res.z;
}

//todo(njoubert): write a big macro
//
//
//#define GPUFIELDOP(name, op)	\
//		__global__ void GPUMRField_reduceKernel##name(	\
//				idx_type tyS,	\
//				idx_type thS,	\
//				idx_type * dtype_indices,	\
//				idx_type * dthread_indices,	\
//				float * dthread_values,	\
//				float * dresult) {	\
//			idx_type myId = threadIdx.x + blockDim.x * blockIdx.x;
//
//AD	if (myId >= tyS)
//				return;
//
//			//look up your section of the values array.
//			idx_type th_idx_s;
//			idx_type th_idx_e;
//
//			//get the start and end of map keys in th_indx array
//			th_idx_s = dtype_indices[myId];
//			if (myId == tyS-1)
//				th_idx_e = thS;
//			else
//				th_idx_e = dtype_indices[myId+1];
//
//			//reduce them!
//			float res = 0;
//			for (idx_type i = th_idx_s; i < th_idx_e; i++) {
//				res += dthread_values[i];
//			}
//
//			//write result to dresult
//			dresult[myId] += res;
//
//		}
//
//
//GPUFIELDOP(ADD, +=)
//GPUFIELDOP(MULTIPLY, *=)

