#ifndef _GPU_PRINT_H
#define _GPU_PRINT_H

#ifdef PROJECT_DEBUG

__shared__ PrintContext * print_context_head[BLOCKSIZE];


__device__ __host__ PrintContext * PrintContext_top(); //defined in kernel.cu

__device__ __host__ PrintContext * PrintContext_head() {
#ifdef __CUDA_ARCH__
	return print_context_head[threadIdx.x];
#else
	return NULL;
#endif
}
__device__ void PrintContext_init() {
	print_context_head[threadIdx.x] = NULL;
}
__device__ PrintContext *  PrintContext_push(PrintContext * new_head) {
	new_head->type = 'u';
	new_head->count = 0;
	new_head->id = -1;
	PrintContext * head = PrintContext_head();
	if(head == NULL) {
		new_head->next = new_head;
	} else {
		PrintContext * first = head->next;
		head->next = new_head;
		new_head->next = first;
	}
	print_context_head[threadIdx.x] = new_head;
	return head;
}

// a -> b -> c
__device__ void PrintContext_pop(PrintContext * old_head) {
	PrintContext * head = PrintContext_head();
	print_context_head[threadIdx.x] = old_head;
	if(old_head != NULL) {
		old_head->count++;
		old_head->next = head->next;
	}
}

__device__ void PrintContext_mark_v(const MeshElement & me) {
	PrintContext * head = PrintContext_head();
	head->id = ID_v(me);
	head->type = 'v';
	head->count = 0;
}
__device__ void PrintContext_mark_e(const MeshElement & me) {
	PrintContext * head = PrintContext_head();
	head->id = ID_e(me);
	head->type = 'e';
	head->count = 0;
}
__device__ void PrintContext_mark_f(const MeshElement & me) {
	PrintContext * head = PrintContext_head();
	head->id = ID_f(me);
	head->type = 'f';
	head->count = 0;
}
__device__ void PrintContext_mark_c(const MeshElement & me) {
	PrintContext * head = PrintContext_head();
	head->id = ID_c(me);
	head->type = 'c';
	head->count = 0;
}

//we have a printf but no sprintf so we need to do this ourself
__device__ __host__ int PrintContext_itos(char * buf, uint32_t i) {
	char nbuf[10];
	int c = 0;
	do {
		nbuf[c++] = '0' + (i % 10);
		i /= 10;
	} while(i != 0);
	int total_chars = c;
	while(c > 0) {
		*buf++ = nbuf[c-1];
		c--;
	}
	return total_chars;
}
__device__ __host__ int PrintContext_sprint(char * buf, const PrintContext * ctx) {
	size_t len = 0;
	*(buf + len) = ctx->type;
	len++;
	len += PrintContext_itos(buf + len, ctx->id);
	*(buf + len) = '-';
	len++;
	len += PrintContext_itos(buf + len,ctx->count);
	return len;
}
__device__ __host__ void PrintContext_print(char * buf) {
	size_t len = PrintContext_itos(buf,PrintContext_top()->count);
	#ifdef __CUDA_ARCH__
	PrintContext * head = PrintContext_head();
	PrintContext * cur = head;
	do {
		cur = cur->next;
		*(buf + len) = ':';
		len++;
		len += PrintContext_sprint(buf + len,cur);
	} while(cur != head);
	head->count++;
	#else
	PrintContext_top()->count++;
	#endif
	buf[len] = '\0';
}
#endif //PROJECT_DEBUG
 
#endif //_GPU_PRINT_H