//LINKER: Expects dev_bytes_accessed, dev_flops_performed to be available

#ifndef GPU_PROFILING_H_
#define GPU_PROFILING_H_

#include <cuda_runtime.h>
#include <cuda.h>


//TODO(njoubert): Also profile Global accesses and profile Field Reductions 

/* 
 * Macros we use for all profiling 
 */
#if defined(PROFILE_COUNTING) || defined(PROFILE_TIMING)
extern FILE* profiling_file;
#endif

/* 
 * Macros we use for profiling counting of memory and flops usage. 
 */
#if defined(PROFILE_COUNTING) && defined(__CUDACC__)
	__global__ void reset_memory_profiling_counters() {
		dev_bytes_accessed = 0;
		dev_flops_performed = 0;
	}
	__host__ __device__ void grdm_count_flops(int cnt) {
#ifdef __CUDA_ARCH__
           dev_flops_performed_shared[threadIdx.x] += cnt;
#endif
        }	
	#define GRHM_RESET_BYTES_ACCESSED \
		reset_memory_profiling_counters<<<1,1>>>();
		
	#define GRHM_OUTPUT_COUNTING_DETAILS(workgroup_id, numcolors) do {\
		unsigned long long int count; \
		unsigned long long int flops; \
		cudaMemcpyFromSymbol(&count,"dev_bytes_accessed",sizeof(unsigned long long int),0,cudaMemcpyDeviceToHost); \
		cudaMemcpyFromSymbol(&flops,"dev_flops_performed",sizeof(unsigned long long int),0,cudaMemcpyDeviceToHost); \
		fprintf(profiling_file, "%d,%llu,%u,%llu\n", workgroup_id, count, numcolors,flops); \
		} while (0);

	#define GRDM_INCREMENT_BYTES_ACCESSED(tp) \
		atomicAdd(&dev_bytes_accessed,sizeof(tp));

	#define GRDM_INCREMENT_FLOPS(cnt) \
	    grdm_count_flops(cnt);
	    
	#define GRDM_KERNEL_BEGIN \
		dev_flops_performed_shared[threadIdx.x] = 0;
	#define GRDM_KERNEL_END \
		atomicAdd(&dev_flops_performed,dev_flops_performed_shared[threadIdx.x]);
#else
	#define GRHM_RESET_BYTES_ACCESSED
	#define GRHM_OUTPUT_COUNTING_DETAILS(workgroup_id,numcolors)
	#define GRDM_INCREMENT_BYTES_ACCESSED(tp)
	#define GRDM_INCREMENT_FLOPS(cnt)
	#define GRDM_KERNEL_BEGIN
	#define GRDM_KERNEL_END
#endif

/* 
 * Macros we use for profiling timing. 
 */
#ifdef PROFILE_TIMING
	#define GRDM_PROFILE_TIMING_START(workgroup_id) \
		start_timer(&timer);

	#define GRDM_PROFILE_TIMING_STOP(workgroup_id) \
		fprintf(profiling_file, "%d,%f\n", workgroup_id, stop_get_timer(&timer));

	
	
	
	

#else
	#define GRDM_PROFILE_TIMING_START(workgroup_id)
	#define GRDM_PROFILE_TIMING_STOP(workgroup_id)
#endif

#endif /* GPU_PROFILING_H_ */
