#ifndef _GPU_THRUSTREDUCTIONS_H
#define _GPU_THRUSTREDUCTIONS_H

#include <thrust/reduce.h>
#include <thrust/functional.h>
#include "Gpu/Operators.h"

/*
 * GPU Implementations of reduction operations. See Fields/ReductionOps.h
 */

namespace LisztGPU {
namespace reductions {

template<typename T>
struct sum : public thrust::binary_function<T,T,T> {
	__host__ __device__ T operator()(const T &lhs, const T &rhs) const {return ReductionImpl<T>::plus(lhs,rhs);}
}; 

template<typename T>
struct product : public thrust::binary_function<T,T,T> {
	__host__ __device__ T operator()(const T &lhs, const T &rhs) const {return ReductionImpl<T>::times(lhs,rhs);}
}; 
	
template<typename T>
struct max : public thrust::binary_function<T,T,T> {
	__host__ __device__ T operator()(const T &lhs, const T &rhs) const {return ReductionImpl<T>::max(lhs,rhs);}
};

template<typename T>
struct min : public thrust::binary_function<T,T,T> {
	__host__ __device__ T operator()(const T &lhs, const T &rhs) const {return ReductionImpl<T>::min(lhs,rhs);}
};

template<typename T> 
struct logical_and : public thrust::binary_function<T,T,T>{
	__host__ __device__ T operator()(const T &lhs, const T &rhs) const {return ReductionImpl<T>::land(lhs,rhs);}
}; 

template<typename T>
struct logical_or : public thrust::binary_function<T,T,T> {
	__host__ __device__ T operator()(const T &lhs, const T &rhs) const {return ReductionImpl<T>::lor(lhs,rhs);}
};

template<typename T>
struct logical_xor : public thrust::binary_function<T,T,T> {
	__host__ __device__ T operator()(const T &lhs, const T &rhs) const {return ReductionImpl<T>::lxor(lhs,rhs);}
};

template<typename T>
struct binary_and : public thrust::binary_function<T,T,T> {
	__host__ __device__ T operator()(const T &lhs, const T &rhs) const {return ReductionImpl<T>::band(lhs,rhs);}
};

template<typename T>
struct binary_or : public thrust::binary_function<T,T,T> {
	__host__ __device__ T operator()(const T &lhs, const T &rhs) const {return ReductionImpl<T>::bor(lhs,rhs);}
};

template<typename T>
struct binary_xor : public thrust::binary_function<T,T,T> {
  __host__ __device__ T operator()(const T &lhs, const T &rhs) const {return ReductionImpl<T>::bxor(lhs,rhs);}
};

} /* namespace reductions */
} /* namespace LisztGPU */

#endif /* _GPU_THRUSTREDUCTIONS_H */
