#ifndef _WORKGROUP_LAUNCHER_H
#define _WORKGROUP_LAUNCHER_H

#define GPU_BOTH __host__ __device__
#define GPU_ONLY __device__

#include "InterferenceGraph/ColoringController.h"

class ColorBatch {
public:
	void init(const Mesh * dM, id_type * start_, id_type * end_) {
		dMesh = dM;
		start = start_;
		nelems = end_ - start_;
	}
	GPU_BOTH id_type size() const {
		return nelems;
	}
	GPU_BOTH id_type kernel_size() const {
		return nelems;
	}
	GPU_ONLY bool elementAt(id_type idx, MeshElement * elem) const {
		if(idx < size()) {
			elem->init(dMesh,start[idx]);
			return true;
		} else return false;
	}
private:
	const Mesh * dMesh;
	id_type * start;
	size_t nelems;
};
class WorkgroupLauncher {
public:
	void init(const Mesh * dM, size_t ncolors_, CRS * crs_) {
		dMesh = dM;
		crs = crs_;
		next_color = 0;
		ncolors = ncolors_;
	}
	bool nextBatch(ColorBatch * batch) {
		if(next_color >= ncolors)
			return false;
		id_type * start = crs->values + crs->row_idx[next_color];
		id_type * end = crs->values + crs->row_idx[next_color + 1];
		next_color++;
		batch->init(dMesh,start,end);
		return true;
	}
	size_t numColors() {
		return ncolors;
	}
private:
	size_t next_color;
	size_t ncolors;
	const Mesh * dMesh;
	CRS * crs;
};

extern Synchronization::ColoringController coloring_controller;
extern Mesh * dummydMesh;

template<typename MeshSet>
void WorkgroupLauncher_forWorkgroup(int wg, const MeshSet & mesh_set, bool isTrivial, WorkgroupLauncher * launcher) {
	CRS * crs;
	int ncolors = coloring_controller.getCRSColoring(wg,mesh_set,isTrivial,&crs);
	launcher->init(dummydMesh,ncolors,crs);	
}

#undef GPU_BOTH
#undef GPU_ONLY

#endif //_WORKGROUP_LAUNCHER_H