
inline Cell::Cell(const CellData & data) : MeshObject(data) {}


class Cell_face_it : public DFSIterBase<Cell_face_it,Face> {
public:
    typedef Face value_type;
    Cell_face_it(const FacetEdge & ife, const Cell & cell) {
        FacetEdge init = ife.towards(cell);
        assert(init.outside() == cell);
        visit(init);

        for(unsigned int i = 0; i < stk->size(); i++) {
            const FacetEdge & fe = (*stk)[i];

            FacetEdge it = fe;
            do {
                it = it.facetedgeCCWAroundFace();
                FacetEdge fe2 = it.facetedgeCCWAroundEdge().flip();
                assert(fe2.outside() == cell);
                visit(fe2);
            } while(it != fe);
        }

    }
private:
    friend class DFSIterBase<Cell_face_it,Face>;
    value_type get(const FacetEdge & fe) const {
        return fe.face();
    }
    const FacetEdge & key(const FacetEdge & fe) const {
        return fe.internal()->face->ife;
    }
};

//Cell_cell iterator assumes that only 1 face separates two cells
//in other words, it iterates across the faces and retrieves
//the cell on the other side
//if there is more than one face between this cell and its neighbor
//that neighboring cell will appear multiple times
class Cell_cell_it {
public:
    typedef Cell value_type;
    Cell_cell_it(const FacetEdge & ife, const Cell & c)
    : fit(ife,c) {
    }
    bool hasNext() const { return fit.hasNext(); }
    value_type next() {
        return fit.next().inside();
    }
private:
    Cell_face_it fit;
};


class Cell_edge_it : public DFSIterBase<Cell_edge_it,Edge> {
public:
    typedef Edge value_type;
    Cell_edge_it(const FacetEdge & begin, const Cell & cell) {
        FacetEdge init = begin.towards(cell);
        visit(init);
        for(unsigned int i = 0; i < stk->size(); i++) {
            const FacetEdge & fe = (*stk)[i];
            assert(fe.outside() == cell);
            visit(fe.facetedgeCCWAroundFace());
            visit(fe.facetedgeCWAroundFace());

            FacetEdge fe2 = fe.facetedgeCCWAroundEdge().flip();

            assert(fe2.outside() == cell);
            visit(fe2.facetedgeCCWAroundFace());
            visit(fe2.facetedgeCWAroundFace());
        }
    }

private:
    friend class DFSIterBase<Cell_edge_it,Edge>;
    value_type get(const FacetEdge & fe) const {
        return fe.edge();
    }
    const FacetEdge & key(const FacetEdge & fe) const {
        return fe.internal()->edge->ife;
    }
};

class Cell_vertex_it : public DFSIterBase<Cell_vertex_it,Vertex> {
public:
    typedef Vertex value_type;
    Cell_vertex_it(const FacetEdge & begin, const Cell & cell) {
        visit(begin.towards(cell));

        for(unsigned int i = 0; i < stk->size(); i++) {
            const FacetEdge & fd = (*stk)[i];

            FacetEdge it = fd;
            assert(fd.outside() == cell);
            do {
              it = it.facetedgeCCWAroundFace();
              assert(it.tail() == fd.head());
              visit(it.internal());
              it = it.facetedgeCCWAroundEdge().flip();
              assert(it.head() == fd.head());
            } while(it != fd);

        }
    }
private:
    friend class DFSIterBase<Cell_vertex_it,Vertex>;
    value_type get(const FacetEdge & fe) const {
        return fe.head();
    }
    const FacetEdge & key(const FacetEdge & fe) const {
        return fe.internal()->vertex->ife;
    }
};

SUBCLASS_MESHSET(Cell,vertex_set,Cell_vertex_it)
SUBCLASS_MESHSET(Cell,edge_set,Cell_edge_it)
SUBCLASS_MESHSET(Cell,face_set,Cell_face_it)
SUBCLASS_MESHSET(Cell,cell_set,Cell_cell_it)

inline Cell::vertex_set Cell::vertices() const {
    return vertex_set(Cell_vertex_it(data()->ife,*this));
}
inline Cell::edge_set Cell::edges() const {
    return edge_set(Cell_edge_it(data()->ife,*this));
}
inline Cell::face_set Cell::faces() const {
    return face_set(Cell_face_it(data()->ife,*this));
}
inline Cell::cell_set Cell::cells() const {
    return cell_set(Cell_cell_it(data()->ife,*this));
}


