/*
 * DFSIterBase.h
 *
 *  Created on: Aug 12, 2009
 *      Author: zdevito
 */

#ifndef DFSITERBASE_H_
#define DFSITERBASE_H_

//base for iterators using DFS
//iterators using this class must provide this functionality:
//1. the function const FacetEdge & key(FacetEdge & value) should return the facet-edge that serves
//as the key which is looked up for a value to check if value has been visited already
//2. the function value_type get(FacetEdge & value) should return the value associated with the facetedge
//3. the child should provide a constructor that fills stack with all the FacetEdges needed to represent the
//current set
//4. friend the DFSBase class so it can access get() and key()

template<typename T, typename VT>
class DFSIterBase {
    typedef std::vector<FacetEdge> list_t;
    //this is a shared pointer to allow for the DFS iterators to be copied
    //and used more than once
    typedef ::std::tr1::shared_ptr<list_t> list_p;
public:
    DFSIterBase()
    : stk(new list_t), tag(VisitorTags::instance().alloc()) {}
    bool hasNext() const {
        return !stk->empty();
    }
    VT next() {
        VT e = me()->get(stk->back());
        stk->pop_back();
        return e;
    }
protected:
    //try to visit key, if it has not been visited, add value to the stack
    void visit(FacetEdge value) {
        if(!VisitorTags::instance().visit(tag,me()->key(value))) {
           stk->push_back(value);
        }
    }
    list_p stk;
    VisitorTags::tag_type tag;
private:
    T * me() {
        return static_cast<T*>(this);
    }
};

#endif /* DFSITERBASE_H_ */
