
inline Edge::Edge(const EdgeData & data_)
: MeshObject(data_), info(data_.ife) {
}
inline Edge::Edge(const EdgeData & data, const FacetEdge & fe_)
: MeshObject(data), info(fe_) {
}

inline Vertex Edge::head() const {
    return info.head();
}
inline Vertex Edge::tail() const {
    return info.tail();
}
inline  Vertex Edge::opposite(const Vertex & v) const {
    return info.opposite(v);
}
inline void Edge::flip() {
    info = info.flip();
}
inline void Edge::towards(const Vertex & v) {
    info = info.towards(v);
}
inline FacetEdge Edge::with(const Face & f) const {
    const FacetEdge & start = info;
    FacetEdge ret = start;
    do {
        ret = ret.facetedgeCCWAroundEdge();
    } while(ret.face() != f && ret != start);
    assert(ret.face() == f);
    return ret;
}

class Edge_vertex_it {
public:
    typedef Vertex value_type;
    Edge_vertex_it(const FacetEdge & i)
    : orig(i), n(0) {}
    bool hasNext() const { return n < 2; }
    value_type next() {
        return (n++ < 1) ? orig.head() : orig.tail();
    }
private:
    const FacetEdge orig;
    int n;
};
class Edge_face_it {
public:
    typedef Face value_type;
    Edge_face_it(const FacetEdge & fe)
    : cur(fe), orig(fe), done(false) {}
    bool hasNext() const {
        return !done;
    }
    value_type  next() {
        cur = cur.facetedgeCCWAroundEdge();
        Face f = cur.face();
        if(cur == orig)
            done = true;
        return f;
    }
private:
    FacetEdge cur;
    FacetEdge orig;
    bool done;
};

class Edge_cell_it {
public:
    typedef Cell value_type;
    Edge_cell_it(const FacetEdge & fe)
    : cur(fe), orig(fe), done(false)  {}
    bool hasNext() const { return !done; }
    value_type next() {
        cur = cur.facetedgeCCWAroundEdge();
        Cell c = cur.outside();
        if(cur == orig) done = true;
        return c;
    }
private:
    FacetEdge cur;
    FacetEdge orig;
    bool done;
};

class Edge_facet_it {
public:
    typedef FacetEdge value_type;
    Edge_facet_it(const FacetEdge & fe) : cur(fe), orig(fe), done(false) {
    }
    bool hasNext() const {
        return !done;
    }
    value_type  next() {
        cur = cur.facetedgeCCWAroundEdge();
        if(cur == orig) done = true;
        return cur;
    }
private:
    FacetEdge cur;
    FacetEdge orig;
    bool done;
};

SUBCLASS_SIZEDMESHSET(Edge,vertex_set,Edge_vertex_it)
SUBCLASS_MESHSET(Edge,face_set,Edge_face_it)
SUBCLASS_MESHSET(Edge,facet_set,Edge_facet_it)
SUBCLASS_MESHSET(Edge,cell_set,Edge_cell_it)


inline Edge::face_set Edge::facesCCW() const {
    return face_set(Edge_face_it(info));
}
inline Edge::face_set Edge::facesCCW(const Vertex & v) const {
    return face_set(Edge_face_it(info.towards(v)));
}
inline Edge::face_set Edge::facesCW() const {
    return face_set(Edge_face_it(info.flip()));
}
inline Edge::face_set Edge::facesCW(const Vertex & v) const {
    return face_set(Edge_face_it(info.towards(v).flip()));
}



inline Edge::cell_set Edge::cellsCCW() const {
    return cell_set(Edge_cell_it(info));
}
inline Edge::cell_set Edge::cellsCCW(const Vertex & v) const {
    return cell_set(Edge_cell_it(info.towards(v)));
}
inline Edge::cell_set Edge::cellsCW() const {
    return cell_set(Edge_cell_it(info.flip()));
}
inline Edge::cell_set Edge::cellsCW(const Vertex & v) const {
    return cell_set(Edge_cell_it(info.towards(v).flip()));
}


inline Edge::vertex_set Edge::vertices() const {
    return vertex_set(Edge_vertex_it(info),2);
}
inline Edge::face_set Edge::faces() const {
    return facesCCW();
}
inline Edge::cell_set Edge::cells() const {
    return cellsCCW();
}
inline Edge::facet_set Edge::facets() const {
    return facet_set(Edge_facet_it(info));
}
