
inline Face::Face(const FaceData & data_)
: MeshObject(data_), info(data_.ife) {
}
inline Face::Face(const FaceData & data_, const FacetEdge & fe_)
: MeshObject(data_), info(fe_) {
}

inline Cell Face::outside() const {
    return info.outside();
}
inline Cell Face::inside() const {
    return info.inside();
}
inline  Cell Face::opposite(const Cell & c) const {
    return info.opposite(c);
}
inline void Face::flip() {
    info = info.flip();
}
inline void Face::towards(const Cell & c) {
    info = info.towards(c);
}
inline FacetEdge Face::with(const Edge & e) const {
    const FacetEdge & start = info;
    FacetEdge ret = start;
    do {
        ret = ret.facetedgeCCWAroundFace();
    } while(ret.edge() != e && ret != start);
    assert(ret.edge() == e);
    return ret;
}

class Face_cell_it {
public:
    typedef Cell value_type;
    Face_cell_it(const FacetEdge & i)
    : orig(i), n(0) { }
    bool hasNext() const { return n < 2; }
    value_type next() {
        return (n++ < 1) ? orig.outside() : orig.inside();
    }
private:
    const FacetEdge orig;
    int n;
};

class Face_edge_it {
public:
    typedef Edge value_type;
    Face_edge_it(const FacetEdge & fe)
    : cur(fe), orig(fe), done(false) {
    }
    bool hasNext() const {
        return !done;
    }
    value_type  next() {
        cur = cur.facetedgeCCWAroundFace();
        Edge e = cur.edge();
        if(cur == orig)
            done = true;
        return e;
    }
private:
    FacetEdge cur;
    FacetEdge orig;
    bool done;
};

class Face_vertex_it {
public:
    typedef Vertex value_type;
    Face_vertex_it(const FacetEdge & fe)
    : cur(fe),  orig(fe), done(false) {
    }
    bool hasNext() const { return !done; }
    value_type next() {
        cur = cur.facetedgeCCWAroundFace();
        Vertex v = cur.head();
        if(cur == orig)
            done = true;
        return v;
    }
private:
    FacetEdge cur;
    FacetEdge orig;
    bool done;

};

class Face_facet_it {
public:
    typedef FacetEdge value_type;
    Face_facet_it(const FacetEdge & fe)
    : cur(fe), orig(fe), done(false) {
    }
    bool hasNext() const {
        return !done;
    }
    value_type  next() {
        cur = cur.facetedgeCCWAroundFace();
        if(cur == orig)
            done = true;
        return cur;
    }
private:
    FacetEdge cur;
    FacetEdge orig;
    bool done;
};

SUBCLASS_MESHSET(Face,vertex_set,Face_vertex_it)
SUBCLASS_MESHSET(Face,edge_set,Face_edge_it)
SUBCLASS_MESHSET(Face,facet_set,Face_facet_it)
SUBCLASS_SIZEDMESHSET(Face,cell_set,Face_cell_it)

inline Face::vertex_set Face::vertices() const {
    return verticesCCW();
}
inline Face::vertex_set Face::verticesCCW() const {
    return vertex_set(Face_vertex_it(info));
}
inline Face::vertex_set Face::verticesCCW(const Cell & c) const {
    return vertex_set(Face_vertex_it(info.towards(c)));
}
inline Face::vertex_set Face::verticesCW() const {
    return vertex_set(Face_vertex_it(info.flip()));
}
inline Face::vertex_set Face::verticesCW(const Cell & c) const {
    return vertex_set(Face_vertex_it(info.towards(c).flip()));
}


inline Face::edge_set Face::edges() const {
    return edgesCCW();
}
inline Face::edge_set Face::edgesCW() const {
    return edge_set(Face_edge_it(info.flip()));
}
inline Face::edge_set Face::edgesCW(const Cell & c) const {
    return edge_set(Face_edge_it(info.towards(c).flip()));
}
inline Face::edge_set Face::edgesCCW() const {
    return edge_set(Face_edge_it(info));
}
inline Face::edge_set Face::edgesCCW(const Cell & c) const {
    return edge_set(Face_edge_it(info.towards(c)));
}

inline Face::cell_set Face::cells() const {
    return cell_set(Face_cell_it(info),2);
}
inline Face::facet_set Face::facets() const {
    return facet_set(Face_facet_it(info));
}
