inline FacetEdge::FacetEdge() {
    data = NULL;
}
inline FacetEdge::FacetEdge(const FacetEdgeData * i) {
    data = i;
}
inline FacetEdge::FacetEdge(const FacetEdge & fe) {
    data = fe.data;
}
inline void FacetEdge::reset(const FacetEdgeData * i) {
    data = i;
}
inline void FacetEdge::reset(const FacetEdge & fe) {
    data = fe.data;
}

inline FacetEdge FacetEdge::towards(const Cell & c) const {
    //makes ret->outside() == c
    FacetEdge fe = (outside() != c) ? flip() : *this;
    assert(fe.outside() == c);
    return fe;
}
inline FacetEdge FacetEdge::towards(const Face & f) const {
    //makes ret->facetedgeCCWAroundEdge().face() == f
    FacetEdge fe = (facetedgeCCWAroundEdge().face() != f) ? flip() : *this;
    assert(fe.facetedgeCCWAroundEdge().face() == f);
    return fe;
}

inline FacetEdge FacetEdge::towards(const Edge & e) const {
    //makes this->facetedgeCCWAroundFace().edge() == e
    FacetEdge fe = (facetedgeCCWAroundFace().edge() != e) ? flip() : *this;
    assert(fe.facetedgeCCWAroundFace().edge() == e);
    return fe;
}
inline FacetEdge FacetEdge::towards(const Vertex & v) const {
    //makes this->head() == v
    FacetEdge fe = (head() != v) ? flip() : *this;
    assert(fe.head() == v);
    return fe;
}

bool FacetEdge::operator==(const FacetEdge & rhs) const {
    return ID() == rhs.ID();
}
bool FacetEdge::operator!=(const FacetEdge & rhs) const {
    return !(*this == rhs);
}

Vertex FacetEdge::head() const {
    return Vertex(*data->vertex);
}
Vertex FacetEdge::tail() const {
    return Vertex(*data->opposite()->vertex);
}
Edge  FacetEdge::edge() const {
    return Edge(*data->edge);
}
Face  FacetEdge::face() const {
    return Face(*data->face);
}
Cell FacetEdge::outside() const {
    return Cell(*data->cell);
}
Cell FacetEdge::inside() const {
    return Cell(*data->opposite()->cell);
}

Cell FacetEdge::opposite(const Cell & c) const {
    Cell o = outside();
    if(o == c) {
        return inside();
    } else {
       assert(inside() == c);
       return o;
    }
}
Vertex FacetEdge::opposite(const Vertex & v) const {
    Vertex h = head();
    if(h == v) {
        return tail();
    } else {
        assert(tail() == v);
        return h;
    }
}
inline FacetEdge FacetEdge::flip() const {
    //returns the facet edge, return->head() == this->tail(), and return->outside() == this->inside()
    return FacetEdge(data->opposite());
}

inline id_type FacetEdge::ID() const {
    //each FacetEdge data stores 2 half-facetedge structures, so the FacetEdge ID is
    //the top 31-bits of the F
    return (data->id >> 1);
}

inline Edge FacetEdge::edgeThroughVertex(const Vertex & v) const {
    return towards(v).edgeThroughVertex();
}
inline Edge FacetEdge::edgeThroughVertex() const {
    return edgeCCWAroundFace();
}

inline Face FacetEdge::faceThroughCell(const Cell & c) const {
    return towards(c).faceCCWAroundEdge();
}
inline Face FacetEdge::faceThroughCell() const {
    return faceCCWAroundEdge();
}

inline Edge FacetEdge::edgeCWAroundFace(const Cell & e) const {
    return towards(e).edgeCWAroundFace();
}
inline Edge FacetEdge::edgeCWAroundFace() const {
    return facetedgeCWAroundFace().edge();
}
inline Edge FacetEdge::edgeCCWAroundFace(const Cell & e) const {
    return towards(e).edgeCCWAroundFace();
}
inline Edge FacetEdge::edgeCCWAroundFace() const {
    return facetedgeCCWAroundFace().edge();
}
inline Vertex FacetEdge::vertexCWAroundFace(const Cell & e) const {
    return towards(e).vertexCWAroundFace();
}
inline Vertex FacetEdge::vertexCWAroundFace() const {
    return flip().vertexCCWAroundFace();
}
inline Vertex FacetEdge::vertexCCWAroundFace(const Cell & e) const {
    return towards(e).vertexCCWAroundFace();
}
inline Vertex FacetEdge::vertexCCWAroundFace() const {
    return head();
}

inline Face FacetEdge::faceCWAroundEdge(const Vertex & e) const {
    return towards(e).faceCWAroundEdge();
}
inline Face FacetEdge::faceCWAroundEdge() const {
    return facetedgeCWAroundEdge().face();
}
inline Face FacetEdge::faceCCWAroundEdge(const Vertex & e) const {
    return towards(e).faceCCWAroundEdge();
}
inline Face FacetEdge::faceCCWAroundEdge() const {
    return facetedgeCCWAroundEdge().face();
}

inline Cell FacetEdge::cellCWAroundEdge() const {
    return flip().cellCCWAroundEdge();
}

inline Cell FacetEdge::cellCWAroundEdge(const Vertex & v) const {
    return towards(v).cellCWAroundEdge();
}

inline Cell FacetEdge::cellCCWAroundEdge(const Vertex & e) const {
    return towards(e).cellCCWAroundEdge();
}
inline Cell FacetEdge::cellCCWAroundEdge() const {
    return outside();
};

inline FacetEdge FacetEdge::facetedgeCWAroundEdge(const Vertex & e) const {
    return towards(e).facetedgeCWAroundEdge();
}
inline FacetEdge FacetEdge::facetedgeCWAroundEdge() const {
    return flip().facetedgeCCWAroundEdge().flip();
};
inline FacetEdge FacetEdge::facetedgeCCWAroundEdge(const Vertex & e) const {
    return towards(e).facetedgeCCWAroundEdge();
}
inline FacetEdge FacetEdge::facetedgeCCWAroundEdge() const {
    //this function actually does something!
    return FacetEdge(data->ccwAroundEdge);
};

inline FacetEdge FacetEdge::facetedgeCWAroundFace(const Cell & e) const {
    return towards(e).facetedgeCWAroundFace();
}
inline FacetEdge FacetEdge::facetedgeCWAroundFace() const {
    return flip().facetedgeCCWAroundFace().flip();
};
inline FacetEdge FacetEdge::facetedgeCCWAroundFace(const Cell & e) const {
    return towards(e).facetedgeCCWAroundFace();
}
inline FacetEdge FacetEdge::facetedgeCCWAroundFace() const {
    //this function actually does something!
    return FacetEdge(data->ccwAroundFace);
};

inline const FacetEdgeData * FacetEdge::internal() const { return data; }
inline std::ostream & operator<<(std::ostream & o, const FacetEdge & fe) {
    return o << "FacetEdge Towards: " << *fe.data << " away from: " << *fe.data->opposite() << endl;
}

