inline FacetEdgeData::FacetEdgeData(id_type id_)
: id(id_),
  edge(NULL), face(NULL),
  vertex(NULL), cell(NULL),
  ccwAroundFace(NULL),
  ccwAroundEdge(NULL) {
      VisitorTags::instance().add(this);
  }

inline FacetEdgeData::FacetEdgeData(id_type i, EdgeData * ed, FaceData * fd, VertexData * vd, CellData * cd,
              FacetEdgeData * ccwAF, FacetEdgeData * ccwAE)
: id(i), edge(ed), face(fd), vertex(vd), cell(cd),
  ccwAroundFace (ccwAF), ccwAroundEdge(ccwAE) {
      VisitorTags::instance().add(this);
}

inline FacetEdgeData * FacetEdgeData::opposite() const {
    //opposite only will work if FacetEdgeData is aligned to some power of two boundary
    BOOST_STATIC_ASSERT((is_power_two<sizeof(FacetEdgeData)>::value));
    return (FacetEdgeData*)( ((arithmetic_pointer_type)this) ^ OTHER_BIT);
}
inline bool FacetEdgeData::initialized() const {
    return vertex && edge && face && cell && ccwAroundFace && ccwAroundEdge;
}

const arithmetic_pointer_type FacetEdgeData::OTHER_BIT = (arithmetic_pointer_type) sizeof(FacetEdgeData);

inline std::ostream & operator<<(std::ostream & o, const FacetEdgeData & ife) {
    #define PRINT(x) ( (x) == NULL ? -1 : (x)->id )
    return o << "FE: " << ife.id <<
    " V: " << PRINT(ife.vertex)<<
    " E: " << PRINT(ife.edge) <<
    " F: " << PRINT(ife.face) <<
    " C: " << PRINT(ife.cell) <<
    " ccwAroundFace: " << PRINT(ife.ccwAroundFace) <<
    " ccwAroundEdge: " << PRINT(ife.ccwAroundEdge) <<
    " opposite: " << PRINT(ife.opposite()) << endl;
    #undef PRINT
}
