

#ifndef HFMESHFACETEDGEDATA_H
#define HFMESHFACETEDGEDATA_H

namespace HalfFacetMeshPrivate {
	
//an object that holds the data from half a facet-edge, it has several special properties:
//1. it has a size that is a power of two, making it a multiple of the cache line size
//2. it is allocated in memory aligned to 2 * sizeof(FacetEdgeData) and the opposite
//   facet edge is aligned right next to it in memory, since it is aligned
//   the opposite facet-edge can be obtained by fliping the sizeof(FaceEdgeData)-th bit
//3. By design, the half facet-edge holds only 1 of each type: vertex,edge,face,cell
//4. It provides an inline tag, visit_tag which can be used by iterators to mark
//   when this facet-edge is visited.  If iterators need to mark other elements
//   like marking a cell visited, they should use the facet-edge obtained 
//   from cell_data.ife, due to (3), no other cell can share this facet-edge
struct FacetEdgeData {
	arithmetic_pointer_type id;
	mutable VisitorTag visit_tag; //handles DFS visiting operations like cells.faces()
	
	EdgeData * edge;
	FaceData * face;
	
	VertexData * vertex;  //let tail be the vertex opposite this vertex on 'edge,'
	                      //then the vertex ccw around cell from tail will be this vertex
	CellData * cell;
	FacetEdgeData * ccwAroundFace;   //fe obtained by rotating around face ccw, from the perspective of cell ccwAroundFace.cell == cell, ccwAroundFace.face = face
	FacetEdgeData * ccwAroundEdge; //fe obtained by rotating around edge ccw, from the perspective of vertex ccwAroundEdge.vertex == vertex, ccwAroundEdge.edge = edge
	const static arithmetic_pointer_type OTHER_BIT; //flip this bit in the address of 'this' to obtain the other half of the facet-edge
	
	FacetEdgeData(id_type id_);
	
	FacetEdgeData(id_type i, EdgeData * ed, FaceData * fd, VertexData * vd, CellData * cd, 
				  FacetEdgeData * ccwAF = NULL, FacetEdgeData * ccwAE = NULL);
	
	FacetEdgeData * opposite() const;
	bool initialized() const;
};

inline std::ostream & operator<<(std::ostream & o, const FacetEdgeData & ife);

}
#endif
