/*
 * Util.h
 *
 *  Created on: Mar 8, 2009
 *      Author: zdevito
 */

#ifndef HFMESHUTIL_H_
#define HFMESHUTIL_H_
//common functions and forward decls
#include "type_traits.h"
#include <vector>
#include <limits>

namespace HalfFacetMeshPrivate {
using std::numeric_limits;
using std::vector;

//meta-function returning the type of the integer that is the same
//size as this architectures pointers
template<int size>
struct arithmetic_type_for_pointer_size;
template<>
struct arithmetic_type_for_pointer_size<4> {
    typedef unsigned int type;
};
template<>
struct arithmetic_type_for_pointer_size<8> {
    typedef unsigned long long type;

};

typedef arithmetic_type_for_pointer_size<sizeof(int*)>::type arithmetic_pointer_type;
const size_t BITS_IN_ARITHMETIC_POINTER = sizeof(arithmetic_pointer_type) << 3;

//meta-function that derives from true-type
//if i is a power of two
//used to make sure FacetEdgeData can find its partner by flipping a bit in its address
template<int i>
struct is_power_two;

template<bool a>
struct is_true;
template<>
struct is_true<true> : public Liszt::true_type {};
template<>
struct is_true<false> : public Liszt::false_type {};
template<int i>
struct is_power_two : public is_true< i != 0 && (i & (i - 1)) == 0 >{};

//structure that keeps track of visitors for DFS


//iterators needed to DFS should first call alloc() to obtain a tag for that iteration
//when the iterator has finished searching, it must call free(tag)
//copying a DFS-based iterator destroys the source iterator, using it will
//lead to undefined results

//the first visitor tag is stored inside the facet-edge to make the common
//case fast but subsequent ones are stored externally
struct VisitorTag {
    typedef arithmetic_pointer_type tag_type;
    tag_type tag;
    VisitorTag()
    : tag(0) {}
};

}
#endif /* UTIL_H_ */
