
inline Vertex::Vertex(const VertexData & data) : MeshObject(data) {}

class Vertex_edge_it : public DFSIterBase<Vertex_edge_it,Edge> {
public:
    typedef Edge value_type;
    Vertex_edge_it(const FacetEdge & ife, const Vertex & vertex) {
        FacetEdge init = ife.towards(vertex);
        assert(init.head() == vertex);
        visit(init);
        for(unsigned int i = 0; i < stk->size(); i++) {
            const FacetEdge & fe = (*stk)[i];
            FacetEdge it = fe;
            do {
               it = it.facetedgeCCWAroundEdge();
               FacetEdge fe2 = it.facetedgeCCWAroundFace().flip();
               assert(fe2.head() == vertex);
               visit(fe2);
           } while(it != fe);
        }
    }
private:
    friend class DFSIterBase<Vertex_edge_it,Edge>;
    value_type get(const FacetEdge & fe) const {
        return fe.edge();
    }
    const FacetEdge & key(const FacetEdge & fe) const {
        return fe.internal()->edge->ife;
    }
};

class Vertex_vertex_it {
public:
    typedef Vertex value_type;
    Vertex_vertex_it(const FacetEdge & ife, const Vertex & v) : eit(ife,v) {

    }
    bool hasNext() const { return eit.hasNext(); }
    value_type next() {
        return eit.next().tail();
    }
private:
    Vertex_edge_it eit;
};

class Vertex_face_it : public DFSIterBase<Vertex_face_it,Face> {
public:
    typedef Face value_type;
    Vertex_face_it(const FacetEdge & begin, const Vertex & vertex) {
        FacetEdge init = begin.towards(vertex);
        visit(init);
        for(unsigned int i = 0; i < stk->size(); i++) {
            const FacetEdge & fe = (*stk)[i];
            assert(fe.head() == vertex);
            visit(fe.facetedgeCCWAroundEdge());
            visit(fe.facetedgeCWAroundEdge());
            
			FacetEdge fe2 = fe.facetedgeCCWAroundFace().flip();
			
            assert(fe2.head() == vertex);
            visit(fe2.facetedgeCCWAroundEdge());
            visit(fe2.facetedgeCWAroundEdge());
        }
    }
private:
    friend class DFSIterBase<Vertex_face_it,Face>;
    value_type get(const FacetEdge & fe) const {
        return fe.face();
    }
    const FacetEdge & key(const FacetEdge & fe) const {
        return fe.internal()->face->ife;
    }
};

class Vertex_cell_it : public DFSIterBase<Vertex_cell_it,Cell> {
public:
    typedef Cell value_type;
    Vertex_cell_it(const FacetEdge & begin, const Vertex & vertex) {
        visit(begin.towards(vertex));
        for(unsigned int i = 0; i < stk->size(); i++) {
            const FacetEdge & fd = (*stk)[i];
            FacetEdge it = fd;
            assert(fd.head() == vertex);
            do {
                it = it.facetedgeCCWAroundEdge();
                //TODO(zach): double check these asserts
                assert(it.inside() == fd.outside());
                visit(it);
                it = it.facetedgeCCWAroundFace().flip();
                assert(it.outside() == fd.outside());
            } while(it != fd);
        }
    }
private:
    friend class DFSIterBase<Vertex_cell_it,Cell>;
    value_type get(const FacetEdge & fe) const {
        return fe.outside();
    }
    const FacetEdge & key(const FacetEdge & fe) const {
        return fe.internal()->cell->ife;
    }
};

SUBCLASS_MESHSET(Vertex,vertex_set,Vertex_vertex_it)
SUBCLASS_MESHSET(Vertex,edge_set,Vertex_edge_it)
SUBCLASS_MESHSET(Vertex,face_set,Vertex_face_it)
SUBCLASS_MESHSET(Vertex,cell_set,Vertex_cell_it)

inline Vertex::vertex_set Vertex::vertices() const {
    return vertex_set(Vertex_vertex_it(data()->ife,*this));
}
inline Vertex::edge_set Vertex::edges() const {
    return edge_set(Vertex_edge_it(data()->ife,*this));
}
inline Vertex::face_set Vertex::faces() const {
    return face_set(Vertex_face_it(data()->ife,*this));
}
inline Vertex::cell_set Vertex::cells() const {
    return cell_set(Vertex_cell_it(data()->ife,*this));
}

