#ifndef COLORINGCACHE_H_
#define COLORINGCACHE_H_

#include <vector>

#include "InterferenceGraph/InterferenceBuilder.h"

//provided by code generations
void createColoringMetaData(const ColoringMesh & cpu_mesh, ::Synchronization::InterferenceBuilder * igraph);

namespace Synchronization {

/* 
 * Main controller interface to the Coloring abstraction providing GPU synchronization.
 * This class will manage and cache the coloring results appropriately.
 */
class ColoringController {
public:
	void init(const ColoringMesh & mesh, unsigned int blocksize);
	template<typename MeshSet>
	int getCRSColoring(WorkGroupID wU, const MeshSet & mesh_set, bool isTrivial,  CRS** crs);
	
	InterferenceBuilder* getInterferenceBuilder();
	
private:
	InterferenceBuilder interference_builder;
};


inline void ColoringController::init(const ColoringMesh & mesh, unsigned int blocksize) {
	this->getInterferenceBuilder()->init(blocksize);
	createColoringMetaData(mesh,this->getInterferenceBuilder());
}

inline InterferenceBuilder* ColoringController::getInterferenceBuilder() {
	return &interference_builder;
}

template<typename MeshSet>
inline int ColoringController::getCRSColoring(WorkGroupID wG, const MeshSet & mesh_set, bool isTrivial, CRS** crs) {
	return interference_builder.template getCRSColoring< MeshSet >(wG,mesh_set,isTrivial,crs);
}

} /* namespace Synchronization */

#endif /* COLORINGCACHE_H_ */

