#include "Coloring.h"

#include "MeshIO/FacetEdgeBuilder.h"
#include "MeshIO/LisztFileReader.h"
#include "CRSMesh/CRSMesh.h"
#include "ForAll.h"

using namespace CRSMesh;

int main(int argc, char ** argv) {
	assert(argc == 2);
	MeshIO::LisztFileReader mesh_file;
	MeshIO::FacetEdgeBuilder builder;
	
	mesh_file.init(argv[1]);
	const MeshIO::LisztHeader & h = mesh_file.header();
	MeshIO::FileFacetEdge * ffe = mesh_file.facetEdges();
	builder.init(h.nV,h.nE,h.nF,h.nC,h.nFE);
	builder.insert(0,h.nFE,ffe);
	builder.validateFullMesh();
	mesh_file.free(ffe);
	Mesh mesh;
	mesh.initFromFacetEdgeBuilder(&builder);
	
	Coloring color;
	
	size_t nNodes = h.nV;
	size_t * sizes = new size_t[nNodes + 1];
	sizes[0] = 0;
	FORALL_SET_SER(v,mesh.vertices())
		sizes[v.ID() + 1] = sizes[v.ID()] + v.edges().size(); 
	ENDSET
	size_t * edges = new size_t[sizes[nNodes]];
	FORALL_SET_SER(v,mesh.vertices())
		size_t * cur = edges + sizes[v.ID()];
		FORALL_SET_SER(e,v.edges())
			*cur++ = e.opposite(v).ID();
		ENDSET
	ENDSET
	size_t * colors = new size_t[nNodes];
	color.run(nNodes,sizes,edges,colors);
	
	FORALL_SET_SER(v,mesh.vertices())
		printf("%d: color %ld\n",v.ID(),colors[v.ID()]);
		FORALL_SET_SER(e,v.edges())
			assert(colors[v.ID()] != colors[e.opposite(v).ID()]); 
		ENDSET
	ENDSET
	for(size_t i = 0; i < nNodes; i++) {
		
	}
	return 0;
}