
class SuperLU {
public:
	typedef uint64_t Index;
	
	void init(size_t nrows, size_t ncols, size_t my_partition, size_t npartitions, Index * partitions) {}
	//called once per run,
	//nrows = number of rows in the matrix A, height of B
	//ncols = number of cols in the matrix A, width of X
	//my_partition = if in distributed memory the current partition this part of the matrix lives on (for now zero)
	//npartitions = number of partitions of memory (for now 1)
	//partitions = an array of npartitions+1 size, where the range [ partitions[i], partitions[i+1] ) specifies the part of the rows owned by partition i
	
	void finalize() {}
	
	//non-zero specifying functions: called before the non-zero pattern changes, all previous non-zero patterns, including those applied to 
	//previously allocated arrays should be discarded in favor of the new nonzero pattern
	void beginNonZeroes();
	
	//these entries are only valid between begin and end NonZeroes calls, and specify that a non-zero lives at a location
	
	void entryA(Index row, Index col);
	void entryX(Index index);
	void entryB(Index index);
	
	template<size_t R, size_t C>
	void entryA(const vec<R,Index> & rows, const vec<C,Index> & cols);
	
	template<size_t N>
	void entryX(const vec<N,Index> & indices);
	
	template<size_t N>
	void entryB(const vec<N,Index> & indices);
	
	//called at the end of a non-zero block
	void endNonZeroes();
	
	class Vector_t {
	public:
		double read(Index i);
		template<size_t N>
		vec<N,double> read(const vec<N,Index> idx);
			
		template<typename Operator>
		void write(Index a, double v);
		template<typename Operator, size_t N>
		void write(const vec<N,Index> & indices, vec<N,double> & values);
		
		
		//ensure the state is correct for upcoming operations
		//the possible states are in Fields/ReductionOps.h
		//if nested is true, this means that this ensure state is being called inside a forall loop
		//and can be safely ignored
		template<bool nested, typename T>
		void ensureState();
		
	};
	
	class Matrix_t {
	public:
		double read(Index row, Index col);
		template<size_t R, size_t C>
		matrix<R,C,double> read(const matrix<R,C,size_t> idx);
		
		template<typename Operator>
		void write(Index r, Index c, double v);
		
		template<typename Operator, size_t R, size_t C>
		void write(const vec<R,Index> & rows, 
		           const vec<C,Index> & cols,
		           ::matrix<R,C,double> & values);
		           
		//ensure the state is correct for upcoming operations
		//the possible states are in Fields/ReductionOps.h
		//if nested is true, this means that this ensure state is being called inside a forall loop
		//and can be safely ignored
		template<bool nested, typename T>
		void ensureState();
	};
	
	//types actually used by the compiler
	typedef shared_ptr<Matrix_t> Matrix;
	typedef shared_ptr<Vector_t> Vector;
	
	
	//allocators, note that even after allocation, a call to change non-zero layouts should still
	//affect already allocated objects
	
	Matrix A();
	Vector x();
	Vector b();
	
	
	//invoke the solver for Ax=b
	void solve(Matrix A, Vector x, Vector b);
	
	//configure the solver
	int iters();
	void max_iters(int i);
	double resid();
	void target_resid(double d);

};
