#ifndef _MPI_MESH_FACADE_H
#define _MPI_MESH_FACADE_H

#include "CRSMesh/CRSMesh.h"
#include "ParallelLoad/Common.h"

class MeshFacade : public CRSMesh::Mesh {
public:
	typedef CRSMesh::Mesh parent_mesh;
	typedef SizedMeshSet<parent_mesh::vertex_set::iterator> vertex_set;
	typedef SizedMeshSet<parent_mesh::edge_set::iterator> edge_set;
	typedef SizedMeshSet<parent_mesh::face_set::iterator> face_set;
	typedef SizedMeshSet<parent_mesh::cell_set::iterator> cell_set;
	
	vertex_set vertices() const {
		parent_mesh::vertex_set vs = me()->vertices(0,firstGhost[VERTEX]);
		return vertex_set(vs.iter(),nElems[VERTEX]);
	}
	edge_set edges() const {
		parent_mesh::edge_set vs = me()->edges(0,firstGhost[EDGE]);
		return edge_set(vs.iter(),nElems[EDGE]);
	}
	face_set faces() const {
		parent_mesh::face_set vs = me()->faces(0,firstGhost[FACE]);
		return face_set(vs.iter(),nElems[FACE]);
	}
	cell_set cells() const {
		parent_mesh::cell_set vs = me()->cells(1,firstGhost[CELL]);
		return cell_set(vs.iter(),nElems[CELL] - 1);
	}
	
	void initFacade(size_t * nE, size_t * nO) {
		std::copy(nE, nE + ELEM_TYPE_SIZE, nElems);
		std::copy(nO, nO + ELEM_TYPE_SIZE, firstGhost);
	}
	
	size_t nElems[ELEM_TYPE_SIZE];
	size_t firstGhost[ELEM_TYPE_SIZE];
private:
	const CRSMesh::Mesh * me() const {
		return static_cast<const CRSMesh::Mesh *>(this);
	}
};

#endif