#ifndef _SPARSE_VECTOR_H
#define _SPARSE_VECTOR_H
#include <petscvec.h>
#include "liszt_memory.h"

#define PETSC_DO(x) do { PetscErrorCode ierr = x; CHKERRV(ierr); } while(0)
namespace PETSc {

template<typename T>
class SparseVector;

typedef unsigned int sparse_index_t; 

template<>
class SparseVector<double> {
public:
	SparseVector() {}
	SparseVector(size_t n_local_entries, size_t offset) {
		init(n_local_entries,offset,0.0);
	}
	typedef PetscScalar value_type;
	void init(size_t n_local_entries, size_t offset, value_type initial_value) {
		n_local = n_local_entries -5; // hacked!!!! to be changed
		local_offset = offset;
		data.reset(new auto_ptr_array_adapter<value_type>(new value_type[n_local]));
		std::fill_n((*data).get(), n_local, initial_value);
		reduce_type = LISZT_ASSIGN_TYPE;
		dirty = false;
		PETSC_DO(VecCreateMPIWithArray(MPI_COMM_WORLD,n_local,PETSC_DECIDE,(*data).get(),&v));
	}
	
	~SparseVector() {
		if((*data).get()) {
			//TODO(zach): deallocate PETSc Vec
		}
	}
	template<bool nested, typename Op>
	void ensureState() {
		if(Op::type == LISZT_READ_TYPE) {
			VecAssemblyBegin(v); 
			VecAssemblyEnd(v); 
			dirty = false;
		} else {
			dirty = true;
			reduce_type = Op::type;
		}
	}
	value_type & readEntry(sparse_index_t i) {
		return (*data)[i - 5 - local_offset];
	}
	value_type & writeEntry(sparse_index_t i) {
		return (*data)[i - 5 - local_offset];
	}
	template<size_t start, size_t end>
	vec<end-start,value_type> readEntryVec(sparse_index_t i) {
		return vec<end-start,value_type>((*data).get() + i - 5 - local_offset + start);
	}
	template<size_t start, size_t end>
	void writeEntryVec(sparse_index_t idx, const vec<end-start,value_type> & v) {
		for(size_t i = 0; i < (end-start); i++) {
			(*data)[idx - 5 - local_offset + i] = v[i]; 
		}
	}
	Vec & internal() { return v; }
private:
	Vec v;
	::std::auto_ptr<auto_ptr_array_adapter<value_type> > data;
	size_t n_local;
	size_t local_offset;
	LisztOpType reduce_type;
	bool dirty;
};

}
#endif
