#ifndef _MATH_HELPER_FUNCTIONS
#define _MATH_HELPER_FUNCTIONS

#ifndef MATH_HELPER_PREFIX
#error "Before including math helpers you must define MATH_HELPER_PREFIX"
#endif


template<size_t N, typename scalar>
MATH_HELPER_PREFIX static inline scalar length(const vec<N,scalar> & v) {
    return sqrt(v.dot(v));
}
template<size_t N, typename scalar>
MATH_HELPER_PREFIX static inline vec<N,scalar> normalize(const vec<N,scalar> & v) {
    return v / length(v);
}
template<size_t N, typename scalar>
MATH_HELPER_PREFIX static inline scalar dot(const vec<N,scalar> & l,const vec<N,scalar> & r) {
    return l.dot(r);
}

template<typename scalar>
MATH_HELPER_PREFIX static inline vec<3,scalar> cross(const vec<3,scalar> & l, const vec<3,scalar> & r) {
    return l.cross(r);
}


template<size_t R, size_t C, typename scalar>
MATH_HELPER_PREFIX static inline matrix<R,C,scalar> outer(const vec<R,scalar> & l, const vec<C,scalar> & r) {
    matrix<R,C,scalar> ret;
    for(int i = 0; i < R; i++)
        for(int j = 0; j < C; j++)
                ret(i,j) = l[i] * r[j];
        return ret;
}

template<typename T, size_t R, size_t C>
MATH_HELPER_PREFIX static inline vec<C,T> row(const matrix<R,C,T> & m, int r) {
        assert(r < R);
        return vec<C,T>(&m.v[r][0]);
}

template<typename T, size_t R, size_t C>
MATH_HELPER_PREFIX static inline vec<R,T> col(const matrix<R,C,T> & m, int c) {
        assert(c < C);
        vec<R,T> v;
        for(size_t i = 0; i < R; i++)
                v[i] = m.v[i][c];
        return v;
}

MATH_HELPER_PREFIX static inline double MATH_PI() { return M_PI; }
MATH_HELPER_PREFIX static inline float MIN_FLOAT() { return FLT_MIN; }
MATH_HELPER_PREFIX static inline float MAX_FLOAT() { return FLT_MAX; }
MATH_HELPER_PREFIX static inline double MIN_DOUBLE() { return DBL_MIN; }
MATH_HELPER_PREFIX static inline double MAX_DOUBLE() { return DBL_MAX; }

#ifdef _SYRAH_FIXED_VECTOR_H_
// NOTE(boulos): You can't use (type) here for the template arguments below.
  #define MAKE_CONVERT(name,type) \
  template<typename T> \
  syrah::FixedVector<type, VecWidth> name(const syrah::FixedVector<T, VecWidth>& other) { return syrah::FixedVector<type, VecWidth>(other); } \
  template<typename T> \
  MATH_HELPER_PREFIX type name(const T & t) { return (type) t; }
#else
  #define MAKE_CONVERT(name,type) \
  template<typename T> \
  MATH_HELPER_PREFIX type name(const T & t) { return (type) t; }
#endif

MAKE_CONVERT(toByte,unsigned char)
MAKE_CONVERT(toChar,char)
MAKE_CONVERT(toDouble,double)
MAKE_CONVERT(toFloat,float)
MAKE_CONVERT(toInt,int)
MAKE_CONVERT(toLong,long)
MAKE_CONVERT(toShort, short int)

#undef MAKE_CONVERT



#endif
