#ifndef _PARALLEL_LOAD_FIELD_WRITER_H
#define _PARALLEL_LOAD_FIELD_WRITER_H
#include <mpi.h>
#include <sstream>
#include "MeshIO/LisztFormat.h"

template<typename T>
class MPIFieldWriter {
public:
	void init(const char * name, MeshIO::IOElemType domain, size_t nElems) {
		MPI_Comm_rank(MPI_COMM_WORLD,&mpi_rank);
		MPI_Comm_size(MPI_COMM_WORLD,&mpi_size);

		MeshIO::LisztType range = GetLisztType<T>::type();
		
		field.domain = domain;
		field.range = range;
		field.nElems = nElems;
		
		std::ostringstream file_name_ss;
		file_name_ss << name << ".lfield";
		std::string file_name = file_name_ss.str();
		
		MPI_File_open(MPI_COMM_WORLD,(char*)file_name.c_str(),MPI_MODE_WRONLY | MPI_MODE_CREATE,MPI_INFO_NULL,&fh);
		MPI_Status status;
		field.name = sizeof(MeshIO::FileField);
		field.data = sizeof(MeshIO::FileField) + strlen(name) + 1;
		
		if(mpi_rank == 0) {
			MPI_File_write(fh,&field,sizeof(MeshIO::FileField),MPI_CHAR,&status);
			MPI_File_seek(fh,field.name,MPI_SEEK_SET);
			MPI_File_write(fh,(void*)name,strlen(name)+1,MPI_CHAR,&status);
		}
	}
	void write(id_type idx, const T & moo) {
		MPI_File_seek(fh, field.data + sizeof(T) * idx, MPI_SEEK_SET);
		MPI_Status status;
		MPI_File_write(fh,(void*) &moo,sizeof(T),MPI_CHAR,&status);
	}
	void close() {
		MPI_File_close(&fh);
	}
private:
	MeshIO::FileField field;
	MPI_File fh;
	int mpi_rank, mpi_size;
};

#endif