#include "MeshIO/LisztFileReader.h"
#include "ParallelLoad/Common.h"
#include "ParallelLoad/NameService.h"
#include "ParallelLoad/TopologyService.h"
#include "ParallelLoad/OwnerService.h"
#include "ParallelLoad/LocalTopology.h"
#include "ParallelLoad/LocalMemory.h"
#include "ParallelLoad/Stencil.h"
#include "CRSMesh/CRSMesh.h"
#include "Forall.h"
#include <sstream>

int main(int argc, char ** argv) {
	assert(argc == 2);
	MPI_Init(&argc,&argv);
    
    std::ostringstream ss;
    int mpi_rank;
    MPI_Comm_rank(MPI_COMM_WORLD,&mpi_rank);
    ss << "log" << "_" << mpi_rank << ".txt";
    int file = fileno(fopen(ss.str().c_str(),"a"));
    dup2(file,1);
    dup2(file,2);
    
	shared_ptr<MeshIO::LisztFileReader> reader(new MeshIO::LisztFileReader());
	reader->init(argv[1]);
	shared_ptr<NameService> name_service(new NameService(&reader->header()));
	shared_ptr<TopologyService> topology_service(new TopologyService(name_service, reader));
	shared_ptr<OwnerService> owner_service( new OwnerService(name_service,topology_service->partitionedTopology()));
	shared_ptr<LocalTopology> local_topology(new LocalTopology(name_service->self(),owner_service->locallyOwnedElements()));
	shared_ptr<LocalMemory> local_memory(new LocalMemory());
	shared_ptr<Stencil> stencil(new Stencil(name_service,topology_service,owner_service,local_topology,local_memory));
	stencil->discoverLocalTopology();
	stencil->discoverLocalMemory();
	
	local_topology->print_debug(name_service);
	CRSMesh::Mesh crmesh;
	local_topology->buildLocalMesh(&crmesh);
	
	
	printf("size = %d\n",(int)crmesh.cells().size());
	FORALL_SET_SER(c,crmesh.cells())
		printf("c = %d\n",(int)c.ID());
	ENDSET
	
	MPI_Finalize();
}