#ifndef _SMP_FIELD_H
#define _SMP_FIELD_H
#include <pthread.h>
#include "liszt_memory.h"
#include "Array.h"

#include "SMP/Synchronization.h"
#include "Layout/Util.h"
#include "Fields/ReductionOps.h"

#include "cache_sim.h"

#ifdef _SYRAH_FIXED_VECTOR_H_
#include "../Vector.h"
#include "../Matrix.h"
#include "../SyrahVector.h"
#include "../SyrahMatrix.h"
#include "../SyrahObjectWrapper.h"
#endif

#if 0
// NOTE(boulos): Most fields are floats and if not floats then
// Vec3f. A value of 6 here means we're going to new up blocks of only
// 64 elements meaning 256 bytes.
const size_t LOG_BLOCK_SIZE = 6;
#else
// This one is at least 1024 elements, so 4k (page size)
#if 0
const size_t LOG_BLOCK_SIZE = 10;
#else
// How about more like 64k elems per block (so 256KB for float field)
const size_t LOG_BLOCK_SIZE = 16;
#endif
#endif
const size_t BLOCK_SIZE = 1u << LOG_BLOCK_SIZE;
const size_t BLOCK_SIZE_MASK = BLOCK_SIZE - 1;

//#define SAMPLE_USAGE

#ifdef SAMPLE_USAGE
const int SAMPLE_INTERVAL = 1000;
#endif
//non-templated information goes in the base class
//this makes it easy to keep a list of all fields
class FieldBase {
public:
        FieldBase() {
                initialized = false;
                state = LISZT_READ_TYPE;
                pthread_mutex_init(&field_l,NULL);

                #ifdef SAMPLE_USAGE
                save_count = 0;
                sample_count = 0;
                #endif
        }

  static SimpleCache* cache_sim;

#define CACHE_SIM 0

  static inline void cacheSim(const void* addr) {
#if CACHE_SIM
     if (cache_sim) {
        cache_sim->fetch(addr);
     }
#endif
  }

        void init(size_t s, const std::string & n, MeshIO::IOElemType t) {
                name_ = n;
                size = s;
                field_type = t;

                #ifdef SAMPLE_USAGE
                usage_count.reset(size);
                std::fill_n(usage_count.get(),size,0);
                #endif
        }

        bool ensureState(LisztOpType new_state,liszt_barrier_t * b) {
                /*if(state != new_state) {

                        if(barrier_wait(b)) {
                                state = new_state;
                        }

                        return true;
                } else return false; */

          // NOTE(boulos): Don't let C++ have a default here...
          return true;
        }

        //protects the entire field, used in initialization
        void lock_all() {
                pthread_mutex_lock(&field_l);
        }
        void unlock_all() {
                pthread_mutex_unlock(&field_l);
        }

        const std::string & name() const { return name_; }

        #ifdef SAMPLE_USAGE
        void save() const {
                char num[16];
                sprintf(num,"%d",save_count);
                std::ostringstream ss;
                ss << name() << "_" << num;
                writeField<int>(field_type, usage_count.get(), sizeof(int), size, ss.str().c_str());
                save_count++;
        }
        #endif //SAMPLE_USAGE

        void touch(size_t i) const {
                #ifdef SAMPLE_USAGE
                usage_count[i]++;
                sample_count++;
                if(sample_count == SAMPLE_INTERVAL) {
                        sample_count = 0;
                        save();
                }
                #endif
        }
        ~FieldBase() {
                #ifdef SAMPLE_USAGE
                save();
                #endif
        }
protected:
        size_t size;
        std::string name_;
        bool initialized;
        LisztOpType state;
        pthread_mutex_t field_l;

        MeshIO::IOElemType field_type;

        #ifdef SAMPLE_USAGE
        mutable Array<int> usage_count;
        mutable size_t save_count;
        mutable size_t sample_count;
        #endif
};

#if !(defined(_SYRAH_FIXED_VECTOR_H_))
//templated information goes here
template<typename MO, typename VT>
class Field : public FieldBase {
public:
        typedef VT value_type;
        typedef MO mesh_elem_type;
        bool init(size_t sz, const std::string & n) {
                if(initialized)
                        return false;
                FieldBase::init(sz,n, ElemTypeToIO(mesh_elem_type::value));
                size = sz;
                array.reset(size);
                initialized = true;
                return true;
        }
        void init(size_t sz, const std::string & n, const value_type & v) {
                init(sz,n);
                for(size_t i = 0; i < sz; i++)
                        internal(i) = v;
        }
        template<typename ME>
        const value_type & operator()(const ME & mo) const {
                assert(MO::value == ME::type::value);
                assert(mo.ID() < size);
                touch(mo.ID());
                return array[mo.ID()];
        }
        template<typename ME>
        value_type & operator[](const ME & mo) {
                assert(MO::value == ME::type::value);
                assert(mo.ID() < size);
                touch(mo.ID());
                return internal(mo.ID());
        }
        value_type & internal(size_t t) {
                assert(t < size);
                return array[t];
        }

        template<typename ME>
        void lock(const ME & mo) {
                //size_t blk = mo.ID() >> LOG_BLOCK_SIZE;
                //lock_lock(&array[blk].lock);
        }
        template<typename ME>
        void unlock(const ME & mo) {
                //size_t blk = mo.ID() >> LOG_BLOCK_SIZE;
                //lock_unlock(&array[blk].lock);
        }
private:
        Array<value_type> array;
};
#else
#include "SyrahField.h"
#endif

#endif
