#ifndef _MESH_WRAPPER_H
#define _MESH_WRAPPER_H
//provides a wrapper around a mesh to give a local view of the mesh elements
//we expect that the type GlobalMesh refers to the type of the overall mesh when this file is included

#include "Liszt/ElemTypes.h"
using namespace ::LisztPrivate::ElemTypes;

static inline void Util_getEvenDistribution(size_t total_size, size_t nparts, size_t i, size_t * start, size_t * psize) {
	size_t size = total_size / nparts;
	size_t rem = total_size % nparts;
	*start = size * i + std::min(rem,i);
	*psize = size + ( (i < rem) ? 1 : 0 );
}

class MeshWrapper {
public:
	MeshWrapper() {}
	void init(const GlobalMesh * m, size_t num_threads, size_t thread_id) {
		mesh = m;
		Util_getEvenDistribution(m->vertices().size(),num_threads,thread_id,&offsets[VERTEX],&sizes[VERTEX]);
		Util_getEvenDistribution(m->edges().size(),num_threads,thread_id,&offsets[EDGE],&sizes[EDGE]);
		Util_getEvenDistribution(m->faces().size(),num_threads,thread_id,&offsets[FACE],&sizes[FACE]);
		Util_getEvenDistribution(m->cells().size(),num_threads,thread_id,&offsets[CELL],&sizes[CELL]);
		offsets[CELL] += 1; //ignore the zero cell
	}
	typedef GlobalMesh::vertex_set vertex_set;
	typedef GlobalMesh::edge_set edge_set;
	typedef GlobalMesh::face_set face_set;
	typedef GlobalMesh::cell_set cell_set;
	
	vertex_set vertices() const {
		return mesh->vertices(offsets[VERTEX],offsets[VERTEX] + sizes[VERTEX]);
	}
	edge_set edges() const {
		return mesh->edges(offsets[EDGE],offsets[EDGE] + sizes[EDGE]);
	}
	face_set faces() const {
		return mesh->faces(offsets[FACE],offsets[FACE] + sizes[FACE]);
	}
	cell_set cells() const {
		return mesh->cells(offsets[CELL],offsets[CELL] + sizes[CELL]);
	}
	
private:
	const GlobalMesh * mesh;
	size_t offsets[ELEM_TYPE_SIZE];
	size_t sizes[ELEM_TYPE_SIZE];
};

#endif