#ifndef _SMP_SYNCHRONIZATION_H
#define _SMP_SYNCHRONIZATION_H
#include <xmmintrin.h>
#include "Liszt/CycleTimer.h"

//spin barriers

typedef struct {
	int num_threads_minus_one;
	volatile int state;
	volatile int lock;
} liszt_barrier_t;

static inline void barrier_init(liszt_barrier_t * b, short num_threads) {
	b->num_threads_minus_one = num_threads - 1;
	b->state = 0;
	b->lock = 0;
	__sync_synchronize();
}

#if 0

static inline bool barrier_wait(liszt_barrier_t * b, bool dostuff = true) {
	if(dostuff) {
		volatile int * v = &b->lock;
		volatile int * s = &b->state;
		__sync_synchronize();
		int current_state = *s;
		int num_waiting_except_me = __sync_fetch_and_add(v,1);
		if(num_waiting_except_me == b->num_threads_minus_one) {
			*v = 0;
			*s = !current_state;
			return true;
		} else {
			while(current_state == *s)
				_mm_pause();
			return false;
		}
	} else {
		vtal::CycleTimer timer;
		double begin = timer.currentSeconds();
		while(timer.currentSeconds() - begin < 1e-6 * 3) {
			_mm_pause();
		}
	}
}


#else


//begin barrier acquisition, this will return true if you are the elected leader of the barrier
//the leader must call barrier_end to release the other threads, but may do things before they are released

//idiom:
//if(barrier_begin(b)) {
// <work before releaseing barrier>
//  barrier_end(b);
//}
static inline void wait_8us() {
	vtal::CycleTimer timer;
	double begin = timer.currentSeconds();
	while(timer.currentSeconds() - begin < 8 * 1e-6) {}
}
static inline void wait_us(double us) {
	vtal::CycleTimer timer;
	double begin = timer.currentSeconds();
	while(timer.currentSeconds() - begin < us * 1e-6) {}
}

static inline bool barrier_begin(liszt_barrier_t * b) {
	volatile int * v = &b->lock;
	volatile int * s = &b->state;
	__sync_synchronize();
	int current_state = *s;
	int num_waiting_except_me = __sync_fetch_and_add(v,1);
	if(num_waiting_except_me == b->num_threads_minus_one) {
		return true;
	} else {
                double wait_time = 8.0;
		while(current_state == *s) wait_us(wait_time *= 1.1);
		return false;
	}
}

static inline bool barrier_end(liszt_barrier_t * b) {
	volatile int * v = &b->lock;
	volatile int * s = &b->state;
	*v = 0;
	__sync_synchronize(); //flush writes between barrier_begin adn barrier_end
	*s = !*s;
}

static inline bool barrier_wait(liszt_barrier_t * b) {
	bool r = barrier_begin(b);
	if(r) barrier_end(b);
	return r;
}

#endif
//spin locks

typedef int liszt_lock_t;

static inline void lock_init(volatile liszt_lock_t *  lck) {
	*lck = 0;
	__sync_synchronize();
}

void lock_lock(volatile liszt_lock_t * lck) {
	while (__sync_lock_test_and_set(lck, 1))
		while (*lck)
			_mm_pause();
}

static inline void lock_unlock(volatile liszt_lock_t * lck) {
	__sync_synchronize();
	*lck = 0;
}


#endif
