/*
 * ScalaStdLib.h
 *
 *  Created on: Sep 2, 2009
 *      Author: zdevito
 */

#ifndef SCALASTDLIB_H_
#define SCALASTDLIB_H_
//This file includes C++ implementations that are used frequently in the scala code generators across
//all the runtimes that use C++ as a base language.

//Code for specific runtimes needs to go in its own header file
//Code that uses common features of the old c++ runtime (descriptors, Liszt object, fluent-file)
//Should go in ScalaCppGlue.h, not here

#define LISZT_EXTERN(ret,name) \
typedef ret name##_return_type; \
static inline ret name

//Runtime-specific functions
//These are left undefined in this header and must be provide elsewhere for correct operation
static inline const Mesh * mesh();
static inline id_type ID_v(const Vertex & v);
static inline id_type ID_e(const Edge & e);
static inline id_type ID_f(const Face & f);
static inline id_type ID_c(const Cell & c);

#ifdef id
static inline id_type __ID(const Vertex & v) {
    return ID_v(v);
}
static inline id_type __ID(const Edge & e) {
    return ID_e(e);
}
static inline id_type __ID(const Face & f) {
    return ID_f(f);
}
static inline id_type __ID(const Cell & c) {
    return ID_c(c);
}
#else
static inline id_type ID(const Vertex & v) {
    return ID_v(v);
}
static inline id_type ID(const Edge & e) {
    return ID_e(e);
}
static inline id_type ID(const Face & f) {
    return ID_f(f);
}
static inline id_type ID(const Cell & c) {
    return ID_c(c);
}
#endif


LISZT_EXTERN(mesh_type::vertex_set,vertices_m)(const mesh_type * m) {
    return m->vertices();
}

LISZT_EXTERN(Vertex::vertex_set,vertices_v)(const Vertex & m) {
    return m.vertices();
}

LISZT_EXTERN(Edge::vertex_set,vertices_e)(const Edge & m) {
    return m.vertices();
}

LISZT_EXTERN(Face::vertex_set,vertices_f)(const Face & m) {
    return m.vertices();
}


LISZT_EXTERN(Face::vertex_set,verticesCCW_f)(const Face & m) {
    return m.verticesCCW();
}


LISZT_EXTERN(Face::vertex_set,verticesCW_f)(const Face & m) {
    return m.verticesCW();
}

LISZT_EXTERN(Cell::vertex_set,vertices_c)(const Cell & m) {
    return m.vertices();
}

LISZT_EXTERN(mesh_type::cell_set,cells_m)(const mesh_type * m) {
    return m->cells();
}

LISZT_EXTERN(Vertex::cell_set,cells_v)(const Vertex & m) {
    return m.cells();
}

LISZT_EXTERN(Edge::cell_set,cells_e)(const Edge & m) {
    return m.cells();
}

LISZT_EXTERN(Edge::cell_set,cellsCCW_e)(const Edge & m) {
    return m.cellsCCW();
}

LISZT_EXTERN(Edge::cell_set,cellsCW_e)(const Edge & m) {
    return m.cellsCW();
}

LISZT_EXTERN(Face::cell_set,cells_f)(const Face & m) {
    return m.cells();
}

LISZT_EXTERN(Cell::cell_set,cells_c)(const Cell & m) {
    return m.cells();
}

LISZT_EXTERN(mesh_type::edge_set,edges_m)(const mesh_type * m) {
    return m->edges();
}

LISZT_EXTERN(Vertex::edge_set,edges_v)(const Vertex & m) {
    return m.edges();
}

LISZT_EXTERN(Face::edge_set,edges_f)(const Face & m) {
    return m.edges();
}

LISZT_EXTERN(Cell::edge_set,edges_c)(const Cell & m) {
    return m.edges();
}

LISZT_EXTERN(Face::edge_set,edgesCCW_f)(const Face & m) {
    return m.edgesCCW();
}

LISZT_EXTERN(Face::edge_set,edgesCW_f)(const Face & m) {
    return m.edgesCW();
}

LISZT_EXTERN(mesh_type::face_set,faces_m)(const mesh_type * m) {
    return m->faces();
}

LISZT_EXTERN(Vertex::face_set,faces_v)(const Vertex & m) {
    return m.faces();
}

LISZT_EXTERN(Edge::face_set,faces_e)(const Edge & m) {
    return m.faces();
}

LISZT_EXTERN(Cell::face_set,faces_c)(const Cell & m) {
    return m.faces();
}

LISZT_EXTERN(Edge::face_set,facesCCW_e)(const Edge & m) {
    return m.facesCCW();
}

LISZT_EXTERN(Edge::face_set,facesCW_e)(const Edge & m) {
    return m.facesCW();
}

//Extract by label functions
#include "LabelExtractXMacro.h"
template<typename SetType>
static inline typename SetType::iterator::value_type label_extract(const SetType & st, int lbl) {
        assert(lbl >= 0 && (unsigned int) lbl < st.size());
        return st.iter()[lbl];
}

#define LABEL_EXTRACT_TEMPLATE(fn,tag,r_type,a_type,plural_fn) \
static inline r_type fn##_##tag(const a_type & e, int lbl) { \
        return label_extract(e.plural_fn(),lbl); \
}

LABEL_EXTRACTORS(LABEL_EXTRACT_TEMPLATE)
#undef LABEL_EXTRACT_TEMPLATE


#define LABEL_EXTRACT_CONST_TEMPLATE(fn,tag,r_type,a_type,zero_fn,one_fn) \
static inline r_type fn##_##tag(const a_type & e, int lbl) { \
        if(lbl == 0) return e.zero_fn(); \
        else return e.one_fn(); \
}

LABEL_EXTRACTORS_CONST(LABEL_EXTRACT_CONST_TEMPLATE)
#undef LABEL_EXTRACT_CONST_TEMPLATE

template<typename Set>
static inline size_t size_v(const Set & s) {
    return s.size();
}

template<typename Set>
static inline size_t size(const Set & s) {
    return s.size();
}


template<typename Set>
static inline size_t size_e(const Set & s) {
    return s.size();
}

template<typename Set>
static inline size_t size_f(const Set & s) {
    return s.size();
}

template<typename Set>
static inline size_t size_c(const Set & s) {
    return s.size();
}

#ifdef _SYRAH_FIXED_VECTOR_H_
template<size_t SIMD_WIDTH, typename SetType>
static inline syrah::FixedVector<int, (int)SIMD_WIDTH> size(const syrah::SIMDObject<SIMD_WIDTH, SetType>& s) {
  syrah::FixedVector<int, (int)SIMD_WIDTH> result;
  for (size_t i = 0; i < SIMD_WIDTH; i++) result[i] = s[i].size();
  return result;
}
#endif

static inline Vertex head(const Edge & e) {
    return e.head();
}

static inline Vertex tail(const Edge & e) {
    return e.tail();
}

static inline Cell outside(const Face & f) {
    return f.outside();
}

static inline Cell inside(const Face & f) {
    return f.inside();
}

static inline Face flip_f(const Face & f) {
    Face f2 = f;
    f2.flip();
    return f2;
}

static inline Edge flip_e(const Edge & e) {
    Edge e2 = e;
    e2.flip();
    return e2;
}

static inline Face towards_f(const Face & f, const Cell & c) {
    Face f2 = f;
    f2.towards(c);
    return f2;
}

static inline Edge towards_e(const Edge & e, const Vertex & v) {
    Edge e2 = e;
    e2.towards(v);
    return e2;
}

static inline std::ostream & operator<<(std::ostream & out, const Vertex & v) {
        return out << "Vertex " << ID_v(v);
}
static inline std::ostream & operator<<(std::ostream & out, const Edge & v) {
        return out << "Edge " << ID_e(v);
}
static inline std::ostream & operator<<(std::ostream & out, const Face & v) {
        return out << "Face " << ID_f(v);
}
static inline std::ostream & operator<<(std::ostream & out, const Cell & v) {
        return out << "Cell " << ID_c(v);
}

#define MATH_HELPER_PREFIX
#include "MathHelperFunctions.h"


#undef LISZT_EXTERN


#endif /* SCALASTDLIB_H_ */
