/*
 * Field.h
 *
 *  Created on: Feb 19, 2009
 *      Author: zdevito
 */

#ifndef SINGLEFIELD_H_
#define SINGLEFIELD_H_
#include <vector>
#include <iostream>
#include <string>

#include "Array.h"
#include "liszt_memory.h"
#include "SetAdaptor.h"

#include "Fields/FieldValueReference.h"

#include "Liszt/ElemTypes.h"
#include "MeshIO/LisztFormat.h"
#include "MeshIO/Common.h"

using std::vector;
using namespace LisztPrivate::ElemTypes;
using namespace std;

namespace SingleField {

//for use in initializer which needs to translate global IDs to local id
//for whatever field implementation is in use
template<typename T>
class GlobalToLocalIdentityFunc {
public:
    typedef T elem_type;
    GlobalToLocalIdentityFunc() {}
    LisztPrivate::id_type operator()(const LisztPrivate::ElemRef<elem_type> & e) const {
        return e.ID();
    }
    LisztPrivate::id_type operator()(LisztPrivate::id_type id) const {
        return id;
    }
};

template<typename E, typename vt> class Field {
public:
    typedef Field<E,vt> this_type;
	typedef vt value_type;
	typedef E elem_type;
	typedef GlobalToLocalIdentityFunc<elem_type> global_to_local_type;
	typedef ::FieldCommon::FieldValueReference<this_type> reference;
	Field(const std::string & name__,vt init, size_t size)
	: dirty(false), name_(name__) {
	    data.array.reset(size);
	    data.size = size;
	    for(size_t i = 0; i < data.size; i++)
	        data.array[i] = init;
	}
public: Field() : dirty(false), name_("undefined") {
        data.size = 0;
        data.array.reset(0);
    }
	void init(const std::string & name__, size_t size) {
		name_ = name__;
		data.array.reset(size);
		data.size = size;
		dirty = true;
	}
    template<typename MO>
	const vt & operator()(const MO & mo) const {
		return data.array[mo.ID()];
	}
    template<typename MO>
    const vt & operator()(const MO & mo) {
        return (*(const this_type*)this)(mo);
    }
    template<typename MO>
	reference operator[](const MO & mo) {
        id_type id = mo.ID();
	    return reference(*this,id);
	}

	const std::string & name() const {
		return name_;
	}
	template<typename Op>
	void update() {}
	template<typename Op>
	void setReductionOp() {}
	template<bool nested, typename Op>
	void ensureState() {}
	value_type * getReadEntry(id_type id) {
	    return &data.array[id];
	}
    const value_type * getReadEntry(id_type id) const {
        return &data.array[id];
    }
    size_t getSize() const {
        return data.size;
    }
    value_type * getWriteEntry(id_type id) {
    	dirty = true;
        return &data.array[id];
    }
    const value_type * getWriteEntry(id_type id) const {
        return &data.array[id];
    }
    bool isDirty() const {
    	return dirty;
    }
    void clean() const {
    	dirty = false;
    }
    this_type & operator=(const this_type & rhs) {
        dirty = rhs.dirty;
        name_ = rhs.name_;
        data.size = rhs.data.size;
        if(rhs.data.array.get()) {
            data.array.reset(data.size);
            for(size_t i = 0; i < data.size; i++)
                data.array[i] = rhs.data.array[i];
        }
        return *this;
    } 
    Field(const this_type & rhs) {
        *this = rhs;
    }
    //used to visualize fields directly from a file
    //currently does not do any sanity checking that the file
    //data fits the current field size, or even is the write type
    void loadFromFieldFile(const char * field_file) {
    	FILE * file = FOPEN(field_file,"r");
    	assert(file);
    	MeshIO::FileField f;
    	fread(&f,sizeof(MeshIO::FileField),1,file);
    	fseeko(file,f.data,SEEK_SET);
    	fread(data.array.get(),sizeof(value_type),data.size,file);
    	fclose(file);
    	dirty = true;
    }
    size_t nReadEntries() const {
    	return data.size;
    }
private:
    mutable bool dirty;
    std::string name_;
    struct {
        size_t size;
        Array<value_type> array;
    } data;
};

} // namespace LocalField

#endif /* FIELD_H_ */
