/*
 * Global.h
 *
 *  Created on: Aug 4, 2009
 *      Author: zdevito
 */

#ifndef SINGLEGLOBAL_H_
#define SINGLEGLOBAL_H_

#include "Vector.h"
#include "Matrix.h"
#include "Fields/ReductionOps.h"

namespace LisztSinglePrivate {

template<typename T>
struct Updater;

template<typename T, bool nested>
class Global {
public:
    typedef T value_type;
    typedef Global<T,nested> this_type;
    Global() {
        reduction_type = LISZT_ASSIGN::type;
    }
    Global(const value_type & v) {
    	rvalue = v;
    }
#define FORWARD_OP(op) \
        this_type & operator op(const value_type & rhs) { \
            value() op rhs; \
            return *this;\
        } \
        this_type & operator op(const this_type & r) { \
            return *this op (const value_type &) r; \
        }
        FORWARD_OP(+=)
        FORWARD_OP(-=)
        FORWARD_OP(*=)
        FORWARD_OP(/=)
        FORWARD_OP(&=)
        FORWARD_OP(|=)
        FORWARD_OP(^=)

#undef FORWARD_OP
        this_type & operator=(const value_type & rhs) { \
            rvalue = rhs;
            return *this;
        }
        this_type & operator=(const this_type & r) {
            return *this = (const value_type &) r;
        }
        this_type & minWith(const value_type & rhs) {
            value() = ReductionImpl<value_type>::min(value(),rhs);
            return *this;
        }
        this_type & minWith(const this_type & r) {
            return minWith( (const value_type &) r);
        }

        this_type & maxWith(const value_type & rhs) {
            value() = ReductionImpl<value_type>::max(value(),rhs);
            return *this;
        }
        this_type & maxWith(const this_type & r) {
            return maxWith( (const value_type &) r);
        }

        this_type & operator++(int) {
            value()++;
            return *this;
        }
        this_type & operator--(int) {
            value()--;
            return *this;
        }
        this_type & operator++() {
            ++value();
            return *this;
        }
        this_type & operator--() {
            --value();
            return *this;
        }
        const value_type operator-() const {
            return - (const value_type &) (*this);
        }
        operator const value_type &() const {
            return value();
        }
        template<typename Op>
        void setReductionOp() {
            rvalue = Op::template identity<value_type>();
            reduction_type = Op::type;
        }
        template<typename Op>
        void update() {}
        
        template<bool nested2, typename Op>
        void ensureState() {}
        
        value_type & readEntry() {
        	return rvalue;
        }
        value_type & writeEntry() {
        	return rvalue;
        }
private:
    const T & value() const {
        return rvalue;
    }
    T & value() {
        return rvalue;
    }
    T rvalue;
    LisztOpType reduction_type;
};

} // namespace
#endif /* GLOBAL_H_ */
