/*
 * MeshTopology.h
 *
 *  Created on: Aug 17, 2009
 *      Author: zdevito
 */

#ifndef SINGLEMESHTOPOLOGY_H_
#define SINGLEMESHTOPOLOGY_H_

#include "Vector.h"
#include "FileMesh/FileMesh.h"

namespace LisztPrivate {
    class LisztSingle;
}

namespace LisztSinglePrivate {
template<typename T>
class MeshStore;

class MeshTopology {
public:
    MeshTopology() : fluent(NULL), fmesh(NULL) {}
    MeshTopology(FileMesh::FluentFile & ffile) {
        fluent = &ffile;
        fluent->init();
        fmesh = NULL;
    }
    MeshTopology(const FileMesh::Mesh & fmesh_) {
        fmesh =  &fmesh_;
        fluent = NULL;
    }
    void init(FileMesh::FluentFile & ffile) {
        fluent = &ffile;
    }
private:
    template<typename Mesh>
    void initMesh(Mesh * mesh) {
        if(fluent) {
            fluent->template load<Mesh>(mesh);
        } else {
            assert(fmesh);
            mesh->init(*fmesh);
        }
    }

    template<typename T>
    friend class MeshStore;

    FileMesh::FluentFile * fluent;
    const FileMesh::Mesh * fmesh;
};

template<typename T>
class MeshStore {
public:
    typedef T mesh_type;

    const mesh_type & mesh() const {
        return mesh_;
    }
    //the ghost mesh is for debugging purposes only and should
    //not be relied upon for actual simulation code
    const mesh_type & ghostMesh() const {
        return mesh_;
    }
    const mesh_type & internal_mesh() const {
        return mesh_;
    }
    // TODO(mbarrien): HACK HACK HACK!
    ElemRef<CellType> outsideCellRef() const {
        return ElemRef<CellType>(0);
    }
private:
    friend class LisztSingle;
    void init(MeshTopology * topology) {
       topology->initMesh(&mesh_);

    }
private:
    mesh_type mesh_;
};

} // namespace LisztOpenMPPrivate
#endif /* MESHTOPOLOGY_H_ */
