/*
 * WatchPoint.h
 *
 *  Created on: Nov 24, 2009
 *      Author: zd
 */

#ifndef WATCHPOINT_H_
#define WATCHPOINT_H_

#include "SingleCore/IObject.h"
#include "SingleCore/interpret.h"

class WatchPoint {
public:
	WatchPoint(const std::string & n, const std::string & sexp)
	: name_(n), exp_string(sexp), watch_exp(interpret_compile(sexp)){}
	const std::string & name() const {
		return name_;
	}
	//returns true if the value changed, if oldv or newv is not null it is filled with
	//the value
	bool eval(const Exp::ctx & context, IObject * oldv, IObject * newv) {
		IObject v = watch_exp->eval(context);
		if(oldv)
			*oldv = prev_value;
		if(newv)
			*newv = v;
		bool changed = (v != prev_value).intValue();
		std::cout << changed << std::endl;
		if(v.isVec()) //force a copy of the vec just in case it is backed by a field that may change
		    prev_value = IObject(v.vecValue());
		else
		    prev_value = v;
		return changed;
	}
	const std::string & exp() const {
		return exp_string;
	}
	void setName(const std::string & n) {
		name_ = n;
	}
	void setExp(const std::string & e) {
		exp_string = e;
		watch_exp = interpret_compile(exp_string);
	}
	void reset() {
		prev_value = IObject();
	}
private:
	std::string name_;
	std::string exp_string;
	ExpRef watch_exp;
	IObject prev_value;
};

#endif /* WATCHPOINT_H_ */
