#ifndef _SYRAH_OBJECT_WRAPPER_H_
#define _SYRAH_OBJECT_WRAPPER_H_

#include <iostream>
#include "/Users/boulos/projects/syrah/src/include/syrah/FixedVector.h"

namespace syrah {
  template<size_t N, typename ElemType>
  class SIMDObject {
  public:
    SYRAH_FORCEINLINE SIMDObject() {}
    SYRAH_FORCEINLINE SIMDObject(const SIMDObject& other) {
      for (size_t i = 0; i < N; i++) data[i] = other.data[i];
    }
    SYRAH_FORCEINLINE SIMDObject& operator=(const SIMDObject& other) {
      for (size_t i = 0; i < N; i++) data[i] = other.data[i];
      return *this;
    }
    SYRAH_FORCEINLINE const ElemType& operator[](int i) const { return data[i]; }
    SYRAH_FORCEINLINE ElemType& operator[](int i) { return data[i]; }

    SYRAH_FORCEINLINE void merge(const SIMDObject& other, const FixedVectorMask<N>& mask) {
      for (size_t i = 0; i < N; i++) {
        if (mask.get(i)) {
          data[i] = other[i];
        }
      }
    }

    ElemType data[N];
  };

  template<size_t N, typename ElemType>
  SYRAH_FORCEINLINE SIMDObject<N, ElemType> select(const SIMDObject<N, ElemType>& a, const SIMDObject<N, ElemType>& b, const FixedVectorMask<(int)N>& mask) {
    SIMDObject<N, ElemType> result;
    for (size_t i = 0; i < N; i++) result.data[i] = (mask.get(i)) ? a[i] : b[i];
    return result;
  }
};



  template<size_t N, typename ElemType>
  std::ostream& operator<<(std::ostream& os, const syrah::SIMDObject<N, ElemType>& wrapper) {
    for (size_t i = 0; i < N; i++) {
      const char* spacing = (i == 0) ? "" : ", ";
      os << spacing << wrapper[i];
    }
    return os;
  }

#endif // _SYRAH_OBJECT_WRAPPER_H_
