#include "MeshIO/LisztFileReader.h"
#include "MeshIO/FacetEdgeBuilder.h"

using namespace MeshIO;
MeshIO::id_t gr;


void test_facet_pool(size_t nfes, bool erase_last) {
	FacetPool pool;
	for(unsigned int i = 0; i < nfes; i++) {
		FacetEdge * fe = pool.alloc();
		fe->edge_face[0] = i;
	}
	if(erase_last) {
		pool.alloc();
		pool.freeLastFacet();
	}
	FacetPool::iterator_t it = pool.iterator();
	FacetEdge * fe;
	unsigned int i = 0;
	while(it.next(&fe)) {
		std::cout << fe->edge_face[0] << std::endl;
		assert(fe->edge_face[0] == i);
		i++;
	}
	assert(i == nfes);
}

struct RotateFilter {
	RotateFilter(lsize_t nv,lsize_t ne, lsize_t nf, lsize_t nc) {
		data[VERTEX_T] = nv;
		data[CELL_T] = nc;
		data[EDGE_T] = ne;
		data[FACE_T] = nf;
	}
	MeshIO::id_t operator()(IOElemType t, MeshIO::id_t id) {
		return (id + 1) % data[t];
	}
	lsize_t data[4];
};

int main(int argc, char ** argv) {
	
	for(int i = 0; i < 2; i++) {
		bool erase = i == 1;
		test_facet_pool(0,erase);
		test_facet_pool(1,erase);
		test_facet_pool(10,erase);
		test_facet_pool(FACET_POOL_PAGE_SIZE - 2,erase);
		test_facet_pool(FACET_POOL_PAGE_SIZE - 1,erase);
		test_facet_pool(FACET_POOL_PAGE_SIZE,erase);
		test_facet_pool(FACET_POOL_PAGE_SIZE + 1,erase);
		test_facet_pool(FACET_POOL_PAGE_SIZE - 2,erase);
		test_facet_pool(FACET_POOL_PAGE_SIZE + 10,erase);
		test_facet_pool(2*FACET_POOL_PAGE_SIZE - 2,erase);
		test_facet_pool(2*FACET_POOL_PAGE_SIZE - 1,erase);
		test_facet_pool(2*FACET_POOL_PAGE_SIZE,erase);
		test_facet_pool(2*FACET_POOL_PAGE_SIZE + 1,erase);
		test_facet_pool(2*FACET_POOL_PAGE_SIZE + 2,erase);
	}
	
	assert(argc == 2);
	LisztFileReader read;
	read.init(argv[1]);
	const LisztHeader & h = read.header();
	
	printf("Mesh Info for %s\nnV = %d\nnE = %d\nnF = %d\nnC = %d\nnFE = %d\nnBoundaries = %d\n",
	       argv[1],h.nV,h.nE,h.nF,h.nC,h.nFE,h.nBoundaries);
	BoundarySetEntry * boundaries = read.boundaries();
	
	for(lsize_t i = 0; i < h.nBoundaries; i++) {
		printf("boundary: name: %s type: %d, start: %d end: %d\n",boundaries[i].name.c_str(),
			   boundaries[i].type,boundaries[i].start,boundaries[i].end);
	}
	read.free(boundaries);
	
	PositionTable * table = read.positions();
	for(lsize_t i = 0; i < h.nV; i++) {
		printf("v = %d position = (%f,%f,%f)\n",i,table->data[i][0],table->data[i][1],table->data[i][2]);
	}
	read.free(table);
	
	FileFacetEdge * fes = read.facetEdges();
	for(lsize_t i = 0; i < h.nFE; i++) {
		printf("fe = %d, f = %d, e = %d,  (v = %d, c = %d) (v = %d, c = %d)\n",
	    i,fes[i].face,fes[i].edge, fes[i].hf[0].vert, fes[i].hf[0].cell, fes[i].hf[1].vert, fes[i].hf[1].cell);
	}
	
	FacetEdgeBuilder builder;
	builder.init(h.nV,h.nE,h.nF,h.nC,h.nFE);
	
	
	builder.insert(0,h.nFE,fes);

	builder.validateFullMesh();
	
	//now to mess with it
	
	for(lsize_t i = 0; i < h.nFE; i++) {
		MeshIO::id_t r = rand() % h.nFE;
		//printf("trying to add %d\n",r); 
		assert(!builder.insert(r,&fes[r]));
	}
	
	builder.clear();
	
	for(lsize_t i = h.nFE; i > 0; i--) {
		MeshIO::id_t r = i - 1;
		//printf("trying to add %d\n",r); 
		assert(builder.insert(r,&fes[r]));
	}
	builder.validateFullMesh();
	
	
	for(int i = 0; i < 1; i++) {
		builder.clear();
		//bozofill : grab fe's at random and try to add them
		//we are done when we finally catch em all pokémon!
		lsize_t nadded = 0;
		for(lsize_t i = 0; nadded < h.nFE; i++) {
			printf("so far %d: %d\n",i,nadded);
			gr = rand() % h.nFE;
			if(gr == 1014) {
				printf("magic");
			}
			if(builder.insert(gr,&fes[gr])) {
				nadded++;
			}
		}
		builder.validateFullMesh();
	}
	
	
	//now test if a filter doesn't majorly crash things
	RotateFilter rf(h.nV,h.nE,h.nF,h.nC);
	builder.filter(rf);
	
	read.free(fes);
	return 0;
}
