#include "FreeImage.h"
#include "MeshIO/Common.h"
#include "MeshIO/LisztFormat.h"
#include <limits.h>
#include <algorithm>
#include <sstream>

using namespace MeshIO;

size_t getdata(const char * fname, int ** data) {
	FILE * file = FOPEN(fname,"r");
	FileField field;
	fread(&field,sizeof(FileField),1,file);
	assert(field.range.type == LISZT_INT); //we only support int for now
	fseeko(file,field.data,SEEK_SET);
	*data = new int[field.nElems];
	fread(*data,sizeof(int),field.nElems,file);
	fclose(file);
	return field.nElems;
}
int main(int argc, char * argv[]) {
	size_t width = atoi(argv[1]);
	size_t start_images = 2;
	//get metadata
	int * min_use = NULL; //INT_MAX;
	int * max_use = NULL; //INT_MIN;
	
	for(size_t i = start_images; i < argc; i++) {
		int * data;
		size_t n = getdata(argv[i],&data);
		if(min_use == NULL) {
			min_use = new int[n];
			max_use = new int[n];
			for(size_t j = 0; j < n; j++) {
				min_use[j] = data[j];
				max_use[j] = data[j];
			}
		} else {
			for(size_t j = 0; j < n; j++) {
				min_use[j] = std::min(min_use[j],data[j]);
				max_use[j] = std::max(max_use[j],data[j]);
			}
		}
		delete [] data;
	}
	//printf("min = %d max = %d\n", min_use, max_use);
	FreeImage_Initialise();
	for(size_t i = start_images; i < argc; i++) {
		int * data;
		size_t n = getdata(argv[i],&data);
		size_t height = n / width + ( (n % width == 0) ? 0 : 1);
		FIBITMAP * bitmap = FreeImage_Allocate(width,height,24);
		RGBQUAD color;
		for(size_t j = 0; j < n; j++) {
			double v;
			if(min_use[j] == max_use[j]) {
				v = 0;
			} else {
				v = 255.0 * (data[j] - min_use[j]) / (double) (max_use[j] - min_use[j]);
			}
			//printf("value = %d %f\n",data[j],v);
			color.rgbRed = v;
			color.rgbGreen = v;
			color.rgbBlue = v;
			FreeImage_SetPixelColor(bitmap,j % width,j / width,&color);
		}
		std::ostringstream ss;
		ss << argv[i] << ".png";
		FreeImage_Save(FIF_PNG,bitmap, ss.str().c_str());
		delete [] data;
	}
	
	FreeImage_DeInitialise();
}