#include <string>
#include <stdlib.h>
#include "MeshIO/FluentParse.h"
#include "MeshIO/VTKParse.h"
#include "MeshIO/OFFParse.h"
#include "MeshIO/DPLParse.h"
#include "MeshIO/LisztFileWriter.h"
#include "MeshIO/BoundaryTable.h"

using namespace MeshIO;
bool endsWith(const std::string & filename, const std::string & ending) {
	return filename.size() >= ending.size() &&
	       std::equal(filename.begin() + filename.size() - ending.size(),
				   filename.end(),ending.begin());
}

int main(int argc, char * argv[]) {
	if (!(argc >= 3)) {
		fprintf(stderr,"usage: %s <input_mesh> <output_mesh>\n",argv[0]);
		return 1;
	}
	const char* filename = argv[1];
	const char* output = argv[2];
	LisztFileWriter writer;
	BoundaryTable boundaries;
	writer.boundaries = &boundaries;
	boundaries.parseArgs(argc - 3, argv + 3);

	if(endsWith(filename,".msh")) {
		FluentParse parse;
		parse.init(&writer,filename,output);
		parse.parseAll();
	} else if(endsWith(filename,".off")) {
		OFFParse off;
		off.init(&writer,filename,output);
		off.parseAll();
	} else if(endsWith(filename,"dpl")) {
		DPLParse dpl;
		dpl.init(&writer,filename,output);
		dpl.parseAll();
	} else if(endsWith(filename,".lmesh")) {
		// TODO(mbarrien): Need liszt mesh parser to be able to add user boundries
		char * abs_path = realpath(filename,NULL);
		execl("/bin/ln","/bin/ln","-s","-f",abs_path,output,NULL);
		fprintf(stderr,"Error copying liszt mesh.");
		return 1;
	} else {
		VTKParse reader;
		reader.init(&writer,filename,output);
		reader.parseAll();
	}

	boundaries.finalizeBoundarySetTable();
	boundaries.writeBoundarySetTable(writer);
	writer.writeHeader();
	writer.close();
	return 0;
}
