/*
 *  ColorMap.h
 *  lisztvis
 *
 *  Created by Zach Devito on 11/21/09.
 *  Copyright 2009 __MyCompanyName__. All rights reserved.
 *
 */
#ifndef _COLOR_MAP_H
#define _COLOR_MAP_H
#include <vector>
#include <OpenGL/gl.h>

static GLubyte default_grey[3] = {128,128,128};
static GLubyte nan_color[3] = {255,0,255};
class ColorMap {
public:
	ColorMap(GLubyte c[][3], size_t s) : colors(c), size(s), default_color(default_grey) {}
	void setColor(int idx, int ncategories) {
		if(ncategories == 1) {
			glColor3ubv(default_color);
		} else {
			assert(idx >= 0 && idx < size);
			glColor3ubv(colors[idx]);
		}
	}
	void setDefaultColor(GLubyte * c) {
		default_color = c;
	}
	void setColor(double v) {
		setColor(v,0.,1.);
	}
	void setColor(double v, double min, double max) {
		if(isnan(v)) {
			glColor3ubv(nan_color);
			return;
		}
		int idx;
		if(min == max)
			idx = size / 2;
		else {
			assert(min < max);
			if(v < min)
				v = min;
			if(v > max)
				v = max;
			double a = (v - min) / (max - min);
			idx = a * size;
			if(idx == size)
				idx--;
		}
		assert(idx >= 0 && idx < size);
		glColor3ubv(colors[idx]);
	}
	void initImages() {
		for(size_t i = 0; i < size; i++) {
			const double IMAGE_SIZE = 15.f;
			NSImage* anImage = [[NSImage alloc] initWithSize:NSMakeSize(IMAGE_SIZE,IMAGE_SIZE)];
			[anImage lockFocus];
				[[NSColor colorWithCalibratedRed: colors[i][0] / 256.f green:colors[i][1] / 256.f blue:colors[i][2] / 256.f alpha: 1.f] set];
				NSRect myRect = NSMakeRect(0.0,0.0,IMAGE_SIZE,IMAGE_SIZE);
				NSRectFill(myRect);
			[anImage unlockFocus];
			images.push_back(anImage);
		}
	}
	const std::vector<NSImage*> & getImages() const {
		return images;
	}
	size_t size;
private:
	GLubyte * default_color;
	GLubyte (*colors)[3];
	std::vector<NSImage*> images;
};


static GLubyte diverging_colors_arr[][3] = {
{5, 48, 97},
{33, 102, 172},
{67, 147, 195},
{146, 197, 222},
{209, 229, 240},
{247, 247, 247},
{253, 219, 199},
{244, 165, 130},
{214, 96, 77},
{178, 24, 43},
{103, 0, 31}
};
static ColorMap diverging_colors(diverging_colors_arr,11);



static GLubyte category_colors_arr[][3] = {
{114, 158, 206},
{255, 158, 74},
{103, 191, 92}, 
{237, 102, 93},
{173, 139, 201},
{168, 120, 110},
{237, 151, 202},
{162, 162, 162},
{205, 204, 93},
{109, 204, 218}, 
{144, 169, 202},
{225, 157, 90},
{122, 193, 108},
{225, 122, 120},
{197, 176, 213},
{196, 156, 148},
{247, 182, 210},
{199, 199, 199},
{219, 219, 141},
{158, 218, 229}
};
static ColorMap category_colors(category_colors_arr,20);

#endif