//
//  LisztController.m
//  lisztvis
//
//  Created by Zach Devito on 11/18/09.
//  Copyright 2009 __MyCompanyName__. All rights reserved.
//

#import "LisztController.h"
#include "LisztClient.h"

Field<VertexType,double3> orig_position;
Field<VertexType,double3> position;
Field<FaceType,double3> face_center;
Field<CellType,double3> cell_center;

Field<VertexType,int> vertex_id;
Field<EdgeType,int> edge_id;
Field<FaceType,int> face_id;
Field<CellType,int> cell_id;

double3 mesh_center;
double mesh_scale;

void RecalcMeshProperties() {

    FORALL_SET_SER(f,mesh()->faces())
    face_center[f] = double3(0.,0.,0.);
	FORALL_SET_SER(v,f.vertices())
	face_center[f] += position(v);
	ENDSET
	face_center[f] /= f.vertices().size();
	ENDSET
	FORALL_SET_SER(c,mesh()->cells())
    cell_center[c] = double3(0.,0.,0.);
	FORALL_SET_SER(v,c.vertices())
	cell_center[c] += position(v);
	ENDSET
	cell_center[c] /= c.vertices().size();
	ENDSET
	
	mesh_center = double3(0.,0.,0.);
	
	double3 maxv(-INFINITY,-INFINITY,-INFINITY);
	double3 minv(INFINITY,INFINITY,INFINITY);
	
	FORALL_SET_SER(v,mesh()->vertices())
	mesh_center += position(v);
	minv = min(minv,position(v));
	maxv = max(maxv,position(v));
	ENDSET
	mesh_center /= mesh()->vertices().size();
	double3 diag = maxv - minv;
	mesh_scale = 1.f/diag.length();
    
}
static void InitBuiltInFieldsAndRun() {	
	LisztClient_liszt()->buildFieldWithConst(0,"vertex_id",0,&vertex_id);
	LisztClient_liszt()->buildFieldWithConst(1,"edge_id",0,&edge_id);
	LisztClient_liszt()->buildFieldWithConst(2,"face_id",0,&face_id);
	LisztClient_liszt()->buildFieldWithConst(3,"cell_id",0,&cell_id);
	FORALL_SET_SER(e,mesh()->vertices()) 
	vertex_id[e] = e.ID();
	ENDSET
	FORALL_SET_SER(e,mesh()->edges()) 
	edge_id[e] = e.ID();
	ENDSET
	FORALL_SET_SER(e,mesh()->faces()) 
	face_id[e] = e.ID();
	ENDSET
	FORALL_SET_SER(e,mesh()->cells()) 
	cell_id[e] = e.ID();
	ENDSET
	
	LisztClient_liszt()->buildFieldWithPositions(4, "fluent_position",&orig_position);
	LisztClient_liszt()->buildFieldWithConst(5,"face_center",double3(0.,0.,0.),&face_center);
	LisztClient_liszt()->buildFieldWithConst(6,"cell_center",double3(0.,0.,0.),&cell_center);
	position.init("<built-in>",mesh()->vertices().size());
    FORALL_SET_SER(v,mesh()->vertices())
    position[v] = orig_position(v);
    ENDSET
	RecalcMeshProperties();
    LisztClient_liszt()->setCurrent();
	//LisztClient_liszt()->addWatchPoint("text_exp","Phi(1)");
	LisztClient_main();
    
    LisztSinglePrivate::LisztSingle::watch_point_ref wp(new WatchPoint("Program Exit",""));
    LisztClient_liszt()->takeSnapshot(wp,IObject(),IObject(),"<Exit>",0);
}

static bool isFieldName(const std::string & str) {
	for(size_t i = 0; i < ELEM_TYPE_SIZE; i++) {
		for(size_t j = 0; j < LisztClient_liszt()->registered_fields[i].size(); j++) {
			if(LisztClient_liszt()->registered_fields[i][j]->name() == str)
				return true;
		}
	}
	return false;
}

static void trim(std::string &s) {
    const std::string c = "\r\n";
    int p2 = s.find_last_not_of(c);
    if (p2 == std::string::npos) return;
    int p1 = s.find_first_not_of(c);
    if (p1 == std::string::npos) p1 = 0;
    s = s.substr(p1, (p2-p1)+1);
}

@implementation LisztController

- (bool) displayEdges {
	return show_edges;
}
- (bool) displayFaces {
	return show_faces;
}
- (bool) displayVerts {
	return show_verts;
}
- (bool) displayCells {
	return show_cells;
}
- (void) awakeFromNib {
	[self setValue: [NSNumber numberWithBool:YES]  forKey: @"show_edges"];
	[self setValue: [NSNumber numberWithBool:YES] forKey: @"show_faces"];
	
	filters[EDGE].init(EDGE);
	filters[FACE].init(FACE);
	filters[CELL].init(CELL);
	filters[VERTEX].init(VERTEX);
	
	encodings[FACE][ENC_SIZE].init(FACE,&filters[FACE]);
	encodings[FACE][ENC_COLOR].init(FACE,&filters[FACE]);
	encodings[FACE][ENC_TEXT].init(FACE,&filters[FACE]);
	encodings[FACE][ENC_VECTOR].init(FACE,&filters[FACE]);
	
	encodings[VERTEX][ENC_SIZE].init(VERTEX,&filters[VERTEX]);
	encodings[VERTEX][ENC_COLOR].init(VERTEX,&filters[VERTEX]);
	encodings[VERTEX][ENC_TEXT].init(VERTEX,&filters[VERTEX]);
	encodings[VERTEX][ENC_VECTOR].init(VERTEX,&filters[VERTEX]);
    position_encoding.init(VERTEX,&filters[VERTEX]);
	
	encodings[EDGE][ENC_SIZE].init(EDGE,&filters[EDGE]);
	encodings[EDGE][ENC_COLOR].init(EDGE,&filters[EDGE]);
	encodings[EDGE][ENC_TEXT].init(EDGE,&filters[EDGE]);
	encodings[EDGE][ENC_VECTOR].init(EDGE,&filters[EDGE]);
	
	encodings[CELL][ENC_SIZE].init(CELL,&filters[CELL]);
	encodings[CELL][ENC_COLOR].init(CELL,&filters[CELL]);
	encodings[CELL][ENC_TEXT].init(CELL,&filters[CELL]);
	encodings[CELL][ENC_VECTOR].init(CELL,&filters[CELL]);
	
    char * mesh_path = (char *) [[[NSBundle mainBundle] pathForResource:@"liszt_mesh" ofType:@"msh"] UTF8String];
	char * argv[] = { "someprogram", mesh_path, "0" };
	
	LisztClient_predescriptor(3, argv);
	InitBuiltInFieldsAndRun();
    [mainWindow makeKeyAndOrderFront:nil];
}
- (IBAction) filterModified: (id) sender {
	int tag = [sender tag];
	std::string cstring([[sender stringValue] UTF8String]);
	trim(cstring);
	if(isFieldName(cstring)) {
		cstring += "(id)";
		[sender setStringValue: toNSString(cstring)];
	}
	if(!filters[tag].setFilterString(cstring)) {
		[sender setStringValue:@""];
	} else {
		NSLog(@"VALID!");
	}
	
	for(size_t i = 0; i <  ENC_TYPE_SIZE; i++)
		encodings[tag][i].calcStats();
	[meshView postRecompile];
	[colorLegend newEncoding: &encodings[tag][ENC_COLOR]];
}
- (void) snapshotSelectionChanged {
	for(size_t i = 0; i < ELEM_TYPE_SIZE; i++) {
		for(size_t j = 0; j < ENC_TYPE_SIZE; j++) {
			encodings[i][j].calcStats();
		}
	}
	[colorLegend reloadData];
	[elementInspector reloadData];
	[meshView postRecompile];
}
- (Filter*) getFilter: (ElemType) type {
	return &filters[type];
}

- (IBAction) encodingModified: (id) sender {
	int tag = [sender tag];
	int eid = tag >> 2;
	int enc = (0x3) & tag;
	std::string cstring([[sender stringValue] UTF8String]);
	trim(cstring);
	if(isFieldName(cstring)) {
		cstring += "(id)";
		[sender setStringValue: toNSString(cstring)];
	}
    Encoding & encoding = (tag < 16) ? encodings[eid][enc] : position_encoding;
	if(!encoding.setExp(cstring)) {
		[sender setStringValue:@""];
	} else {
		if(enc == ENC_SIZE || enc == ENC_COLOR) {
			NSString * key = 
			(eid == VERTEX) ? @"show_verts" :
			(eid == EDGE) ? @"show_edges" :
			(eid == FACE) ? @"show_faces" :
			(eid == CELL) ? @"show_cells" : @"error";
			[self setValue: [NSNumber numberWithBool:YES] forKey: key];
		}
		NSLog(@"VALID!");
	}
	if(enc == ENC_COLOR) {
		[colorLegend newEncoding: &encodings[eid][enc]];
	}
	[meshView postRecompile];
}
- (IBAction) visibilityModified: (id) sender {
	[meshView postRecompile];
}
- (IBAction) showLegend: (id) sender {
	[legend makeKeyAndOrderFront:self];
}
- (Encoding*) getEncoding: (EncodingTypes) t forType: (ElemType) e {
    if(t == ENC_POS)
        return &position_encoding;
	else
        return &encodings[e][t];
}
- (void)tabView:(NSTabView *)tabView didSelectTabViewItem:(NSTabViewItem *)tabViewItem {
	NSLog([tabViewItem label]);
	if([[tabViewItem label] isEqualToString:@"Verts"]) {
		[colorLegend newEncoding: &encodings[VERTEX][ENC_COLOR]];
	} else if([[tabViewItem label] isEqualToString:@"Edges"]) {
		[colorLegend newEncoding: &encodings[EDGE][ENC_COLOR]];
	} else if([[tabViewItem label] isEqualToString:@"Faces"]) {
		[colorLegend newEncoding: &encodings[FACE][ENC_COLOR]];
	} else if([[tabViewItem label] isEqualToString:@"Cells"]) {
		[colorLegend newEncoding: &encodings[CELL][ENC_COLOR]];
	}
	[meshView postRecompile];
}
- (IBAction) runSimulation: (id) sender {
	
	LisztClient_liszt()->resetSnapshots();
	
	InitBuiltInFieldsAndRun();
	
	for(size_t i = 0; i <  ELEM_TYPE_SIZE; i++) {
		for(size_t j = 0; j <  ENC_TYPE_SIZE; j++)	
			encodings[i][j].calcStats();
		[colorLegend newEncoding: &encodings[i][ENC_COLOR]];
	}

	[snapshot_table reloadData];
	
	[edge_field_table reloadData];
	[face_field_table reloadData];
	[cell_field_table reloadData];
	[vert_field_table reloadData];
	
	[meshView postRecompile];
	[snapshot_table deselectAll:self];
	NSLog(@"Selected row: %d",[snapshot_table selectedRow]);
	[snapshot_table reloadData];
	[elementInspector reloadData];
}
- (void) setElementSelection: (ElemType) et withID: (id_type) idx atPosition: (NSPoint) point {
	[elementInspector setElementSelection:et withID: idx atPosition: point];
}
@end
