//
// File:		LisztGLView.m
//
// Abstract:	Basic OpenGL View with Renderer information (This example was modified to create the GL window for Liszt
//              Rotation/translation/etc. were implemented using this pre-packaged Apple code.
// Version:		1.1 - minor fixes.
//				1.0 - Original release.
//				
//
// Disclaimer:	IMPORTANT:  This Apple software is supplied to you by Apple Inc. ("Apple")
//				in consideration of your agreement to the following terms, and your use,
//				installation, modification or redistribution of this Apple software
//				constitutes acceptance of these terms.  If you do not agree with these
//				terms, please do not use, install, modify or redistribute this Apple
//				software.
//
//				In consideration of your agreement to abide by the following terms, and
//				subject to these terms, Apple grants you a personal, non - exclusive
//				license, under Apple's copyrights in this original Apple software ( the
//				"Apple Software" ), to use, reproduce, modify and redistribute the Apple
//				Software, with or without modifications, in source and / or binary forms;
//				provided that if you redistribute the Apple Software in its entirety and
//				without modifications, you must retain this notice and the following text
//				and disclaimers in all such redistributions of the Apple Software. Neither
//				the name, trademarks, service marks or logos of Apple Inc. may be used to
//				endorse or promote products derived from the Apple Software without specific
//				prior written permission from Apple.  Except as expressly stated in this
//				notice, no other rights or licenses, express or implied, are granted by
//				Apple herein, including but not limited to any patent rights that may be
//				infringed by your derivative works or by other works in which the Apple
//				Software may be incorporated.
//
//				The Apple Software is provided by Apple on an "AS IS" basis.  APPLE MAKES NO
//				WARRANTIES, EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION THE IMPLIED
//				WARRANTIES OF NON - INFRINGEMENT, MERCHANTABILITY AND FITNESS FOR A
//				PARTICULAR PURPOSE, REGARDING THE APPLE SOFTWARE OR ITS USE AND OPERATION
//				ALONE OR IN COMBINATION WITH YOUR PRODUCTS.
//
//				IN NO EVENT SHALL APPLE BE LIABLE FOR ANY SPECIAL, INDIRECT, INCIDENTAL OR
//				CONSEQUENTIAL DAMAGES ( INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
//				SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
//				INTERRUPTION ) ARISING IN ANY WAY OUT OF THE USE, REPRODUCTION, MODIFICATION
//				AND / OR DISTRIBUTION OF THE APPLE SOFTWARE, HOWEVER CAUSED AND WHETHER
//				UNDER THEORY OF CONTRACT, TORT ( INCLUDING NEGLIGENCE ), STRICT LIABILITY OR
//				OTHERWISE, EVEN IF APPLE HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// Copyright ( C ) 2003-2007 Apple Inc. All Rights Reserved.
//
#define id Id_hacks
#include "LisztClient.h"
#undef id

#import <GLUT/glut.h>
#include "ColorMap.h"
#import "LisztGLView.h"
#import "trackball.h"
#include <sstream>
// ==================================

extern Field<VertexType,double3> position;
extern Field<FaceType,double3> face_center;
extern Field<CellType,double3> cell_center;

extern Field<VertexType,int> vertex_id;
extern Field<EdgeType,int> edge_id;
extern Field<FaceType,int> face_id;
extern Field<CellType,int> cell_id;


//TODO: add boundary set classification field

extern double3 mesh_center;
extern double mesh_scale;
// simple cube data
GLint cube_num_vertices = 8;

GLfloat cube_vertices [8][3] = {
{1.0, 1.0, 1.0}, {1.0, -1.0, 1.0}, {-1.0, -1.0, 1.0}, {-1.0, 1.0, 1.0},
{1.0, 1.0, -1.0}, {1.0, -1.0, -1.0}, {-1.0, -1.0, -1.0}, {-1.0, 1.0, -1.0} };

GLfloat cube_vertex_colors [8][3] = {
{1.0, 1.0, 1.0}, {1.0, 1.0, 0.0}, {0.0, 1.0, 0.0}, {0.0, 1.0, 1.0},
{1.0, 0.0, 1.0}, {1.0, 0.0, 0.0}, {0.0, 0.0, 0.0}, {0.0, 0.0, 1.0} };

GLint num_faces = 6;

short cube_faces [6][4] = {
{3, 2, 1, 0}, {2, 3, 7, 6}, {0, 1, 5, 4}, {3, 0, 4, 7}, {1, 2, 6, 5}, {4, 5, 6, 7} };

recVec gOrigin = {0.0, 0.0, 0.0};

// single set of interaction flags and states
GLint gDollyPanStartPoint[2] = {0, 0};
GLfloat gTrackBallRotation [4] = {0.0f, 0.0f, 0.0f, 0.0f};
GLboolean gDolly = GL_FALSE;
GLboolean gPan = GL_FALSE;
GLboolean gTrackball = GL_FALSE;
LisztGLView * gTrackingViewInfo = NULL;

// time and message info
CFAbsoluteTime gMsgPresistance = 10.0f;

// error output
GLString * gErrStringTex;
float gErrorTime;

// ==================================

#pragma mark ---- OpenGL Capabilities ----

// GL configuration info globals
// see GLCheck.h for more info
CGDisplayCount gNumDisplays = 0;

#pragma mark ---- Utilities ----

static CFAbsoluteTime gStartTime = 0.0f;

// set app start time
static void setStartTime (void)
{	
	gStartTime = CFAbsoluteTimeGetCurrent ();
}

// ---------------------------------

// return float elpased time in seconds since app start
static CFAbsoluteTime getElapsedTime (void)
{	
	return CFAbsoluteTimeGetCurrent () - gStartTime;
}

#pragma mark ---- Error Reporting ----

// error reporting as both window message and debugger string
void reportError (char * strError)
{
    NSMutableDictionary *attribs = [NSMutableDictionary dictionary];
    [attribs setObject: [NSFont fontWithName: @"Monaco" size: 9.0f] forKey: NSFontAttributeName];
    [attribs setObject: [NSColor whiteColor] forKey: NSForegroundColorAttributeName];
	
	gErrorTime = getElapsedTime ();
	NSString * errString = [NSString stringWithFormat:@"Error: %s (at time: %0.1f secs).", strError, gErrorTime];
	NSLog (@"%@\n", errString);
}

// ---------------------------------

// if error dump gl errors to debugger string, return error
GLenum glReportError (void)
{
	GLenum err = glGetError();
	if (GL_NO_ERROR != err)
		reportError ((char *) gluErrorString (err));
	return err;
}
template<typename MO>
static GLuint getName(const MO & mo) {
	return (MO::type::value << 30) | (mo.ID() + 1);
}
#pragma mark ---- OpenGL Utils ----

// ---------------------------------

// draw our simple cube based on current modelview and projection matrices
static void drawCube (GLfloat fSize)
{
	long f, i;
	if (1) {
		glColor3f (1.0, 0.5, 0.0);
		glBegin (GL_QUADS);
		for (f = 0; f < num_faces; f++)
			for (i = 0; i < 4; i++) {
				glColor3f (cube_vertex_colors[cube_faces[f][i]][0], cube_vertex_colors[cube_faces[f][i]][1], cube_vertex_colors[cube_faces[f][i]][2]);
				glVertex3f(cube_vertices[cube_faces[f][i]][0] * fSize, cube_vertices[cube_faces[f][i]][1] * fSize, cube_vertices[cube_faces[f][i]][2] * fSize);
			}
		glEnd ();
	}
	if (1) {
		glColor3f (0.0, 0.0, 0.0);
		for (f = 0; f < num_faces; f++) {
			glBegin (GL_LINE_LOOP);
			for (i = 0; i < 4; i++)
				glVertex3f(cube_vertices[cube_faces[f][i]][0] * fSize, cube_vertices[cube_faces[f][i]][1] * fSize, cube_vertices[cube_faces[f][i]][2] * fSize);
			glEnd ();
		}
	}
}

void writeText(const double3 & pos, const std::string & str) {
	glRasterPos3f(pos.x(),pos.y(),pos.z());
	for(size_t i = 0; i < str.size(); i++) {
		glutBitmapCharacter(GLUT_BITMAP_HELVETICA_12,str.at(i));
	}
}
void writeIObject(const double3 & pos, const IObject & obj) {
	std::ostringstream str;
	str << obj;
	writeText(pos,str.str());
}

void drawVectorMark(const double3 & startPos, const IObject & vobj, double referenceSize) {
	if(!vobj.isVec(3))
		return;
	glColor3ub(200,200,200);
	IObject::vec_type vo = vobj.vecValue();
	
	double3 v(vo[0].doubleValue(0.),vo[1].doubleValue(1.),vo[2].doubleValue(0.));
	v.normalize();
	
	double3 offset = startPos + referenceSize * v;
	glLineWidth(1.0);
	glBegin(GL_LINES);
	glVertex3f(startPos.x(),startPos.y(),startPos.z());
	glVertex3f(offset.x(),offset.y(),offset.z());
	glEnd();
	glPointSize(2.0);
	glBegin(GL_POINTS);
	glVertex3f(offset.x(),offset.y(),offset.z());
	glEnd();
	
}
template<typename MO>
void drawTextMark(const double3 & pos, const Encoding & enc, const MO & mo) {
	glColor3f(1.f,1.f,1.f);
	IObject txt = enc(mo);
	if(!txt.isUnknown())
		writeIObject(pos,txt);
}

GLubyte default_colors[ELEM_TYPE_SIZE][3] = {
{126,126,126},
{31, 120, 180},
{126,126,126},
{168,168,168},
};
template<typename MO>
void setColorEncoding(const Encoding & enc, const MO & mo) {
	if(enc.categorical()) {
		id_type cat = enc.category(mo);
		category_colors.setDefaultColor(default_colors[MO::type::value]);
		category_colors.setColor(cat,enc.numCategories());
	} else {
		double colorv = enc.linearScale(mo).doubleValue(.5);
		diverging_colors.setColor(colorv);
	}
}

template<typename MO>
double drawPointMark(const double3 & pos, const Encoding & size_enc, const MO & mo, double referenceSize, double referenceSizeVec) {
	IObject sizeo = size_enc(mo);
	double s = size_enc.linearScale(mo).doubleValue(.5);
	if(s < .01)
		s = .01;
	double size = referenceSize * .1 + .9 * referenceSize * s;
	glPushMatrix();
	glTranslatef(pos.x(),pos.y(),pos.z());
	glutSolidCube(size);
	glColor3f(0,0,0);
	glutWireCube(size);
	glPopMatrix();
	return size/2 + referenceSize * .1;
}

struct BBox {
	BBox() {
		double mind = std::numeric_limits<double>::min();
		double maxd = std::numeric_limits<double>::max();
		min_ = double3(maxd,maxd,maxd);
		max_ = double3(mind,mind,mind);
	}
	
	BBox & unionWith(const double3 & p) {
		min_.min(p);
		max_.max(p);
		return *this;
	}
	double3 & min() { return min_; }
	double3 & max() { return max_; }
	
	double3 min_;
	double3 max_;
};

// ===================================

@implementation LisztGLView

// pixel format definition
+ (NSOpenGLPixelFormat*) basicPixelFormat
{
    NSOpenGLPixelFormatAttribute attributes [] = {
        NSOpenGLPFAWindow,
        NSOpenGLPFADoubleBuffer,	// double buffered
        NSOpenGLPFADepthSize, (NSOpenGLPixelFormatAttribute)32, // 16 bit depth buffer
        (NSOpenGLPixelFormatAttribute)nil
    };
    return [[[NSOpenGLPixelFormat alloc] initWithAttributes:attributes] autorelease];
}
- (void) postRecompile {
	NSLog(@"Mesh Recompile Posted");
	requires_redraw = true;
    [self setNeedsDisplay: YES];
}
- (void) renderMesh {
	if(requires_redraw) {
		NSLog(@"Mesh Recompile Performed");
		NSAttributedString * mylegend = [legend getText];
		if(legendString) {
			[legendString setString: mylegend];
		} else {
			legendString = [[GLString alloc] initWithAttributedString: mylegend withTextColor:[NSColor colorWithDeviceRed:1.0f green:1.0f blue:1.0f alpha:1.0f] withBoxColor:[NSColor colorWithDeviceRed:0.5f green:0.5f blue:0.5f alpha:0.5f] withBorderColor:[NSColor colorWithDeviceRed:0.8f green:0.8f blue:0.8f alpha:0.8f]];
		}
		glDeleteLists(mesh_list, 1);
		glNewList(mesh_list,GL_COMPILE_AND_EXECUTE);
			[self drawMesh];
		glEndList();
		requires_redraw = false;
	} else {
		glCallList(mesh_list);
	}
}
- (void) drawMesh
{
    /* pre-calc position encoding */
    if(1) {
        Encoding & pos_enc = *[controller getEncoding:ENC_POS forType:VERTEX];
        FORALL_SET_SER(v,mesh()->vertices())
        IObject value = pos_enc(v);
        if(value.isVec(3)) {
             IObject::vec_type vo = pos_enc(v).vecValue();
            IObject x = vo[0];
            IObject y = vo[1];
            IObject z = vo[2];
            if(x.isDouble() && y.isDouble() && z.isDouble()) {
                 position[v] = double3(x.doubleValue(),y.doubleValue(),z.doubleValue());
            }
        }
        ENDSET
        RecalcMeshProperties();
    }
	glPushMatrix();
	glScalef(mesh_scale,mesh_scale,mesh_scale);
	glTranslatef(-mesh_center.x(), -mesh_center.y(),-mesh_center.z());
	glLineWidth(1.0f);
	glBegin(GL_LINES);
	// X axis
	glColor3f(1, 0, 0);
	glVertex3f(0, 0, 0);
	glVertex3f(2, 0, 0);
	// Y axis
	glColor3f(0, 1, 0);
	glVertex3f(0, 0, 0);
	glVertex3f(0, 2, 0);
	// Z axis
	glColor3f(0, 0, 1);
	glVertex3f(0, 0, 0);
	glVertex3f(0, 0, 2);
	
	glEnd();
	
    glColor3f(.5,.5,.5);
    glPointSize(5.0f);
	
	/*face encodings */ if(1) {
		Encoding & size_enc = *[controller getEncoding:ENC_SIZE forType:FACE];
		Encoding & color_enc = *[controller getEncoding:ENC_COLOR forType:FACE];
		Encoding & text_enc = *[controller getEncoding:ENC_TEXT forType:FACE];
		Encoding & vec_enc = *[controller getEncoding:ENC_VECTOR forType:FACE];
		FORALL_SET_SER(f,mesh()->faces())
		glLoadName(getName(f));
		if([controller getFilter:FACE]->shouldDisplay(f.ID())) {
			if([controller displayFaces]) {
				setColorEncoding(color_enc,f);				
				IObject sizeo = size_enc(f);
				double s = size_enc.linearScale(f).doubleValue(.35f);
				if(s < .01)
					s = .01;
				glBegin(GL_POLYGON);
				FORALL_SET_SER(v,f.vertices())
				float3 posv = face_center(f) + s*(position(v) - face_center(f));
				glVertex3f(posv.x(),posv.y(),posv.z());
				ENDSET
				glEnd();
				
				glColor3f(0.,0.,0.);
				glBegin(GL_LINE_LOOP);
				FORALL_SET_SER(v,f.vertices())
				float3 posv = face_center(f) + s*(position(v) - face_center(f));
				glVertex3f(posv.x(),posv.y(),posv.z());
				ENDSET
				glEnd();
			}
			double reflength = 1.0/mesh_scale * .01;
			drawVectorMark(face_center(f),vec_enc(f),reflength);
			drawTextMark(face_center(f),text_enc,f);
		}
		ENDSET
	}
	/*edge encodings */ if(1) {
		Encoding & size_enc = *[controller getEncoding:ENC_SIZE forType:EDGE];
		Encoding & color_enc = *[controller getEncoding:ENC_COLOR forType:EDGE];
		Encoding & text_enc = *[controller getEncoding:ENC_TEXT forType:EDGE];
		Encoding & vec_enc = *[controller getEncoding:ENC_VECTOR forType:EDGE];
		FORALL_SET_SER(e,mesh()->edges())
		glLoadName(getName(e));
		if([controller getFilter:EDGE]->shouldDisplay(e.ID())) {
			double3 head = position(e.head());
			double3 tail = position(e.tail());
			double3 pos = .5*(head + tail);
			if([controller displayEdges]) {
				setColorEncoding(color_enc,e);
				
				IObject sizeo = size_enc(e);
				double s = size_enc.linearScale(e).doubleValue(1.f);
				if(s < .01)
					s = .01;
				double3 toHead = head - pos;
				head = pos + s*toHead;
				tail = pos - s*toHead;
				glBegin(GL_LINES);
					glVertex3f(head.x(),head.y(),head.z());
					glVertex3f(tail.x(),tail.y(),tail.z());
				glEnd();			
			}
			double reflength = 1.0/mesh_scale * .01;
			drawVectorMark(pos,vec_enc(e),reflength);
			drawTextMark(pos,text_enc,e);
		}
		ENDSET
	}
	
	/*vertex*/ if(1) {
		Encoding & size_enc = *[controller getEncoding:ENC_SIZE forType:VERTEX];
		Encoding & color_enc = *[controller getEncoding:ENC_COLOR forType:VERTEX];
		Encoding & text_enc = *[controller getEncoding:ENC_TEXT forType:VERTEX];
		Encoding & vec_enc = *[controller getEncoding:ENC_VECTOR forType:VERTEX];
		FORALL_SET_SER(v,mesh()->vertices()) {
			glLoadName(getName(v));
			if([controller getFilter:VERTEX]->shouldDisplay(v.ID())) {
				double3 text_pos = position(v);
				double reflength = 1.0/mesh_scale * .01;
				if([controller displayVerts]) {
					//double3 refsizev = ( position(v.vertices().element()) - position(v) );
					 //refsizev.length();
					setColorEncoding(color_enc,v);
					double offset = drawPointMark(position(v),size_enc,v, reflength, reflength);
					text_pos +=  offset * double3(0.,1.,0.);
				}
				drawVectorMark(position(v),vec_enc(v),reflength);
				drawTextMark(text_pos, text_enc,v);
			}
		} ENDSET
	}
	/*cell*/ if(1) {
		Encoding & size_enc = *[controller getEncoding:ENC_SIZE forType:CELL];
		Encoding & color_enc = *[controller getEncoding:ENC_COLOR forType:CELL];
		Encoding & text_enc = *[controller getEncoding:ENC_TEXT forType:CELL];
		Encoding & vec_enc = *[controller getEncoding:ENC_VECTOR forType:CELL];
		FORALL_SET_SER(c,mesh()->cells()) {
			glLoadName(getName(c));
			if([controller getFilter:CELL]->shouldDisplay(c.ID())) {
				double3 text_pos = cell_center(c);
				double reflength = 1.0/mesh_scale * .04;
				if([controller displayCells]) {
					setColorEncoding(color_enc,c);
					double3 center = cell_center(c);
					double s = size_enc.linearScale(c).doubleValue(.35f);
					BBox bbox;
					FORALL_SET_SER(f,c.faces())
						glBegin(GL_POLYGON);
						FORALL_SET_SER(v,f.verticesCCW())
							double3 pos = center + s*(position(v) - center);
							bbox.unionWith(pos);
							glVertex3f(pos.x(),pos.y(),pos.z());
						ENDSET
						glEnd();

					ENDSET
					glColor3f(0.f,0.f,0.f);
					FORALL_SET_SER(f,c.faces())
						glBegin(GL_LINE_LOOP);
						FORALL_SET_SER(v,f.verticesCCW())
							double3 pos = center + s*(position(v) - center);
							glVertex3f(pos.x(),pos.y(),pos.z());
						ENDSET
						glEnd();
					ENDSET
					/*
					double3 refsizev = ( position(c.vertices().element()) - cell_center(c) );
					double offset = drawPointMark(cell_center(c),size_enc,c, reflength,.5 * reflength);
					*/
					text_pos[1] = bbox.max().y() + .01f * (bbox.max().y() - center.y());
				}
				drawVectorMark(cell_center(c),vec_enc(c),1.1*reflength);
				drawTextMark(text_pos, text_enc,c);
			}
		} ENDSET
	}
	glPopMatrix();
}
// ---------------------------------

// update the projection matrix based on camera and view info
- (void) updateProjection: (BOOL) reset
{
	GLdouble ratio, radians, wd2;
	GLdouble left, right, top, bottom, near, far;
	
    [[self openGLContext] makeCurrentContext];
	
	// set projection
	glMatrixMode (GL_PROJECTION);
	if(reset)
		glLoadIdentity ();
	near = .0001;
	if (near < 0.00001)
		near = 0.00001;
	far = 100;
	//far = -camera.viewPos.z + shapeSize * 0.5;
	if (far < 1.0)
		far = 1.0;
	radians = 0.0174532925 * camera.aperture / 2; // half aperture degrees to radians 
	wd2 = near * tan(radians);
	ratio = camera.viewWidth / (float) camera.viewHeight;
	if (ratio >= 1.0) {
		left  = -ratio * wd2;
		right = ratio * wd2;
		top = wd2;
		bottom = -wd2;	
	} else {
		left  = -wd2;
		right = wd2;
		top = wd2 / ratio;
		bottom = -wd2 / ratio;	
	}
	glFrustum (left, right, bottom, top, near, far);
	[self updateCameraString];
}

// ---------------------------------

// updates the contexts model view matrix for object and camera moves
- (void) updateModelView
{
    [[self openGLContext] makeCurrentContext];
	
	// move view
	glMatrixMode (GL_MODELVIEW);
	glLoadIdentity ();
	gluLookAt (camera.viewPos.x, camera.viewPos.y, camera.viewPos.z,
			   camera.viewPos.x + camera.viewDir.x,
			   camera.viewPos.y + camera.viewDir.y,
			   camera.viewPos.z + camera.viewDir.z,
			   camera.viewUp.x, camera.viewUp.y ,camera.viewUp.z);
	
	// if we have trackball rotation to map (this IS the test I want as it can be explicitly 0.0f)
	if ((gTrackingViewInfo == self) && gTrackBallRotation[0] != 0.0f) 
		glRotatef (gTrackBallRotation[0], gTrackBallRotation[1], gTrackBallRotation[2], gTrackBallRotation[3]);
	else {
	}
	// accumlated world rotation via trackball
	glRotatef (worldRotation[0], worldRotation[1], worldRotation[2], worldRotation[3]);
	[self updateCameraString];
}

// ---------------------------------

// handles resizing of GL need context update and if the window dimensions change, a
// a window dimension update, reseting of viewport and an update of the projection matrix
- (void) resizeGL
{
	NSRect rectView = [self bounds];
	
	// ensure camera knows size changed
	if ((camera.viewHeight != rectView.size.height) ||
	    (camera.viewWidth != rectView.size.width)) {
		camera.viewHeight = rectView.size.height;
		camera.viewWidth = rectView.size.width;
		
		glViewport (0, 0, camera.viewWidth, camera.viewHeight);
		[self updateProjection: YES];  // update projection matrix
	}
}

// ---------------------------------

// move camera in z axis
-(void)mouseDolly: (NSPoint) location
{
	GLfloat dolly = (gDollyPanStartPoint[1] -location.y) * -camera.viewPos.z / 300.0f;
	camera.viewPos.z += dolly;
	if (camera.viewPos.z == 0.0) // do not let z = 0.0
		camera.viewPos.z = 0.0001;
	if(camera.viewPos.z > 3.5) 
		camera.viewPos.z = 3.5;
	if(camera.viewPos.z < -3.5)
		camera.viewPos.z = -3.5;
	gDollyPanStartPoint[0] = location.x;
	gDollyPanStartPoint[1] = location.y;
}

// ---------------------------------

// move camera in x/y plane
- (void)mousePan: (NSPoint) location
{
	GLfloat panX = (gDollyPanStartPoint[0] - location.x) / (900.0f / -camera.viewPos.z);
	GLfloat panY = (gDollyPanStartPoint[1] - location.y) / (900.0f / -camera.viewPos.z);
	camera.viewPos.x -= panX;
	camera.viewPos.y -= panY;
	gDollyPanStartPoint[0] = location.x;
	gDollyPanStartPoint[1] = location.y;
}

// ---------------------------------

// sets the camera data to initial conditions
- (void) resetCamera
{
	camera.aperture = 40;
	camera.rotPoint = gOrigin;
	
	camera.viewPos.x = 0.0;
	camera.viewPos.y = 0.0;
	camera.viewPos.z = -2.0;
	camera.viewDir.x = -camera.viewPos.x; 
	camera.viewDir.y = -camera.viewPos.y; 
	camera.viewDir.z = -camera.viewPos.z;
	
	camera.viewUp.x = 0;  
	camera.viewUp.y = 1; 
	camera.viewUp.z = 0;
}

// ---------------------------------

// per-window timer function, basic time based animation preformed here
- (void)animationTimer:(NSTimer *)timer
{
	BOOL shouldDraw = NO;
	if (fAnimate) {
		CFTimeInterval deltaTime = CFAbsoluteTimeGetCurrent () - time;
		
		if (deltaTime > 10.0) {
			shouldDraw = NO;
		}
		else {
			shouldDraw = YES; // force redraw
		}
	}
	time = CFAbsoluteTimeGetCurrent (); //reset time in all cases
	if (YES == shouldDraw) 
		[self drawRect:[self bounds]]; // redraw now instead dirty to enable updates during live resize
}

#pragma mark ---- Text Drawing ----

// ---------------------------------

- (void) updateCameraString
{ // update info string texture
	static recCamera savedCamera; 
	
	// This is a compromise between a heavy comparison
	// and updating the camera texture when not needed
	// what is faster the comparison or the texture update
	// only empirical data on a particular configuration
	// will yield a real answer
	
	if  ( (savedCamera.viewPos.x == camera.viewPos.x) &&
		 (savedCamera.viewPos.y == camera.viewPos.y) &&
		 (savedCamera.viewPos.z == camera.viewPos.z) &&
		 (savedCamera.viewDir.x == camera.viewDir.x) &&
		 (savedCamera.viewDir.y == camera.viewDir.y) &&
		 (savedCamera.viewDir.z == camera.viewDir.z) &&
		 (savedCamera.aperture == camera.aperture) )
	{
		return; // Don't update texture! (which usually is more expensive than the comparison above)
	} else {
		NSString * string = [NSString stringWithFormat:@"Camera at (%0.1f, %0.1f, %0.1f) looking at (%0.1f, %0.1f, %0.1f) with %0.1f aperture", camera.viewPos.x, camera.viewPos.y, camera.viewPos.z, camera.viewDir.x, camera.viewDir.y, camera.viewDir.z, camera.aperture];
		if (camStringTex)
			[camStringTex setString:string withAttributes:stanStringAttrib];
		else {
			camStringTex = [[GLString alloc] initWithString:string withAttributes:stanStringAttrib withTextColor:[NSColor colorWithDeviceRed:1.0f green:1.0f blue:1.0f alpha:1.0f] withBoxColor:[NSColor colorWithDeviceRed:0.5f green:0.5f blue:0.5f alpha:0.5f] withBorderColor:[NSColor colorWithDeviceRed:0.8f green:0.8f blue:0.8f alpha:0.8f]];
		}
	}
	
	savedCamera = camera;
}

// ---------------------------------

// draw text info using our GLString class for much more optimized text drawing
- (void) drawInfo
{	
	GLint matrixMode;
	GLboolean depthTest = glIsEnabled (GL_DEPTH_TEST);
	GLfloat height, width, messageTop = 10.0f;
	
	height = camera.viewHeight;
	width = camera.viewWidth;
	
	// set orthograhic 1:1  pixel transform in local view coords
	glGetIntegerv (GL_MATRIX_MODE, &matrixMode);
	glMatrixMode (GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity ();
	glMatrixMode (GL_MODELVIEW);
	glPushMatrix();
	glLoadIdentity ();
	glScalef (2.0f / width, -2.0f /  height, 1.0f);
	glTranslatef (-width / 2.0f, -height / 2.0f, 0.0f);
	
	glColor4f (1.0f, 1.0f, 1.0f, 1.0f);
	[camStringTex drawAtPoint:NSMakePoint (10.0f, messageTop)];
	if([legend showLegend])
		[legendString drawAtPoint:NSMakePoint(10.0f, height - [legendString frameSize].height - 10.f)];
	// reset orginal martices
	glPopMatrix(); // GL_MODELVIEW
	glMatrixMode (GL_PROJECTION);
	glPopMatrix();
	glMatrixMode (matrixMode);
	
	glDisable (GL_TEXTURE_RECTANGLE_EXT);
	//glDisable (GL_BLEND);
	glEnable(GL_BLEND);
	if (depthTest)
		glEnable (GL_DEPTH_TEST);
	glReportError ();
}

#pragma mark ---- IB Actions ----


#pragma mark ---- Method Overrides ----

-(void)keyDown:(NSEvent *)theEvent
{
    NSString *characters = [theEvent characters];
    if ([characters length]) {
        unichar character = [characters characterAtIndex:0];
		switch (character) {
			case 'h':
				// toggle help
				fDrawHelp = 1 - fDrawHelp;
				[self setNeedsDisplay: YES];
				break;
			case 'c':
				// toggle caps
				fDrawCaps = 1 - fDrawCaps;
				[self setNeedsDisplay: YES];
				break;
		}
	}
}

// ---------------------------------

- (void)mouseDown:(NSEvent *)theEvent // trackball
{
    if ([theEvent modifierFlags] & NSControlKeyMask) // send to pan
		[self rightMouseDown:theEvent];
	else if ([theEvent modifierFlags] & NSAlternateKeyMask) // send to dolly
		[self otherMouseDown:theEvent];
	else {
		mouse_clicked = true;
		NSPoint location = [self convertPoint:[theEvent locationInWindow] fromView:nil];
		location.y = camera.viewHeight - location.y;
		gDolly = GL_FALSE; // no dolly
		gPan = GL_FALSE; // no pan
		gTrackball = GL_TRUE;
		startTrackball (location.x, location.y, 0, 0, camera.viewWidth, camera.viewHeight);
		gTrackingViewInfo = self;
	}
}

// ---------------------------------

- (void)rightMouseDown:(NSEvent *)theEvent // pan
{
	NSPoint location = [self convertPoint:[theEvent locationInWindow] fromView:nil];
	location.y = camera.viewHeight - location.y;
	if (gTrackball) { // if we are currently tracking, end trackball
		if (gTrackBallRotation[0] != 0.0)
			addToRotationTrackball (gTrackBallRotation, worldRotation);
		gTrackBallRotation [0] = gTrackBallRotation [1] = gTrackBallRotation [2] = gTrackBallRotation [3] = 0.0f;
	}
	gDolly = GL_FALSE; // no dolly
	gPan = GL_TRUE; 
	gTrackball = GL_FALSE; // no trackball
	gDollyPanStartPoint[0] = location.x;
	gDollyPanStartPoint[1] = location.y;
	gTrackingViewInfo = self;
}

// ---------------------------------

- (void)otherMouseDown:(NSEvent *)theEvent //dolly
{
	NSPoint location = [self convertPoint:[theEvent locationInWindow] fromView:nil];
	location.y = camera.viewHeight - location.y;
	if (gTrackball) { // if we are currently tracking, end trackball
		if (gTrackBallRotation[0] != 0.0)
			addToRotationTrackball (gTrackBallRotation, worldRotation);
		gTrackBallRotation [0] = gTrackBallRotation [1] = gTrackBallRotation [2] = gTrackBallRotation [3] = 0.0f;
	}
	gDolly = GL_TRUE;
	gPan = GL_FALSE; // no pan
	gTrackball = GL_FALSE; // no trackball
	gDollyPanStartPoint[0] = location.x;
	gDollyPanStartPoint[1] = location.y;
	gTrackingViewInfo = self;
}

// ---------------------------------

- (void)mouseUp:(NSEvent *)theEvent
{
	if (gDolly) { // end dolly
		gDolly = GL_FALSE;
	} else if (gPan) { // end pan
		gPan = GL_FALSE;
	} else if (gTrackball) { // end trackball
		gTrackball = GL_FALSE;
		if (gTrackBallRotation[0] != 0.0)
			addToRotationTrackball (gTrackBallRotation, worldRotation);
		gTrackBallRotation [0] = gTrackBallRotation [1] = gTrackBallRotation [2] = gTrackBallRotation [3] = 0.0f;
	} 
	gTrackingViewInfo = NULL;
	if(mouse_clicked) {
		mouse_clicked = false;
		[self mouseClicked: theEvent];
	}
}
- (void) mouseClicked:(NSEvent *) theEvent {
	NSPoint location = [self convertPoint:[theEvent locationInWindow] fromView:nil];
	const int BUF_SIZE = 64 * 5 * 5 * 4;
	GLuint select_buf[BUF_SIZE];
	GLint viewport[4];
	[[self openGLContext] makeCurrentContext];
	
	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();
	glGetIntegerv(GL_VIEWPORT, viewport);
	gluPickMatrix(location.x, location.y, 5.0, 5.0, viewport);
	[self updateProjection:NO];
	
	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	[self updateModelView];
	
	glSelectBuffer(BUF_SIZE, select_buf);
	glRenderMode(GL_SELECT);
	glInitNames();
	glPushName(0);
	[self renderMesh];
	glFlush();
	
	GLint hits = glRenderMode(GL_RENDER);
	
	glMatrixMode(GL_PROJECTION);
	glPopMatrix();
	glMatrixMode(GL_MODELVIEW);
	glPopMatrix();
	
	if(hits > 0) {
		GLuint * hit = &select_buf[0];
		GLuint hit_id = 0;
		GLuint minz = std::numeric_limits<GLuint>::max();
		for(int i = 0; i < hits; i++) {
			assert(hit[0] == 1);
			if(hit[1] < minz) {
				hit_id = hit[3];
				minz = hit[1];
			}
			hit += 4;
		}
		
		NSLog(@"CLICKED %f %f %d %d", location.x, location.y, hit_id >> 30, hit_id & ~(3 << 30));
		if(hit_id == 0)
			return;
		id_type elem_id = (hit_id & ~(3 << 30)) - 1;
		id_type elem_type = hit_id >> 30;
		ElemType et = (elem_type == VERTEX) ? VERTEX : (elem_type == EDGE) ? EDGE : (elem_type == FACE) ? FACE : (elem_type == CELL) ? CELL : CELL;
		[controller setElementSelection:et withID:elem_id atPosition: [NSEvent mouseLocation]];
	}
	glReportError();
}
// ---------------------------------

- (void)rightMouseUp:(NSEvent *)theEvent
{
	[self mouseUp:theEvent];
}

// ---------------------------------

- (void)otherMouseUp:(NSEvent *)theEvent
{
	[self mouseUp:theEvent];
}

// ---------------------------------

- (void)mouseDragged:(NSEvent *)theEvent
{
	mouse_clicked = false;
	NSPoint location = [self convertPoint:[theEvent locationInWindow] fromView:nil];
	location.y = camera.viewHeight - location.y;
	if (gTrackball) {
		rollToTrackball (location.x, location.y, gTrackBallRotation);
		[self setNeedsDisplay: YES];
	} else if (gDolly) {
		[self mouseDolly: location];
		[self updateProjection: YES];  // update projection matrix (not normally done on draw)
		[self setNeedsDisplay: YES];
	} else if (gPan) {
		[self mousePan: location];
		[self setNeedsDisplay: YES];
	}
}

// ---------------------------------

- (void)scrollWheel:(NSEvent *)theEvent
{
	float wheelDelta = [theEvent deltaX] +[theEvent deltaY] + [theEvent deltaZ];
	if (wheelDelta)
	{
		GLfloat deltaAperture = wheelDelta * -camera.aperture / 200.0f;
		/*
		camera.aperture += deltaAperture;
		if (camera.aperture < 0.1) // do not let aperture <= 0.1
			camera.aperture = 0.1;
		if (camera.aperture > 179.9) // do not let aperture >= 180
			camera.aperture = 179.9;*/
		camera.viewPos.z -= deltaAperture;
		if(camera.viewPos.z == 0.f)
			camera.viewPos.z = .001f;
		if(camera.viewPos.z > 3.5) 
			camera.viewPos.z = 3.5;
		if(camera.viewPos.z < -3.5)
			camera.viewPos.z = -3.5;
		[self updateProjection: YES]; // update projection matrix
		[self setNeedsDisplay: YES];
	}
}

// ---------------------------------

- (void)rightMouseDragged:(NSEvent *)theEvent
{
	[self mouseDragged: theEvent];
}

// ---------------------------------

- (void)otherMouseDragged:(NSEvent *)theEvent
{
	[self mouseDragged: theEvent];
}

// ---------------------------------

- (void) drawRect:(NSRect)rect
{		
	// setup viewport and prespective
	[self resizeGL]; // forces projection matrix update (does test for size changes)
	[self updateModelView];  // update model view matrix for object
	
	// clear our drawable
	glClearColor(.1,.1,.1,1);
	glClear (GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	
	
	// model view and projection matricies already set
	
	//drawCube (.5f); // draw scene
	//[self drawMesh];
    [self renderMesh];
	
	if (fInfo)
		[self drawInfo];
	
	if ([self inLiveResize] && !fAnimate)
		glFlush ();
	else
		[[self openGLContext] flushBuffer];
	glReportError ();
}

// ---------------------------------

// set initial OpenGL state (current context is set)
// called after context is created
- (void) prepareOpenGL
{
    GLint swapInt = 1;
	
    [[self openGLContext] setValues:&swapInt forParameter:NSOpenGLCPSwapInterval]; // set to vbl sync
	
	// init GL stuff here
	glEnable(GL_DEPTH_TEST);
	
	glShadeModel(GL_SMOOTH);    
	//glEnable(GL_CULL_FACE);
	glFrontFace(GL_CCW);
	glEnable(GL_POLYGON_OFFSET_FILL);
	glPolygonOffset (1.0f, 1.f);
	
	glClearColor(0.0f, 0.0f, 0.0f, 0.0f);
	[self resetCamera];
	shapeSize = 7.0f; // max radius of of objects
	mesh_list = glGenLists(1);
	// init fonts for use with strings
	NSFont * font =[NSFont fontWithName:@"Helvetica" size:12.0];
	stanStringAttrib = [[NSMutableDictionary dictionary] retain];
	[stanStringAttrib setObject:font forKey:NSFontAttributeName];
	[stanStringAttrib setObject:[NSColor whiteColor] forKey:NSForegroundColorAttributeName];
	[font release];
	
}
// ---------------------------------

// this can be a troublesome call to do anything heavyweight, as it is called on window moves, resizes, and display config changes.  So be
// careful of doing too much here.
- (void) update // window resizes, moves and display changes (resize, depth and display config change)
{
	[super update];
}

// ---------------------------------

-(id) initWithFrame: (NSRect) frameRect
{
	NSOpenGLPixelFormat * pf = [LisztGLView basicPixelFormat];
	
	self = [super initWithFrame: frameRect pixelFormat: pf];
    return self;
}

// ---------------------------------

- (BOOL)acceptsFirstResponder
{
	return YES;
}

// ---------------------------------

- (BOOL)becomeFirstResponder
{
	return  YES;
}

// ---------------------------------

- (BOOL)resignFirstResponder
{
	return YES;
}

// ---------------------------------
- (void) awakeFromNib
{
	requires_redraw = true;
	mouse_clicked = false;
	setStartTime (); // get app start time
	
	// set start values...
	fInfo = 1;
	fAnimate = 1;
	time = CFAbsoluteTimeGetCurrent ();  // set animation time start time

	// start animation timer
	//timer = [NSTimer timerWithTimeInterval:(1.0f/60.0f) target:self selector:@selector(animationTimer:) userInfo:nil repeats:YES];
	//[[NSRunLoop currentRunLoop] addTimer:timer forMode:NSDefaultRunLoopMode];
	//[[NSRunLoop currentRunLoop] addTimer:timer forMode:NSEventTrackingRunLoopMode]; // ensure timer fires during resize
    [self setNeedsDisplay: YES];
}


@end
